# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.



from qt import *
from qtui import QWidgetFactory

from GUI.rightdockwindow import RightDockWindow

from GUI.Workspace import *
from GUI.EditPrefs import EditPrefs
from GUI.Pixmaps.Icons import *
from GUI.HelpViewer import HelpViewer
from GUI.zmlcreator import * 
from GUI.AboutViewer import *
from GUI.SciCraftSplash import SciCraftSplash
from GUI.tools.madesigner import MADesigner

from Library.Util.Config import Config
from Library.Util.gui import EventReceiver
from Library.Util.rtester import RTester, RInstallError
from Library.Util.file import CacheDirectory

from System.EventServer import EventServer
from System.NodeRepository import NodeRepository
from System.NodeStatusDataList import NodeStatusDataList
from System.Plugin.PluginRepository import PluginRepository

from Library.Nodes import EditNodes, PlotNodes, FileHandlers, TestNodes
from Library.Nodes import UserInputNodes

try:
    from Library.debug.pycute import PyCute
except:
    PyCute = None

import sys
import signal
import thread
import time
import os

connect = QObject.connect

class Application(QObject, EventReceiver):

    
    def __init__(self, args, run):
        """init method"""
        QObject.__init__(self)
      
        EventReceiver.__init__(self)
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'mainwindow.ui'))
        self._widget.installEventFilter(self)
        self._widget.setGeometry(Config.appX, Config.appY, Config.appWidth,
                                 Config.appHeight)
        #style used in the qApp
        self._update_style() 
        self._widget.setEraseColor(QColor(Config.bgR, Config.bgG, Config.bgB))
        self.actions_menu = self._widget.child('Actions')
        self.window_menu = self._widget.child('Window')

        self._current_ma_designers = []

        # setup recently opened files
        for child in self._widget.children():
            if isinstance(child, QPopupMenu):
                for i in range(child.count()):
                    if str(child.text(child.idAt(i))) == "RecentFilesToken":
                        self._recent_files_menu = child
        connect(self._recent_files_menu, SIGNAL("aboutToShow()"), self._update_recent_files)
        connect(self._recent_files_menu, SIGNAL('activated(int)'), self._open_recent_file)

        splash = SciCraftSplash()

        self.thread_ident = thread.get_ident()
      
        self.progress_bar_widget = 0
        self.workspace = None
        
        eventServer = EventServer()

        # make sure events are handled by the EventReceiver interface
        eventServer.addEventHandler(self.post_event)

        node_modules = [EditNodes, PlotNodes, FileHandlers, TestNodes,
                        UserInputNodes]

        userfunctionpath = Config.loadpath
        functionpath = Config.systemfunctiondir

        pluginRepository = PluginRepository()
        
        splash.display_message("Loading available nodes")
        self.nodeRepository = NodeRepository(pluginRepository, node_modules,
                                             userfunctionpath, functionpath,
                                             event_server = eventServer)
        
        self.openPath = ""
        
        splash.display_message("Setting up user interface")
        self._widget.statusBar()

        self._cache = CacheDirectory(Config.cachepath)
        self._cache.set_enabled(Config.cacheenabled)
        
        splash.display_message("Initializing workspace")
        self._node_status_data_list = NodeStatusDataList(None)
        self.window_dict = {}
        self.checked_window = None

        splash.display_message("Setting up initial workspace")
        self.workspace = Workspace(self, eventServer, self.nodeRepository, 
                                   self._cache)

        self._init_connections()

        if args:
            for filename in args:
                filename = filename.replace('\x1b[00m', '')
                splash.display_message("Loading module diagram %s" % filename)
                if os.path.isfile(filename):                
                    self.openPath = os.path.realpath(filename)
                    self.workspace.load_diagram(filename, True)
        else:
            splash.display_message("Setting up empty diagram")
            self.workspace.new_diagram()

        dock_window = RightDockWindow(self._widget, 
                                      self.nodeRepository, 
                                      self.workspace,
                                      self.get_node_status_data_list())

        
        self._widget.moveDockWindow(dock_window, Qt.DockRight)
        dock_window.show()
        self._node_tree = dock_window.get_node_tree()
        splash.display_message("Starting SciCraft")

        # remove existing central widget before adding workspace
        self._widget.centralWidget().hide()
        self._widget.setCentralWidget(self.workspace)

        connect(qApp, SIGNAL("lastWindowClosed()"), self._clean_up_and_quit)

        self._update_play_and_stop_buttons()

        # C-c can kill SciCraft directly
        signal.signal(signal.SIGINT, self._clean_up_and_quit)
        # handle hangup
        if os.name == "posix":
            signal.signal(signal.SIGHUP, self._clean_up_and_quit)

        self._widget.statusBar().setFixedHeight(27)
        splash.hide()
        if run:
            self.slot_run_diagram()

    def _clean_up_and_quit(self, sig=None, stack_frame=None):
        self._slot_close_all_windows()
        self.save_config()

        # handle all pending events (info to user etc.)
        qApp.processEvents()

        try:
            RTester.r_install_test()
            import rpy
            rpy.r('quit(save="no")')
        except RInstallError, e:
            print e
            
        # probably never get here if rpy quits the entire application
        qApp.quit()

    def create_new_ma_designer(self, filename=None):
        designer = MADesigner(None, filename)
        designer.show()
        connect(designer, PYSIGNAL("new_editor(filename)"), self.create_new_ma_designer)
        self._current_ma_designers.append(designer)

    def show(self):
        self._widget.show()
        
    def _init_connections(self):
        self.window_menu.setCheckable(1)

        self.pointer_button = self._widget.child('modePointerAction')
        self.link_button = self._widget.child('modeLinkAction')
        self.comment_button = self._widget.child('modeInsertCommentAction')
        self.play_action = self._widget.child('actionsRunAction')
        self.stop_action = self._widget.child('actionsStopAction')

        connect(self.pointer_button, SIGNAL("activated()"), self.slot_point_diagram)
        connect(self.link_button, SIGNAL("activated()"), self.slot_link_diagram)
        connect(self.comment_button, SIGNAL("activated()"), self.slot_comment_diagram)
        connect(self.play_action, SIGNAL("activated()"), self.workspace.slot_run_diagram)
        connect(self.stop_action, SIGNAL("activated()"), self.workspace.slot_stop_diagram)
        connect(self._widget.child('actionsValidateAction'), SIGNAL("activated()"), self.workspace.slot_validate_diagram)
        connect(self._widget.child('fileNewAction'), SIGNAL("activated()"), self.slot_file_new)
        connect(self._widget.child('fileOpenAction'), SIGNAL("activated()"), self.slot_file_open)        
        connect(self._widget.child('fileSaveAction'), SIGNAL("activated()"), self.slot_file_save)        
        connect(self._widget.child('fileSaveAsAction'), SIGNAL("activated()"), self.slot_file_save_as)
        connect(self._widget.child('fileZMLCreatorAction'), SIGNAL("activated()"), self.slot_creator)
        connect(self._widget.child('fileEPSCreatorAction'), SIGNAL("activated()"), self.slot_eps_creator)
        connect(self._widget.child('filePreferencesAction'), SIGNAL("activated()"), self.slot_edit_prefs)
        connect(self._widget.child('fileCloseDiagramAction'), SIGNAL("activated()"), self._slot_close_diagram)

        connect(self._widget.child('windowMinimizeAllAction'), SIGNAL("activated()"), self._slot_minimize_alle)
        connect(self._widget.child('windowMaximizeAllAction'), SIGNAL("activated()"), self._slot_maximize_all)
        connect(self._widget.child('windowCloseAllAction'), SIGNAL("activated()"), self._slot_close_all_windows)

        connect(self._widget.child('helpHelpAction'), SIGNAL("activated()"), self.slot_help_scicraft)
        connect(self._widget.child('helpAboutQtAction'), SIGNAL("activated()"), self.slot_help_qt)        
        connect(self._widget.child('helpAboutSciCraftAction'), SIGNAL("activated()"),self.slot_help_about)
        connect(self._widget.child('toolsMADesignerAction'), SIGNAL("activated()"),self.create_new_ma_designer)
        connect(self._widget.child('hideHiddenLinksAction'), SIGNAL("activated()"), self._slot_hide_hidden_links)
        connect(self._widget.child('showHiddenLinksAction'), SIGNAL('activated()'), self._slot_show_hidden_links)
        
        if PyCute:
            python_interaction = QAction("Python interaction",
                                         "&Python Interaction",
                                         QAccel.stringToKey("CTRL+I"),
                                         self._widget)
            python_interaction.addTo(self._widget.child('Help'))
            connect(python_interaction, SIGNAL("activated()"), self._python_interaction)

    def _python_interaction(self):
        locals_ = {"app" : self, "workspace" : self.workspace,
                   "windows" : self.window_dict, "os" : os, "sys" : sys,
                   "qt" : qt}

        pycute = PyCute(locals_)
        pycute.show()

    def get_node_status_data_list(self):
        return self._node_status_data_list

    def get_node_tree(self):
        return self._node_tree

    def handle_event(self, event):
        """Called by EventReceiver interface to handle events in gui thread."""
        class_ = event.get_event_class()
        type_ = event.getType()
        if class_ == "DiagramEvent":
            self._update_play_and_stop_buttons()
        elif isinstance(event, InfoEvent):
            EventPopup(self._widget, event, modal=True)
        elif class_ == "ErrorEvent":
            EventPopup(self._widget, event)

    def _disable_stop(self):
        """Disables all buttons/menuitems that stops the diagram."""
        self.stop_action.setEnabled(False)
        
    def _disable_play(self):
        """Disables all buttons/menuitems that runs the diagram."""
        self.play_action.setEnabled(False)

    def _enable_stop(self):
        """Enables all buttons/menuitems that stops the diagram."""
        self.stop_action.setEnabled(True)
        
    def _enable_play(self):
        """Enables all buttons/menuitems that runs the diagram."""
        self.play_action.setEnabled(True)

    def _update_play_and_stop_buttons(self):
        """Updates the play and stop buttons after what state the current diagram is in."""
        time.sleep(0.1) # HACK: to be sure Modulediagram has a stable state...
        try:
            diagram = self.workspace.get_current_diagram()
            if len(self.window_dict) == 0:
                #no windows (diagrams)
                self._disable_play()
                self._disable_stop()
            elif len(diagram.get_nodes()) == 0:
                #no nodes in diagram
                # special case if diagram is empty (really HACK because stop-button
                # gets enabled after play-button has been pressed... sometimes...)
                self._enable_play()
                self._disable_stop()
            else:
                if not diagram.getIsRunning():
                    #diagram is NOT running
                    self._enable_play()
                    self._disable_stop()
                elif diagram.getIsInterrupted():
                    #diagram is interrupted
                    self._disable_play()
                    self._disable_stop()
                elif diagram.get_has_crashed():
                    # A node has crashed
                    self._enable_play()
                    self._disable_stop()                    
                else:
                    #diagram is running
                    self._disable_play()
                    self._enable_stop()
        except AttributeError, e:
            #AttributeError: when called from self.addWindow() self.workspace may not be created yet
            if e == 'workspace' and len(self.window_dict) > 0:
                self._enable_play()
                self._disable_stop()
    
    def set_progress_bar(self, progress):

        if self.progress_bar_widget == progress:
            return

        if self.progress_bar_widget and self.progress_bar_widget != progress:
            self.progress_bar_widget.hide()
            self._widget.statusBar().removeWidget(self.progress_bar_widget)
            
        self.progress_bar_widget = progress
        self.progress_bar_widget.resize(205, 24)
        self._widget.statusBar().addWidget(self.progress_bar_widget, 205, 1)
        self.progress_bar_widget.show()


    def get_actions_menu(self):
        self.actions_menu.clear()
        return self.actions_menu


    def _update_style(self):
        style = get_style() #get_style() is a function in the Config module
        qApp.setStyle(style) # qApp is a qApplication "shortcut", derived from Qt


    def _update_canvas(self):
        self.workspace.update_canvas()

        
    def slot_file_new(self):
        self.workspace.new_diagram()
        self._update_play_and_stop_buttons()


    def slot_file_open(self):
        
        fileName = QFileDialog.getOpenFileName(self.openPath,
                                               "SciCraft diagrams (*.zmd);;All files (*)",
                                               self._widget, "Open Diagram",
                                               "Choose a file").latin1()
        if fileName:
            fileName = str(fileName)
            self.openPath = os.path.realpath(fileName)
            if os.path.isfile(fileName):                
                self.workspace.load_diagram(fileName)
        
                
    def slot_file_save(self):
        self.workspace.save_diagram()
                

    def slot_file_save_as(self):
        self.workspace.save_as_diagram()
        
    def slot_edit_prefs(self):
        old_load_path = Config.loadpath
        # show widget
        prefs = EditPrefs(self._widget, self._cache)

        if prefs.exec_loop() == QDialog.Accepted:
            self._update_canvas()
            self._update_style()
            # has user changed loadpath?
            if Config.loadpath != old_load_path:
                node_modules = [EditNodes, PlotNodes, FileHandlers, TestNodes]
                userfunctionpath = Config.loadpath
                functionpath = Config.systemfunctiondir
                self.nodeRepository.updatePaths(node_modules, userfunctionpath, functionpath)
                self._node_tree.updateNodeList(dict(self.nodeRepository.getNodeList()))

            self._cache.set_path(Config.cachepath)
            self._cache.set_enabled(Config.cacheenabled)

    def slot_creator(self):
        zmlcreator = ZMLCreator(self._widget)

    def slot_eps_creator(self):
        self.workspace.save_eps_diagram()
              
    def slot_stop_diagram(self):
        self.workspace.slot_stop_diagram()
    
    def slot_run_diagram(self):
        self.workspace.slot_run_diagram()

    def _reset_toggle_tools(self, button_pressed):
        if (self.pointer_button.isOn() and button_pressed != self.pointer_button):
            self.pointer_button.toggle()
        if (self.link_button.isOn() and button_pressed != self.link_button):
            self.link_button.toggle()
        if (self.comment_button.isOn() and button_pressed != self.comment_button):
            self.comment_button.toggle()

    def enable_pointer_tool(self):
        self._reset_toggle_tools(self.pointer_button)
        if not self.pointer_button.isOn():
            self.pointer_button.toggle()
    
    def slot_point_diagram(self):
        self.enable_pointer_tool()
        self.workspace.set_pointing_tool_as_current_tool()
        
    def enable_link_tool(self):
        self._reset_toggle_tools(self.link_button)
        if not self.link_button.isOn():
            self.link_button.toggle()
    
    def slot_link_diagram(self):
        self.enable_link_tool()
        self.workspace.set_linking_tool_as_current_tool()

    def enable_comment_tool(self):
        self._reset_toggle_tools(self.comment_button)
        if not self.comment_button.isOn():
            self.comment_button.toggle()

    def slot_comment_diagram(self):
        self.enable_comment_tool()
        self.workspace.set_comment_tool_as_current_tool()
            
    def slot_help_scicraft(self):
        HelpViewer(self._widget, os.path.join(Config.htmldocpath, "index.html"),
                   Config.htmldocpath)
    
    
    def slot_help_about(self):
        viewer = AboutViewer(self._widget)
          
        
    def slot_help_qt(self):
        QMessageBox.aboutQt(self._widget, "About Qt")

    def _slot_minimize_alle(self):
        self.workspace.minimize_all_windows()
        for diagram in self.workspace.diagrams.values():
            diagram.showMinimized()

    def _slot_maximize_all(self):
        self.workspace.maximize_all_windows()

    def _slot_close_all_windows(self):
        self.workspace.close_all_windows()


    def save_config(self):
        Config.appWidth = self._widget.width()
        Config.appHeight = self._widget.height()

        if self._widget.x() >= 0:
            Config.appX = self._widget.x()
        if self._widget.y() >= 20:
            Config.appY = self._widget.y()
        else:
            Config.appY = 20
            
        write_config()

    def add_window(self, windowId, name):
        id = self.window_menu.insertItem(name, self.slot_window_clicked) # caller = self.workspace, not inited yet
                                                                    # when this function is called for the first time...
        self.window_menu.setItemParameter(id, int(windowId))
        if self.checked_window:
            self.window_menu.setItemChecked(self.checked_window, 0)
        self.window_menu.setItemChecked(id, 1)
        self.checked_window = id
        self.window_dict[windowId] = id
        self._update_play_and_stop_buttons()
                
    def remove_window(self, windowId):
        self.window_menu.removeItem(self.window_dict[windowId])
        del self.window_dict[windowId]
        self._update_play_and_stop_buttons()

    def slot_window_clicked(self, window_id):
        id = self.window_dict[str(window_id)]
        self.window_menu.setItemChecked(self.checked_window, 0)
        self.window_menu.setItemChecked(id, 1)
        self.checked_window = id
        self.workspace.window_change(window_id)
        self._update_play_and_stop_buttons()

    def _update_recent_files(self):
        self._recent_files_menu.clear()
        existing_files = Config.recentFiles.split(os.pathsep)
        for filename in filter(None, existing_files):
            self._recent_files_menu.insertItem(filename)

    def _open_recent_file(self, item_id):
        filename = str(self._recent_files_menu.text(item_id))
        self.openPath = os.path.realpath(filename)
        if os.path.isfile(filename):                
            self.workspace.load_diagram(filename)
        
    def _slot_close_diagram(self):
        self.workspace.close_active_window()

    def _slot_show_hidden_links(self):
        self.workspace.show_hidden_links()

    def _slot_hide_hidden_links(self):
        self.workspace.hide_hidden_links()
    

    def eventFilter(self, object_, event):
        """Filter that cheks for close events.

        The filter will return True if the event should be
        stopped and False otherwise.
        """
        
        return False

    
