# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os
import sys
import time
import thread
import tempfile

from qt import Qt, QScrollView, QColor, QSizePolicy, QRect, QPoint, QPixmap
from qt import QCursor, QMessageBox, QPen
from qtcanvas import QCanvasView, QCanvas, QCanvasRectangle

from GUI.Canvas.CanvasMarkLine import CanvasMarkLine
from GUI.Canvas.CanvasLine import CanvasLine
from GUI.Canvas.CanvasNode import CanvasNode
from GUI.Canvas.canvascomment import CanvasComment
from GUI.Canvas.tools import PointingTool, LinkingTool, CommentTool
from GUI.guinode import GUINode
from GUI.links import Link, LinkCreationError
from GUI.Nodes import PlotNode
from GUI.ProgressDockWindow import ProgressDockWindow
from GUI.EventPopup import EventPopup
from dprint import dprint, flags
from Library.Util.Config import Config
from Library.Util.gui import EventReceiver
from GUI.Pixmaps.Icons import Icons
from GUI.Pixmaps.cursors import Cursors
from GUI.Nodes.spreadsheetnode import SpreadSheetGUINode
from GUI.Nodes.rdissection import RDissectionGUINode
from System.events import StartDiagramEvent
from GUI.Nodes.multipleoutputnode import OutputGUINode, FileWriterGUINode
from GUI.Nodes.multipleinputnode import InputGUINode
from GUI.Nodes.guiuserquerynode import GUIUserQueryNode
from GUI.Nodes.guistructdissectionnode import GUIStructDissectionNode
    
# Helper functions for CanvasView when setting positions to new nodes


def bounding_rect(gui_elements):
    """Return bounding rectangle of a set of gui elements. This function
    requires that each element in the given list has functions left(), right(),
    top(), bottom().

    Returns a QRect
    """
    if not gui_elements:
        return QRect(0,0,0,0)

    first_rect = gui_elements[0].boundingRect()
    xmin = first_rect.left()
    ymin = first_rect.top()
    xmax = first_rect.right()
    ymax = first_rect.bottom()

    for gui_element in gui_elements[1:]:
        iter_rect = gui_element.boundingRect()
        xmin_ = iter_rect.left()
        ymin_ = iter_rect.top()
        xmax_ = iter_rect.right()
        ymax_ = iter_rect.bottom()
        
        xmin = min(xmin, xmin_)
        ymin = min(ymin, ymin_)
        xmax = max(xmax, xmax_)
        ymax = max(ymax, ymax_)

    return QRect(QPoint(xmin, ymin), QPoint(xmax, ymax))


def find_free_position(node, nodes, area_size):
    """Find free (x, y) position in area for placing node with node_size."""
    
    free_x = 0
    free_y = 0
    increment_x = 22 # horizontal space between nodes.
    increment_y = 30 # vertical spacing...
    
    while True:
        collided = False 
        for other_node in nodes:
            node_rect = (free_x, free_x + node.width(), free_y, free_y + node.height())
            if is_overlapping(node_rect, other_node):
                collided = True
        if not collided:
            return (free_x, free_y)
        free_x += increment_x
        
        if area_size[0]-free_x < (node.width()+increment_x):
            
            free_y += increment_y
            free_x = 0
            if area_size[1]-free_y < (node.height()):
                free_x = 0
                free_y = 0
                return (free_x, free_y)
            

def is_overlapping(node_rect, other_node):
    """Return if rectangle overlaps GUINode.

    node_rect consists of (x1, x2, y1, y2).
    other_node is GUINode, supporting x, y, width and height methods.
    """
    return max(node_rect[0], other_node.x()) < min(node_rect[1], other_node.x() + other_node.width()) and \
        max(node_rect[2], other_node.y()) < min(node_rect[3], other_node.y() + other_node.height())


def find_canvas_size(nodes_size, visible_size, visible_scrollbars_size):
    """Calculate size of canvas to show scrollbars only when necessary.

    nodes_size is (max_x, max_y) of all nodes
    visible_size is (width, height) of view without scrollbars
    visible_scrollbars_size is (width, height) of view with scrollbars

    Return (width, height) tuple of best canvas size.
    """
    if nodes_size[0] <= visible_size[0] and nodes_size[1] <= visible_size[1]:
        return visible_size
    else:
        canvas_width = max(nodes_size[0], visible_scrollbars_size[0])
        canvas_height = max(nodes_size[1], visible_scrollbars_size[1])
        return (canvas_width, canvas_height)


class CanvasView(QCanvasView, EventReceiver):

    def __init__(self, workspace, application, module_diagram,
                 event_server, filename, caption):

        self._workspace = workspace
        QCanvasView.__init__(self, self._workspace, "canvas view",
                             Qt.WDestructiveClose )
        EventReceiver.__init__(self)

        self.thread_ident = thread.get_ident()
        # canvas must be assigned to variable to keep from being deleted by GC
        self._canvas = QCanvas()
        self.setCanvas(self._canvas)
        
        self._workspace = workspace
        self._application = application        
        
        self._caption = caption
        
        self.event_server = event_server
        self._module_diagram = module_diagram

        self.canvas().setBackgroundColor(QColor(Config.bgR, Config.bgG, Config.bgB))
        
        self.setVScrollBarMode(QScrollView.Auto)
        self.setHScrollBarMode(QScrollView.Auto)

        event_server.addEventHandler(self.post_event)

        self._comments = [] # List with CanvasComment objects from module GUI.Canvas.canvascomment
        self._selected_items = [] # List of the currently selected items in the Canvas
        self._selected_link = None # The currently selected link
        self.nodes = {} # list with GUI.guinode.GUINode objects, indexed by node ModuleDiagramID
        self.links = []

        self._connect_line = 0
        self.selection_rectangle = QCanvasRectangle(0, 0, 0, 0, self._canvas)
        self.selection_rectangle.hide()
        self.selection_rectangle.setPen(QPen(Qt.DotLine))

        self.viewport().setMouseTracking(1)
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        self.missing_nodes = 0
        self.progress_bar = ProgressDockWindow(self._application._widget)
        self._node_status_data_list = self._application.get_node_status_data_list()
        self.nodelistbox = None
        self.activate() #sets filename
        self._module_diagram.setFilename(filename)

        self.setIcon(QPixmap(Icons.s))
        
        # enable drag and drop
        self.setAcceptDrops(1)
        
        self.populate_diagram()
    
        self._resize_canvas_to_fit()

        fd, self._temp_save = tempfile.mkstemp()
        os.close(fd)
        self.update_temp_file()

        self._pointing_tool = PointingTool(self)
        self._linking_tool  = LinkingTool(self)
        self._comment_tool  = CommentTool(self)
        self._current_tool  = None
        self._previous_tool = None
        self.set_pointing_tool_as_current_tool()


    def _reset_linking_tool(self):
        if self._current_tool == self._linking_tool:
            self.delete_connect_line()
            self._current_tool.reset_start_of_link()


    def set_pointing_tool_as_current_tool(self):
        self._reset_linking_tool()
        self._current_tool = self._pointing_tool
        self._application.enable_pointer_tool()
        self.unsetCursor()


    def set_linking_tool_as_current_tool(self):
        self._current_tool = self._linking_tool
        self._application.enable_link_tool()
        self.set_canvas_view_cursor(Cursors.cross)


    def set_comment_tool_as_current_tool(self):
        self._reset_linking_tool()
        self._current_tool = self._comment_tool
        self._application.enable_comment_tool()
        self.unsetCursor()


    def get_current_tool(self):
        return self._current_tool


    def set_previous_tool(self, tool):
        """Sets the previous tool used for this CanvasView"""

        self._previous_tool = tool


    def get_previous_tool(self):
        return self._previous_tool


    def update_temp_file(self):
        try:
            file_ = open(self._temp_save, 'w')
            file_.write(self._module_diagram.toXML())
        except IOError, e:
            pass

    
    def populate_diagram(self):
        for node in self._module_diagram.get_nodes():
            new_node = self._create_gui_node(node)
            self.nodes[node.getId()] = new_node
            new_node.getCanvasNode().newZ(2*len(self.nodes)-1)
            name = node.getNodeType()
            if name == "PlotNode":
                new_node.openPlotWindow(None)

        # creates canvas comments
        for comment in self._module_diagram.get_canvas_comments():
            gui_comment = CanvasComment(self, comment, False)
            gui_comment.redraw()
            self._comments.append(gui_comment)

        # gets a list of (from outputport, to inputport)-tuples.
        node_pairs = self._module_diagram.getConnections()

        # fill this list with new node->node connections
        linkable = {}

        for (output, input) in node_pairs:
            
            # get nodes
            fromnode = output.getOwnerNode()
            tonode = input.getOwnerNode()
                
            # if nodes haven't been added before, add them!
            if not linkable.has_key((fromnode.getId(), tonode.getId())):

                # position is simply where from the node the connection should be drawn
                topos = output.getPosition()
                linkable[fromnode.getId(), tonode.getId()] = (topos[0], topos[1])
        
        # add links between nodes with no port connections
        for (snode_id, enode_id) in self._module_diagram.get_connectionless_links():
            if not linkable.has_key((snode_id,enode_id)):
                topos = self.nodes[enode_id].getDiagramNode().getPosition()
                linkable[snode_id, enode_id] = (topos[0], topos[1])

        # create the actual objects!
        for keys in linkable:
            fromnode = self.nodes[keys[0]]
            tonode = self.nodes[keys[1]]        

            self.create_link(fromnode, tonode, linkable[keys][0], linkable[keys][1])

            self.first_point = 0
            self.first_node = 0                                    


    def create_canvas_comment(self, point, created_on_user_interaction = False):
        comment_data = self._module_diagram.add_canvas_comment("new comment", point.x(), point.y())
        new_comment = CanvasComment(self, comment_data, created_on_user_interaction)
        self._comments.append(new_comment)
        return new_comment


    def remove_canvas_comment(self, comment):
        self._module_diagram.del_canvas_comment(comment.get_canvas_comment_data())
        self._comments.remove(comment)

    def get_canvas_comments(self):
        return self._comments
    
    def _create_gui_node(self, node):
        """Factoryfunction for creating GUINode related to specific middleware Node."""
        
        node_class = node.getNodeRepName()
        if node_class.startswith("PlotNodes.PlotNode"):
            return PlotNode.PlotNode(self, node, self.event_server)
        elif node_class.startswith("PlotNodes.Plot2DNode"):
            from GUI.Nodes.guiplot2dnode import GUIPlot2DNode
            return GUIPlot2DNode(self, node, self.event_server)
        elif node_class.startswith("EditNodes.SpreadSheetNode"):
            return SpreadSheetGUINode(self, node)
        elif node_class.startswith("EditNodes.MoleculeEditorNode"):
            # imported here in case pymol doesnt exist, in which case it
            # raises an exception 
            from GUI.Nodes.MoleculeEditorNode import MoleculeEditorNode
            return MoleculeEditorNode(self, node)
        elif node_class.startswith("EditNodes.RDissectionNode"):
            return RDissectionGUINode(self, node)
        elif node_class.startswith("FileHandlers.InputNode"):
            return InputGUINode(self, node)
        elif node_class.startswith("FileHandlers.OutputNode"):
            return OutputGUINode(self, node)
        elif node_class.startswith("EditNodes.GaussdalNode"):
            from GUI.Nodes.gaussdal_node import Gaussdal_gui_node
            return Gaussdal_gui_node(self, node)
        elif node_class.startswith("PlotNodes.Customplots"):
            from GUI.Nodes.guiplot2dnode import GUIPlot2DNode
            return GUIPlot2DNode(self, node, self.event_server)
        elif node_class.startswith("FileHandlers.FileWriterNode"):
            return FileWriterGUINode(self, node)
        elif node_class.startswith("UserInputNodes.UserQueryNode"):
            return GUIUserQueryNode(self, node)
        elif node_class.startswith('EditNodes.StructDissectionNode'):
            return GUIStructDissectionNode(self, node)
        else:
            return GUINode(self, node)


    def create_node(self, node_class, x=None, y=None,):
        node = self._module_diagram.add_node_by_name(node_class, None)
        position_set = False

        if x is not None and y is not None:
            position_set = True
            node.setPosition(x, y)
            
        new_node = self._create_gui_node(node)
        
        if not position_set:
            new_position = find_free_position(new_node, self.nodes.values(), (self._workspace.width(), self._workspace.height()))
            new_node.setPosition(new_position[0], new_position[1])        
                        
        self.nodes[node.getId()] = new_node
        # Update node status dock window.
        self._node_status_data_list.update_list()
        
        new_node.getCanvasNode().newZ(2*len(self.nodes)-1)
        self.set_selected_items([new_node])
        self.dehighlight_none_selected_items()
        self.highlight_selected_items()
            
        dprint(flags.GUI.CREATE.NODE, "Creating a new GUI node. ID: " + str(node.getId()) + ", class: " + str(node_class))
        return new_node

        
    def remove_node(self, node):
        diagram_node = node.getDiagramNode()
        id = diagram_node.getId()
        self.nodes[id].wipePoints()
        del self.nodes[id]
        self._module_diagram.remove_node(diagram_node)
        # Update status list 
        self._node_status_data_list.update_list() 


    def create_connect_line(self, pos, node):
        self.bring_to_front(node)
        z = node.z()
        self._connect_line = CanvasMarkLine(self, self.canvas(), pos, z)


    def delete_connect_line(self):
        if self._connect_line:
            self._connect_line.delete()
            self._connect_line = 0


    def update_connect_line(self, point):
        self._connect_line.updatePosition(point)


    def create_link(self, from_node, to_node, from_point, to_point):
        """Create a link between two GUINode items.

        Creates a link at the exit/entry-point closest to given QPoint
        at each node.
        """
        for link in self.links: #first check if the link already exists
            if (from_node, to_node) == link.get_nodes():
                return

        connection = Link(self, self.canvas(), from_node, to_node, from_point, to_point)
        self.links.append(connection)
        return connection


    def get_links(self):
        return self.links


    def create_rectangle(self, point):
        self.selection_rectangle.setX(point.x())
        self.selection_rectangle.setY(point.y())
        self.selection_rectangle.setSize(0, 0)
        self.selection_rectangle.show()


    def update_rectangle(self, point):
        self.selection_rectangle.setSize(point.x()-
                                         self.selection_rectangle.x(),
                                         point.y()-
                                         self.selection_rectangle.y())


    def delete_rectangle(self):
        self.selection_rectangle.hide()
        

    def get_gui_elements_within_selection_rectangle(self):
        """Returns the gui elements within the selection rectangle."""
        if self.selection_rectangle:
            items = self.canvas().collisions(self.selection_rectangle.rect().normalize())
            
            node_selection = filter(lambda item : isinstance(item, CanvasNode), items)
            comment_selection = filter(lambda item : isinstance(item, CanvasComment), items)

            node_selection = map(lambda item : item.getParent(), node_selection)

            gui_element_selection = []
            gui_element_selection.extend(node_selection)
            gui_element_selection.extend(comment_selection)
            return gui_element_selection


    def set_canvas_view_cursor(self, pixmap_data):
        cursor_pixmap = QPixmap(pixmap_data)
        cursor = QCursor(cursor_pixmap)
        self.setCursor(cursor)


    def _resize_canvas_to_fit(self):
        canvas_items = self.nodes.values()
        canvas_items.extend(self._comments)
        rect = bounding_rect(canvas_items)
        
        xmax = rect.right()
        ymax = rect.bottom()

        visible_size = (self.visibleWidth(), self.visibleHeight())

        visible_scrollbars_size = (self.visibleWidth() - self.verticalScrollBar().width(),
                                   self.visibleHeight() - self.horizontalScrollBar().height())
        
        canvas_width, canvas_height = find_canvas_size((xmax, ymax),
                                                       visible_size,
                                                       visible_scrollbars_size)
        self.canvas().resize(canvas_width, canvas_height)


    def resizeEvent(self, e):
        """Method overridden from QCanvasView"""
        QCanvasView.resizeEvent(self, e)

        width = e.size().width()
        height = e.size().height()

        self._resize_canvas_to_fit()
            
        self._module_diagram.setVariable("canvasSize", (self.canvas().width(), self.canvas().height()))


    def _get_collided_gui_element(self, point):
        """Gets the gui element of interest with highest z-order in the
        collision list. If none found, the returned item is None.

        parameters:
           * point : The location where the mouse currently is

        """
        item_list = self.canvas().collisions(point)
        for item in item_list:
            if isinstance(item, CanvasNode) or isinstance(item, CanvasLine):
                return item.getParent()
            if isinstance(item, CanvasComment):
                return item
        return None
            
        
    def contentsMousePressEvent(self, e): # QMouseEvent e
        """Method overridden from QCanvasView"""
        point = self.inverseWorldMatrix().map(e.pos())
        gui_element = self._get_collided_gui_element(point)
        
        if e.button() == Qt.LeftButton: 
            try:
                self._current_tool.left_mouse_pressed(point, gui_element)
                self.progress_bar.setup()    
            except LinkCreationError, e:
                QMessageBox.warning(self, "Link creation error", str(e))

        elif e.button() == Qt.RightButton:
            self._current_tool.right_mouse_pressed(point, gui_element)

        self.setFocus() # need this since the dockwindows will grab the keyboard focus, do _not_ use grabKeyboard()
        self._canvas.update()
                
                
    def contentsMouseReleaseEvent(self, e):
        """Method overridden from QCanvasView"""
        point = self.inverseWorldMatrix().map(e.pos())
        gui_element = self._get_collided_gui_element(point)
        
        if e.button() == Qt.LeftButton:
            try:
                self._current_tool.left_mouse_released(point, gui_element)
                self.progress_bar.setup()
            except LinkCreationError, e:
                QMessageBox.warning(self, "Link creation error", str(e))
        elif e.button() == Qt.RightButton:
            self.progress_bar.setup()

        self._canvas.update()


    def contentsMouseMoveEvent(self, e):
        """Method overridden from QCanvasView"""

        point = self.inverseWorldMatrix().map(e.pos())
        self._current_tool.mouse_moved(point)
        if e.state() == Qt.LeftButton:
            self._canvas.update()


    def contentsMouseDoubleClickEvent(self, e):
        """Method overridden from QCanvasView"""

        point = self.inverseWorldMatrix().map(e.pos())
        gui_element = self._get_collided_gui_element(point)
        if e.button() == Qt.LeftButton:
            self._current_tool.left_mouse_double_clicked(point, gui_element)


    def keyPressEvent(self, e):        
        """Method overridden from QCanvasView"""
        self._current_tool.key_pressed(e.key())

                
    def bring_to_front(self, item):
        """A node has two z-values; the node icon and label have value z and the label text z + 1. """
        old_z = item.z()
        item.newZ(2 * len(self.nodes) - 1)
        for node in self.nodes.values():
            view_node = node.getCanvasNode()
            if view_node != item and view_node.z() > old_z:
                view_node.newZ(view_node.z() - 2)


    def update(self):
        """updates the background color. Method overridden from QCanvasView."""
        
        self.canvas().setBackgroundColor(QColor(Config.bgR, Config.bgG, Config.bgB))

        for line in self.links: # updates line-color
            line.update()
            
        self.canvas().update()


    def save_node_settings(self):
        """Save GUI settings for all nodes."""
        for node in self.nodes.values():
            node.saveGuiSettings()


    def get_nodes(self):
        return self.nodes


    def get_diagram_id(self):
        return self._module_diagram.getId()

    
    def handle_event(self, event):
        """Function that will be called by eventserver with all events sent 
            from the diagram. All events that is handled by this eventHandler
            will be in the main event thread through EventReceiver(superclass).
        """
        dprint(flags.GUI.CANVAS.CANVASVIEW,
               "[CanvasView]:",
               "handle_event receieved event", event,
               event.get_event_class(), event.getType(), repr(event.getDescription()))

        # need to check if this event is for the current diagram
        owner_id = event.get_module_diagram_id()
        if owner_id != self._module_diagram.getId():
            return
        
        class_ = event.get_event_class()
        type_ = event.getType()
        description = event.getDescription()
        if class_ == "DiagramEvent":
            if type_ == "START":
                if isinstance(event, StartDiagramEvent):
                    num_nodes = event.get_num_nodes()
                else:
                    num_nodes = len(self.nodes)
                    
                self.progress_bar.setup(num_nodes)
                self.progress_bar.startProgress()
            if type_ == "FINISH":
                self.progress_bar.setFinished()
            if type_ == "HALT":
                self.progress_bar.setHalted()
            if type_ == "EXCEPTION":
                if not self.progress_bar.isCrashed():
                    self.progress_bar.setCrashed()

        elif class_ == "NodeEvent":
            # Update node status on all node events.
            self._node_status_data_list.update_list()
            if type_ == "FINISH":
                if not self._module_diagram.getIsInterrupted():
                    self.progress_bar.addProgress(1)
            elif type_ == "EXCEPTION":
                self.progress_bar.setCrashed()
                EventPopup(self._application._widget, event)
            elif type_ == "INFO":
                EventPopup(self._application._widget, event)

    
    def activate(self, filename = 0):
        if filename:
            self._module_diagram.setFilename(filename)
            self._caption = os.path.basename(filename)
            self.setCaption(self._caption)
        else:
            self.setCaption(self._caption)
            
        self._application.set_progress_bar(self.progress_bar)
        self._node_status_data_list.update_list()
        self.progress_bar.refreshBar()


    def closeEvent(self, e):
        """Method overridden from QCanvasView"""
        self._module_diagram.on_close()
        self._workspace.remove_canvas(self._module_diagram.getId())            
        e.accept()

        
    def get_filename(self):
        return self._module_diagram.getFilename()


    def to_xml(self):
        """Create XML from module diagram."""
        return self._module_diagram.toXML()


    def to_eps(self, filename,colour_mode):
        """Creates an eps-image file of the scicraft diagram through pyx"""
        
        try:
            from pyx import color, deco, canvas, unit, box, path, style

            #pyx canvas with unit "points"
            unit.set(defaultunit="pt")
            c = canvas.canvas()
        
            for link in self.links:
                mi = 0
                while mi < (len(link.line_points)-1):
                    p1 = link.line_points[mi]
                    p2 = link.line_points[mi+1]
                    
                    #if last point, add arrow-head
                    if mi == (len(link.line_points)-2):
                        #removes whitespace at end of line before arrow-head - must see to believe :)
                        add_x2=0
                        if p1.x()>=p2.x():
                            add_x2 = -4
                        else:
                            add_x2 = 4
    
                        c.stroke(path.line(p1.x(),0-p1.y(), p2.x()+add_x2, 0-p2.y()),[deco.earrow.normal])
                    else:
                        #Just draw line
                        c.stroke(path.line(p1.x(),0-p1.y(), p2.x(), 0-p2.y()))

                    mi+=1
                

            for node in self.nodes.values():
                new_name = node.getDiagramNode().getName().replace("Node","")

                node_color= color.rgb.white
                if(colour_mode):
                    if node.nodeClass.startswith("EditNodes."):
                        node_color = color.cmyk.Peach
                    elif node.nodeClass.startswith("FileHandlers."):
                        node_color = color.cmyk.Goldenrod
                    elif node.nodeClass.startswith("Functions."):
                        node_color = color.cmyk.Periwinkle
                    elif node.nodeClass.startswith("LogicNodes."):
                        node_color = color.cmyk.Dandelion
                    elif node.nodeClass.startswith("PlotNodes."):
                        node_color = color.cmyk.OliveGreen
                    elif node.nodeClass.startswith("TestNodes."):
                        node_color = color.cmyk.RedOrange

                #pyx node
                myrect0 = box.rect(node.x(), 0-node.y()-40,node.width(),node.height())
                c.stroke(myrect0.path(bezierradius=3),[color.cmyk.Black,style.linewidth.thick])
                c.fill(myrect0.path(bezierradius=3),[node_color])

                #pyx node-label
                myrect1 = box.rect(node.x()+1.2,0-node.y()-48.3,5.5*len(new_name),9.7)
                c.stroke(myrect1.path(bezierradius=0.0),[color.rgb.black])
                c.fill(myrect1.path(bezierradius=0.0),[color.rgb.white])

                #pyx label-text
                c.text(node.x()+2.2,0-node.y()-46.3,new_name,[color.rgb.black])

            try:
                c.writeEPSfile(filename)
            except AttributeError:
                exit = QMessageBox.information(self, "Message", "The diagram is empty, nothing to export.")

        except ImportError, e:
            exit = QMessageBox.information(self, "Message", "PyX must be installed for this to work.")
    

    def validate_diagram(self):
        """Validate module diagram.
        Return:
                a list consisting of:
                    - List of NodeInputPort which doesn't have their data set
                    - List of nodes which have not defined all parameters"""
        
        return self._module_diagram.validate()
        

    def run_diagram(self):
        """Start execution of module diagram."""
        self._module_diagram.start()

    
    def get_module_diagram(self):
        return self._module_diagram

    
    def dragEnterEvent(self, event):
        """Method overridden from QCanvasView"""
        # Drag-and-drops entering this QWidget are accepted"""
        
        if event.format().startswith("text/plain"):
            if str(event.encodedData("text/plain")).startswith("SciCraft node"):
                event.accept()
        
    
    def dropEvent(self, event):
        """Method overridden from QCanvasView. Create a new node where the node icon is dropped"""
        
        if event.format().startswith("text/plain"):
            if str(event.encodedData("text/plain")).startswith("SciCraft node"):
                dragObject = event.source().dragObject

                if not (dragObject.pixmapWidth or dragObject.pixmapHeight):
                    return
                
                dropNodeName = dragObject.nodeName
                dropPos = self.viewportToContents(event.pos())
                dropX = dropPos.x() - (dragObject.pixmapWidth / 2) - 2
                dropY = dropPos.y() - (dragObject.pixmapHeight / 2) - 2

                if dropX < 0:
                    dropX = 0
                elif dropX + dragObject.pixmapWidth > self.canvas().width():
                    dropX = self.canvas().width() - dragObject.pixmapWidth        
               
                if dropY < 0:
                    dropY = 0
                elif dropY + dragObject.pixmapHeight > self.canvas().height():
                    dropY = self.canvas().height() - dragObject.pixmapHeight
               
                self.create_node(dropNodeName, dropX, dropY)
                dragObject.source().nodeItem = None
                dragObject.source().dragObject = None


    def _check_save_needed(self):
        """Returnes True if the diagram needs to be changed"""
        try:
            file_ = open(self._temp_save, 'r')
            xml = file_.read()
        except IOError:
            print 'Could not read temp file'
            
        return not(xml == self._module_diagram.toXML())


    def show_hidden_links(self):
        for link in self.links:
            link.show_hidden_links()

                
    def hide_hidden_links(self):
        for link in self.links:
            link.hide_hidden_links()


    def set_selected_link(self, link):
        """ Sets the currently selected link.

        parameters:
           * link : the selected link of type Link
        """
        self._selected_link = link


    def get_selected_link(self):
        return self._selected_link


    def reset_selected_link(self):
        self._selected_link = None


    def set_selected_items(self, item_list):
        """ Sets the currently selected items.

        parameters:
           * item_list: a list of currently selected items.
        """
        self._selected_items = item_list


    def get_selected_items(self):
        return self._selected_items


    def reset_selected_items(self):
        self._selected_items = []


    def dehighlight_none_selected_items(self):
        # dehighlights all nodes and links which aren't selected
        for link in self.links:
            link.deSelect()
        for node in self.nodes.values():
            if not node in self._selected_items:
                node.deSelect()
        for comment in self._comments:
            if not comment in self._selected_items:
                comment.deselect()

    def highlight_selected_items(self):
        # highlights the node(s) currently selected
        if self._selected_link:
            self._selected_link.select()
        elif self._selected_items:
            for gui_element in self._selected_items:
                gui_element.select()          

#check = QMessageBox.question(self, "Diagram changed",
#                                         "Save changes to diagram %s" % str(self.caption()),
#                                         "&Save", "&Discard", "Cancel")#, 
