# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.links import Link
from GUI.guinode import GUINode
from GUI.Canvas.canvascomment import CanvasComment

from qt import QPoint, Qt


class AbstractTool:
    """This class is must be inherited by the other tools in this module. This
    class also defines standard behaviour functionality not dependent on tools."""

    
    def __init__(self, canvas_view):
        self._canvas_view = canvas_view

    def left_mouse_released(self, pos, gui_element):
        pass

    def mouse_moved(self, pos):
        pass


    def left_mouse_double_clicked(self, pos, gui_element):
        pass

    
    def right_mouse_pressed(self, point, gui_element):
        self._canvas_view.reset_selected_items()
        self._canvas_view.reset_selected_link()
        
        if isinstance(gui_element, Link):
            self._canvas_view.set_selected_link(gui_element)
        elif isinstance(gui_element, GUINode):
            self._canvas_view.set_selected_items([gui_element])
        else:
            return

        self._canvas_view.dehighlight_none_selected_items()
        self._canvas_view.highlight_selected_items()
                    
        gui_element.contextMenu(point)


    def key_pressed(self, key):
        if key == Qt.Key_Delete:
            # the user pressed del
            if self._canvas_view.get_selected_link():
                self._canvas_view.get_selected_link().delete()
                self._canvas_view.reset_selected_link()
            elif self._canvas_view.get_selected_items():
                for gui_element in self._canvas_view.get_selected_items():
                    gui_element.delete()
                self._canvas_view.reset_selected_items()
            
        elif key == Qt.Key_Escape:
            #user pressed ESC-key
            if isinstance(self._canvas_view.get_current_tool(), LinkingTool):
                self._canvas_view.delete_connect_line()
                self.reset_start_of_link()
                self._canvas_view.set_pointing_tool_as_current_tool()
            elif isinstance(self._canvas_view.get_current_tool(), CommentTool):
                self._canvas_view.set_pointing_tool_as_current_tool()

            self._canvas_view.reset_selected_link()
            self._canvas_view.reset_selected_items()
            self._canvas_view.dehighlight_none_selected_items()


class PointingTool(AbstractTool):
    def __init__(self, canvas_view):
        AbstractTool.__init__(self, canvas_view)
        # the current last position of the move of the gui element
        self._current_last_pos = None
        self._move_initiated = False
        self._selection_initiated = False

    def left_mouse_double_clicked(self, pos, gui_element):
        if isinstance(gui_element, Link):
            gui_element.linkDialog()
        elif isinstance(gui_element, GUINode):
            gui_element.nodeDialog()
        elif isinstance(gui_element, CanvasComment):
            gui_element.comment_dialog()
        else:
            pass

    def left_mouse_pressed(self, pos, item_pressed):
        gui_element = None
        
        if isinstance(item_pressed, Link):
            # if the item is of type Link we should select it and reset
            # rubberband selection
            self._canvas_view.set_selected_link(item_pressed)
            self._canvas_view.reset_selected_items()
            self._canvas_view.dehighlight_none_selected_items()
            self._canvas_view.highlight_selected_items()
            return
        elif isinstance(item_pressed, GUINode) or \
                 isinstance(item_pressed, CanvasComment):
            # if the item is of type GUINode or CanvasComment we should set the
            # gui_element parameter
            gui_element = item_pressed
        elif item_pressed == None:
            # if there isn't given an item, then we continue since user is about
            # to make a selection or just clicking on an empty space
            gui_element = None
        else:
            # if the item is of unknown type we should not handle it
            return

        if not gui_element:
            # if no gui element is pressed, we wish to make a new
            # selection. link should also be reset
            self._canvas_view.reset_selected_items()
            self._canvas_view.reset_selected_link()
        if self._canvas_view.get_selected_items().count(gui_element) == 0:
            # if the gui element pressed is not a part of the selection, we wish
            # to cancel the selection
            self._canvas_view.reset_selected_items()
        if not self._canvas_view.get_selected_items():
            # if there is currently not a selection we should either create one
            # by rubberband selection or by just adding a pressed gui element to
            # the selection (and activate moving on that element)
            if gui_element:
                # rubberband selection not activated from this point
                self._canvas_view.reset_selected_link()
                self._current_last_pos = pos
                self._move_initiated = True
                self._canvas_view.set_selected_items([gui_element])
                self._canvas_view.dehighlight_none_selected_items()
                self._canvas_view.highlight_selected_items()
            else:
                # rubberband selection activated from this point
                self._selection_initiated = True
                self._canvas_view.create_rectangle(pos)
        else:
            # if there is a selection and the gui_element pressed is a part of
            # it we should active moving on the selection
            if self._canvas_view.get_selected_items().count(gui_element) == 1:
                self._current_last_pos = pos
                self._move_initiated = True

                
    def left_mouse_released(self, pos, gui_element = None):
        if self._selection_initiated:
            # if a selection is initiated we should get the selected gui
            # elements, highlight them, and then reset selection initiation
            selected_items = self._canvas_view.get_gui_elements_within_selection_rectangle()
            self._canvas_view.set_selected_items(selected_items)
            self._canvas_view.delete_rectangle()
            self._canvas_view.dehighlight_none_selected_items()
            self._canvas_view.highlight_selected_items()
            self._selection_initiated = False
        else:
            # if a selection isn't initiated we should reset move variables
            self._current_last_pos = None
            self._move_initiated = False

        
    def mouse_moved(self, pos):
        if self._move_initiated:

            from GUI.Canvas.CanvasView import bounding_rect

            move_to_x = pos.x()
            move_to_y = pos.y()
        
            # find bounding rect (xmin, ymin, xmax, ymax) of selected gui
            # elements
            selection_rect = bounding_rect(self._canvas_view.get_selected_items())
            xmin = selection_rect.left()
            ymin = selection_rect.top()
            xmax = selection_rect.right()
            ymax = selection_rect.bottom()

            # checking if outside boundaries. if it is update "move_to"-variables
            if xmax + (pos.x() - self._current_last_pos.x()) >= self._canvas_view.canvas().width():
                move_to_x = self._current_last_pos.x() + self._canvas_view.canvas().width() - xmax - 1
            elif xmin + (pos.x() - self._current_last_pos.x()) <= 0:
                move_to_x = self._current_last_pos.x() - xmin

            if ymax + (pos.y() - self._current_last_pos.y()) >= self._canvas_view.canvas().height():
                move_to_y = self._current_last_pos.y() + self._canvas_view.canvas().height() - ymax - 1
            elif ymin + (pos.y() - self._current_last_pos.y()) <= 0:
                move_to_y = self._current_last_pos.y() - ymin

            # updates the position for gui elements
            for gui_element in self._canvas_view.get_selected_items():
                x_inc = move_to_x - self._current_last_pos.x()
                y_inc = move_to_y - self._current_last_pos.y()
                gui_element.move(x_inc, y_inc)
                
            self._current_last_pos = QPoint(move_to_x, move_to_y)
        else:
            if self._selection_initiated:
                self._canvas_view.update_rectangle(pos)
                selected_items = self._canvas_view.get_gui_elements_within_selection_rectangle()
                self._canvas_view.set_selected_items(selected_items)
                self._canvas_view.dehighlight_none_selected_items()
                self._canvas_view.highlight_selected_items()

    
class LinkingTool(AbstractTool):
    def __init__(self, canvas_view):
        AbstractTool.__init__(self, canvas_view)
        self._link_created_from_context_menu = False
        self._start_node = None
        self._start_pos = None

    def left_mouse_pressed(self, position, gui_node):
        """Handle mouse pressed on node."""
        if isinstance(gui_node, GUINode):
            # the item pressed is of type GUINode and we should handle it
            pass
        elif gui_node == None:
            # the item pressed is a blank space and we should handle it
            pass
        else:
            # if the item pressed is of unknown type we shouldn't handle it
            return
        
        if not self._start_node:
            if gui_node:
                self.set_start_of_link(gui_node, position)
        elif gui_node == None or gui_node == self._start_node:
            self.reset_start_of_link()
            self._canvas_view.delete_connect_line()
            self._update_current_tool_if_link_created_from_context_menu()

            
    def left_mouse_released(self, position, gui_node):
        """Handle mouse released on node."""
            
        # if a start node hasn't been activated a link shouldn't be created
        if not self._start_node:
            return
        # if start_node equals end_node we shouldn't create a link
        if self._start_node == gui_node:
            self._update_current_tool_if_link_created_from_context_menu()
            return
        # no link created when releasing on a non-node
        if gui_node == None or not isinstance(gui_node, GUINode):
            self.reset_start_of_link()
            self._canvas_view.delete_connect_line()
            self._update_current_tool_if_link_created_from_context_menu()
            return

        self._canvas_view.create_link(self._start_node, gui_node,
                                      self._start_pos, position)
        self._start_node = None
        self._canvas_view.delete_connect_line()
        self._update_current_tool_if_link_created_from_context_menu()


    def mouse_moved(self, position):
        if self._start_node:
            self._canvas_view.update_connect_line(position)


    def set_link_created_from_context_menu(self):
        self._link_created_from_context_menu = True


    def set_start_of_link(self, start_node, start_pos):
        self._start_node = start_node
        self._start_pos = start_pos
        self._canvas_view.create_connect_line(start_pos, self._start_node)
        self._canvas_view.update_connect_line(start_pos)


    def reset_start_of_link(self):
        self._start_node = None
        self._start_pos = None


    def _update_current_tool_if_link_created_from_context_menu(self):
        """if the link was created from a context menu we should set the current
        tool to the previously used tool"""
        
        if self._link_created_from_context_menu == True:
            previous_tool = self._canvas_view.get_previous_tool()
            if previous_tool:
                if isinstance(previous_tool, PointingTool):
                    self._canvas_view.set_pointing_tool_as_current_tool()
                elif isinstance(previous_tool, LinkingTool):
                    self._canvas_view.set_linking_tool_as_current_tool()
                self._link_created_from_context_menu = False
                self._canvas_view.set_previous_tool(None)


class CommentTool(AbstractTool):
    def __init__(self, canvas_view):
        AbstractTool.__init__(self, canvas_view)


    def left_mouse_pressed(self, point, gui_element=None):
        comment = self._canvas_view.create_canvas_comment(point, True)
        self._canvas_view.set_pointing_tool_as_current_tool()
