#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QApplication
from GUI.guinode import GUINode
from Library.Util.gui import size, unsize, unqstring
from Library.Util.gui import UiWrapper, ComboWrapper, InputOutputView


class RDissectionGUINode(GUINode):

    def nodeDialog(self):
        """Called by CanvasView."""
        self.window = PickWindow(self._cv, self.node)
        self.window.show()


class PickWindow:
    """The GUI window for this node."""

    def __init__(self, parent, libnode):
        self._libnode = libnode
        self._parent = parent

        # Setup widgets
        self._ui = UiWrapper(parent, "rdissect")
        self._ui.autoconnect(self)
        self._ui.lsv_Input.setRootIsDecorated(True)
        self._mainwin = self._ui.get_widget()
        forest = self._libnode.get_forest()
        selection = self._libnode.get_selection()
        selection_name = self._libnode.get_selection_name(selection)
        self._ioview = InputOutputView(self._ui.lsv_Input, self._ui.lsv_Output, forest, selection)
        self._ioview.set_widgets(self._ui.btn_Clear, self._ui.btn_Remove, self._ui.btn_Add)
        self._combo = ComboWrapper(self._ui.cmb_Load, synclist=[self._ui.btn_Delete])
        self._selectioncollection = self._libnode.get_selectioncollection()
        self._combo.fill(self._selectioncollection.keys())
        if selection_name and selection_name in self._combo:
            self._combo.set_selected(selection_name)
        if self._combo.has_data():
            self.cmb_Load(self._ui.cmb_Load.currentText())
        else:
            self._ui.cmb_Load.setEnabled(False)
            self._ui.btn_Delete.setEnabled(False)
        self.sync_enabled()
        
        # Position the window nicely on the screen
        self._desksize = size(QApplication.desktop())
        self._mainsize = size(self._mainwin)
        self._mainwin.move(*unsize((self._desksize - self._mainsize) / 3))

    def show(self):
        """Is called by the guinode."""
        self._mainwin.show()

    def update_savebtn(self, qstring=None):
        """Disables or enables the save-button, depending on conditions."""
        if qstring:
            name = unqstring(qstring)
        else:
            name = unqstring(self._ui.txt_Name.text())
        ok_name = bool(name) and name not in self._combo
        has_output = self._ioview.has_data()
        saveable = ok_name and has_output
        if saveable != self._ui.btn_Save.isEnabled():
            self._ui.btn_Save.setEnabled(saveable)

    def sync_enabled(self):
        """Disable the useless buttons."""
        self._ioview.sync_enabled()

    def cmb_Load(self, qstring):
        """Changes the current selection."""
        name, selection = self._libnode.load_from_gui(unqstring(qstring))
        self._ioview.clear()
        self._ioview.set_selection(selection)
        if name in self._combo:
            self._ioview.set_dirty(False)
        else:
            self._ui.txt_Name.setText(name) 
            self._ioview.set_dirty(True)
        self.update_savebtn()
        self.sync_enabled()

    def btn_Delete(self):
        """Removes a selection from the combobox."""
        name = self._combo.remove_selected()
        if self._combo.has_data():
            self.cmb_Load(self._ui.cmb_Load.currentText())
        self._libnode.remove_from_gui(name)
        self.sync_enabled()

    def btn_Save(self):
        """Saves a selection."""
        name = unqstring(self._ui.txt_Name.text())
        self._libnode.save_from_gui((name, self._ioview.get_active_selection()))
        self._combo.append(name)
        self._ui.txt_Name.clear()
        self._ioview.set_dirty(False)
        self.update_savebtn()

    def btn_Ok(self):
        """Accepts the current selection."""

        dirty = self._ioview.is_dirty()
        loaded = self._combo.has_data()
        named = self._ui.btn_Save.isEnabled()
        empty = not self._ioview.has_data()

        # Get the correct name
        name = None
        if dirty:
            if named:
                if not empty:
                    self.btn_Save()
                    name = self._combo.get_selected()
            else:
                if loaded:
                    name = self._combo.get_selected()
                    if name.endswith("(unsaved)") or name.endswith("*"):
                        name += "*"
                    else:
                        name += " (unsaved)"
                    self._libnode.save_from_gui((name, self._ioview.get_active_selection()))
                else:
                    name = "Untitled"
        elif loaded:
            name = self._combo.get_selected()

        # No can do if we don't have a name
        if name is None:
            self._mainwin.close()
            return
            
        # Send the active selection to the libnode
        active_selection = name, self._ioview.get_active_selection()
        self._libnode.ok_from_gui(active_selection)

        # That's all folks!
        self._mainwin.close()

    def btn_Cancel(self):
        """Discards the current selection."""
        self._mainwin.close()

    def btn_Add(self):
        """Adds an output port."""
        self._ioview.add_selected()
        self.update_savebtn()
        self.sync_enabled()

    def btn_Remove(self):
        """Removes an output port."""
        self._ioview.remove_selected()
        self.update_savebtn()
        self.sync_enabled()

    def btn_Clear(self):
        """Removes all output ports."""
        self._ioview.clear()
        self.update_savebtn()
        self.sync_enabled()

    def lsv_Output_doubleClicked(self, item):
        """Adds an output port."""
        self._ioview.remove_from_output(item) 
        self.update_savebtn()
        self.sync_enabled()

    def lsv_Input_doubleClicked(self, item):
        """Removes an output port."""
        self._ioview.add_to_output(item)
        self.update_savebtn()
        self.sync_enabled()

    def lsv_Input_selectionChanged(self):
        """Updates the buttons."""
        self.sync_enabled()

    def lsv_Output_selectionChanged(self):
        """Updates the buttons."""
        self.sync_enabled()

    def txt_Name_changed(self, qstring):
        """Only enable the Save button under certain conditions."""
        self.update_savebtn(qstring)

    def txt_Name_returnPressed(self):
        """Pressing return should be like pressing the Save button."""
        self.btn_Save()
