# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QObject, SIGNAL, PYSIGNAL
import Numeric
import types

connect = QObject.connect
emit = QObject.emit

def _is_string_list(data):
    if type(data) in types.StringTypes:
        return False
    try:
        if type(data[0]) in types.StringTypes:
            return True
        else:
            return False
    except:
        return False

def _is_string_matrix(data):
    try:
        if type(data) in types.StringTypes:
            return False
        if type(data[0]) in types.StringTypes:
            return False
        return type(data[0][0]) in types.StringTypes
    
    except:
        return False

def _is_listish(data):
    """Is data vector/list or single column/row matrix?"""
    if type(data) in types.StringTypes:
        return False
    
    if _is_string_list(data):
        return True

    if _is_string_matrix(data):
        return 1 in (len(data), (data[0]))

    if Numeric.rank(data) == 2:
        return 1 in Numeric.shape(data)
    
    return Numeric.rank(data) == 1

def _is_matrix(data):
    """Returns True for all matrices, even with single row/column."""
    if _is_string_list(data):
        return False
    
    return _is_string_matrix(data) or Numeric.rank(data) == 2


class _DataSelectorModel:
    """Hold set of selectable ports and current selected port."""

    # enumeration of selection modes in matrices
    ROW, COLUMN = range(2)

    def __init__(self):
        self._ports = []
        self._current_port = None
        self._current_port_index = 0
        self._mode = _DataSelectorModel.COLUMN
        self._index = 1

    def set_index(self, index):
        self._index = index

    def get_index(self):
        return self._index

    def set_mode(self, mode):
        self._mode = mode

    def get_mode(self):
        return self._mode

    def set_active_ports(self, ports):
        self._ports = []

        if not ports: return

        selectable_ports = []

        for port in ports:
            container = port.getData()
            if not container:
                selectable_ports.append(port)
            else:
                data = container.getData()
                if _is_listish(data) or _is_matrix(data):
                    selectable_ports.append(port)

        # sort ports by name
        names = [(port.getName(), port) for port in selectable_ports]
        names.sort()
        self._ports = zip(*names)[1]

    def select_port_by_index(self, index):
        self._current_port_index = index + 1
        if index < 0:
            self._current_port = None
        else:
            self._current_port = self._ports[index]

    def get_current_id(self):
        if not self._current_port:
            return None
        else:
            return self._current_port.getId()

    def get_current_name(self):
        if not self._current_port:
            return None
        else:
            return self._current_port.getName()

    def select_port_by_id(self, port_id):
        """Selects port and returns index of selected port."""
        for index, port in enumerate(self._ports):
            if port.getId() == port_id:
                self._current_port = port
                return index
            
        self._current_port = None
        return None

    def get_port_names_and_descriptions(self):
        """Return (name, description) tuples for each port as a list."""
        pairs = []
        for port in self._ports:
            container = port.getData()
            if container:
                pairs.append((port.getName(), container.get_data_size_str()))
            else:
                pairs.append((port.getName(), "no data"))

        return pairs

    def get_current_data(self):
        if not self._current_port:
            return None

        container = self._current_port.getData()

        if not container:
            return None

        return container.getData()

    def get_current_port_index(self):
        return self._current_port_index

    def get_mode_string(self):
        if self._mode == 0:
            return 'row'
        else:
            return 'column'
    

class DataSelector(QObject):
    """Updated selector for plot sources.

    Allows user to always see the available input ports of a Node,
    while allowing plots to get the data directly, without having to
    fiddle with updating comboboxes and row ranges etc.

    Usage:
        # create with optional initiall settings for port, mode and
        # index

        port_id = port.getId()
        column_mode = True
        index = 5
        
        selector = DataSelector(port_id, column_mode, index)

        # connect to your existing widgets
        selector.set_source_combo(source_combo)
        selector.set_mode_combo(mode_combo)
        selector.set_index_spinbox(index_spinbox)

        # keep it updated as ports change
        selector.set_active_ports(port_list)

        # possibly connect change-events to functions saving state
        QObject.connect(selector, PYSIGNAL("dataChanged()"), my_callback)

        ...

        # to get the user selected data:
        value_list = selected.get_list()


    DataSelector is a combined MVC-style Controller and Facade.
    """

    def __init__(self, port_id=None, column_mode=True, index=1):
        QObject.__init__(self)
        self._source_combo = None
        self._index_spinbox = None
        self._mode_combo = None
        self._initial_port = port_id

        # remember last index user set explicitly and try to keep it
        self._last_set_index = 1
        self._model = _DataSelectorModel()
        # allow us the inhibit signals temporarily
        self._inhibit_change_signal = False

        self._set_column_mode(column_mode)
        self._model.set_index(index)

    def _set_column_mode(self, value):
        """Set True to define column mode, otherwise row mode."""
        if value:
            index = 0
            self._model.set_mode(_DataSelectorModel.COLUMN)
        else:
            self._model.set_mode(_DataSelectorModel.ROW)
            index = 1

        if self._mode_combo:
            self._mode_combo.setCurrentItem(index)

        self._update_index_choice(self._model.get_current_data())

    def is_column_mode(self):
        return self._model.get_mode() == _DataSelectorModel.COLUMN 

    def get_source_id(self):
        """Get ID of currently selected datasource/port."""
        return self._model.get_current_id()

    def get_source_name(self):
        """Get name of currently selected datasource/port."""
        return self._model.get_current_name()

    def set_source_combo(self, source_combo):
        """Set combo box to respond to and modify for sources."""
        self._source_combo = source_combo
        connect(self._source_combo, SIGNAL('activated(int)'), self._source_changed)
        self._update_widget_with_ports()

    def set_mode_combo(self, mode_combo):
        """Set combo box to respond to and fill with mode choices."""
        if mode_combo.count() == 0:
            mode_combo.insertItem("column")
            mode_combo.insertItem("row")
        if self._model.get_mode() == _DataSelectorModel.COLUMN:
            mode_combo.setCurrentItem(0)
        else:
            mode_combo.setCurrentItem(1)
            
        connect(mode_combo, SIGNAL('activated(const QString &)'), self._mode_changed)
        self._mode_combo = mode_combo

        if self._model.get_current_id() == None:
            mode_combo.setDisabled(True)
        else:
            self._update_mode_choice(self._model.get_current_data())

        
    def set_index_spinbox(self, index_spinbox):
        index_spinbox.setValue(self._model.get_index())
        self._last_set_index = self._model.get_index()
        connect(index_spinbox, SIGNAL('valueChanged(int)'), self._index_changed)
        self._index_spinbox = index_spinbox
        if self._model.get_current_id() == None:
            index_spinbox.setDisabled(True)
        else:
            self._update_index_choice(self._model.get_current_data())

    def _select_port_by_id(self, port_id):
        """Change current port to port having port_id."""
        index = self._model.select_port_by_id(port_id)

        if not self._source_combo:
            return
        
        if index != None:
            self._source_combo.setCurrentItem(index+1)
        else:
            self._source_combo.setCurrentItem(0)

        self._source_changed(self._source_combo.currentItem())

    def set_active_ports(self, ports):
        """Allow user to choose source only within currently active ports."""
        self._model.set_active_ports(ports)
        self._update_widget_with_ports()
        if self._initial_port:
            self._select_port_by_id(self._initial_port)
            self._initial_port = None
            
    def _update_widget_with_ports(self):
        """Update available choices to user based on known ports.

        Known ports are set with set_active_ports().
        """
        if not self._source_combo:
            return
        
        self._source_combo.clear()

        selected = self._model.get_current_id()
        entries = self._model.get_port_names_and_descriptions()
            
        if entries:
            self._source_combo.insertItem("No source selected")

            for portname, description in entries:
                self._source_combo.insertItem("%s (%s)" % (portname, description))

            if selected != None:
                self._select_port_by_id(selected)
        else:
            self._source_combo.insertItem("No sources available")

    def _mode_changed(self, mode_string):
        """Row/column mode changed."""
        if mode_string == self._model.get_mode_string():
            return 
        self._set_column_mode(str(mode_string) == "column")
        self.emit(PYSIGNAL("dataChanged()"), ())
        
    def _source_changed(self, source_index):
        """Data source changed."""

        if source_index == self._model.get_current_port_index():
            # don't want to send signal if port is reselected
            return 
        
        self._inhibit_change_signal = True
        try:
            self._model.select_port_by_index(source_index-1)

            if source_index == 0:
                if self._mode_combo:
                    self._mode_combo.setEnabled(False)
                if self._index_spinbox:
                    self._index_spinbox.setEnabled(False)
                return

            data = self._model.get_current_data()
            self._update_index_choice(data)
            self._update_mode_choice(data)

            self.emit(PYSIGNAL("dataChanged()"), ())
        finally:
            self._inhibit_change_signal = False

    def _index_changed(self, value):
        self._last_set_index = value
        self._model.set_index(value)

        # make sure it doesnt emit if this is just a result of
        # an updated source
        if not self._inhibit_change_signal:
            self.emit(PYSIGNAL("dataChanged()"), ())

    def _update_mode_choice(self, data):
        if not self._mode_combo:
            return

        self._mode_combo.setEnabled(True)

        if _is_listish(data):
            self._mode_combo.setDisabled(True)

    def _update_index_choice(self, data):
        if not self._index_spinbox:
            return
        
        last_index = self._last_set_index
        try:
            self._index_spinbox.setMinValue(1)
            self._index_spinbox.setMaxValue(99999)
            
            self._index_spinbox.setValue(last_index)
            self._index_spinbox.setEnabled(True)

            if not data:
                return

            if _is_listish(data):
                self._index_spinbox.setEnabled(False)
                return

            if _is_string_matrix(data):
                rows, columns = len(data), len(data[0])
            else:
                rows, columns = Numeric.shape(data)
                
            if self._model.get_mode() == _DataSelectorModel.ROW:
                self._index_spinbox.setMaxValue(rows)
            else:
                self._index_spinbox.setMaxValue(columns)
        finally:
            # make sure this method doesn't change last_set_index
            # indirectly
            self._last_set_index = last_index

    def get_list(self):
        """Get list contained in selected data source.

        This will get the list of values addressed by port, mode and
        index. If source data is a list, or is a single row/column in
        a matrix, then current mode and index will be ignored.

        When no data is available, None is returned.
        """ 
        data = self._model.get_current_data()

        if data == None:
            return None

        index = self._index_spinbox.value()

        if _is_string_matrix(data):
            rows, columns = len(data), len(data[0])
            if rows == 1:
                return data[0]
            if columns == 1:
                return zip(*data)[0]
            
            if self._model.get_mode() == _DataSelectorModel.ROW:
                return data[index-1]
            else:
                return zip(*data)[index-1]

        if Numeric.rank(data) == 1 or _is_string_list(data):
            return data

        # assume rank is 2
        rows, columns = Numeric.shape(data)
        if rows == 1:
            return data[0]

        if columns == 1:
            return Numeric.transpose(data)[0].tolist()

        if self._model.get_mode() == _DataSelectorModel.ROW:
            return Numeric.take(data, [index-1], 0)[0].tolist()

        return Numeric.transpose(Numeric.take(data, [index-1], 1))[0].tolist()

    def get_index(self):
        """Get currently selected index into source data."""
        return self._model.get_index()
