# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Pixmaps.Icons import Icons
from GUI.Plots.plot2d.dataselector import DataSelector
from Library.Util.Config import Config
from Library.Util.gui import ValueListViewItem
from Library.plot.plotmap import MappableSelection

from itertools import izip
from qt import QObject, SIGNAL, QListView, QListViewItem, QListViewItemIterator
from qt import PYSIGNAL, Qt, QPixmap
from qtui import QWidgetFactory
from sets import Set

import os
import rpy

connect = QObject.connect

class GoBrowser(MappableSelection):
    """Graphical tree-browser for Gene Ontology mapped to plots."""
    def __init__(self, parent, ontology, lookup_function):
        """Create browser for Gene Ontology using lookup_function for lookups.

        ontology is a GeneOntology providing named/structured
        classification of genes given GOIDs.  lookup_function is a
        callback supporting being called as lookup_function(type_id,
        id_list) and return list of [(id, genename, [GO-term ID,
        ...]), ...].
        """
        self._aspect = 'gene indices'
        MappableSelection.__init__(self, [self._aspect])
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          "gobrowser.ui"),
                                             None, parent)
        self._dbwidget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                            "gobrowser_dataselection.ui"),
                                               None, self._widget)

        # create listviews (_GoBrowserListView)
        tabwidget = self._widget.child('go_tree_tabwidget')
        self._molecular_list_view = _GoBrowserListView(tabwidget.page(0),
                                                  'molecular_tree')
        self._biological_list_view = _GoBrowserListView(tabwidget.page(1),
                                                   'biological_tree')
        self._cellular_list_view = _GoBrowserListView(tabwidget.page(2),
                                                 'cellular_tree')
        tabwidget = self._widget.child('go_tree_tabwidget')
        tabwidget.page(0).layout().addWidget(self._molecular_list_view, 0, 0)
        tabwidget.page(1).layout().addWidget(self._biological_list_view, 0, 0)
        tabwidget.page(2).layout().addWidget(self._cellular_list_view, 0, 0)

        self._target_type_combo = self._dbwidget.child('target_type_combo')
        self._master_type_combo = self._dbwidget.child('master_type_combo')
        self._available_genes = self._widget.child('available_genes')
        self._selected_genes = self._widget.child('selected_genes')
        self._tree_genes = self._widget.child('tree_genes')
        # last looked up gene data - list of tuples (gene_id,
        # gene_name, [goid1, goid2, ...])
        self._target_gene_data = {}
        self._master_gene_data = {}
        
        # holds _OntologyGraph instances for master input
        self._master_bio_subgraph = None
        self._master_mol_subgraph = None
        self._master_cel_subgraph = None
        self._master_input_changed = False
        self._master_type = ""

        self._ontology = ontology
        download_button = self._widget.child('download_button')
        connect(download_button, SIGNAL('clicked()'),
                self._show_download_selector)
        db_connect_button = self._dbwidget.child('download_button')
        connect(db_connect_button, SIGNAL("clicked()"),
                self._lookup_db)
        self._selection_tree_button = self._widget.child('selection_tree_button')
        connect(self._selection_tree_button, SIGNAL('clicked()'),
                self._slot_selection_tree)
        self._full_tree_button = self._widget.child('full_tree_button')
        connect(self._full_tree_button, SIGNAL("clicked()"),
                self._slot_full_tree)

        connect(self._biological_list_view, SIGNAL('selectionChanged()'),
                self._slot_selection_changed)
        connect(self._molecular_list_view, SIGNAL('selectionChanged()'),
                self._slot_selection_changed)
        connect(self._cellular_list_view, SIGNAL('selectionChanged()'),
                self._slot_selection_changed)
        connect(self._widget.child('cutoff_action'),
                SIGNAL('activated()'),
                self._slot_cutoff_dialog)

        self._lookup_function = lookup_function

        self._target_data_selector = DataSelector()
        self._target_data_selector.set_source_combo(self._dbwidget.child('target_source_combo'))
        self._target_data_selector.set_mode_combo(self._dbwidget.child('target_mode_combo'))
        self._target_data_selector.set_index_spinbox(self._dbwidget.child('target_index_spinbox'))

        self._master_data_selector = DataSelector()
        self._master_data_selector.set_source_combo(self._dbwidget.child('master_source_combo'))
        self._master_data_selector.set_mode_combo(self._dbwidget.child('master_mode_combo'))
        self._master_data_selector.set_index_spinbox(self._dbwidget.child('master_index_spinbox'))

        connect(self._master_data_selector, PYSIGNAL('dataChanged()'),
                self._slot_master_input_changed)
        connect(self._master_type_combo, SIGNAL('activated(const QString &)'),
                self._slot_master_type_changed)

        self._pvalue_cutoff = False
        self._depth_cutoff = False
        self._cutoff_value = 0.0

    def _slot_master_input_changed(self):
        self._master_input_changed = True

    def _slot_master_type_changed(self, master_type):
        if master_type != self._master_type:
            self._master_input_changed = True
            self._master_type = str(master_type)

    def _slot_cutoff_dialog(self):
        self._cutoff_dialog = CutOffDialog(self._widget,
                                           (self._depth_cutoff or self._pvalue_cutoff),
                                           self._pvalue_cutoff,
                                           self._cutoff_value)
        result_ = self._cutoff_dialog.show()
        if result_:
            self._pvalue_cutoff = self._cutoff_dialog.is_pvalue()
            self._depth_cutoff = self._cutoff_dialog.is_depth()
            self._cutoff_value = self._cutoff_dialog.get_cutoff_value()

        self._update_cutoff()

    def _slot_selection_tree(self):
        indices = self.get_highlighted_indices_by_aspect(self._aspect)
        self._generate_target_list_views(indices)

    def _slot_full_tree(self):
        self._generate_target_list_views(None)

    def _gene_items(self):
        gene_items = self._biological_list_view._gene_items
        gene_items.extend(self._molecular_list_view._gene_items)
        gene_items.extend(self._cellular_list_view._gene_items)
        return gene_items

    def _slot_selection_changed(self):
        selected_items = self._biological_list_view.get_gene_selection()
        selected_items.extend(self._molecular_list_view.get_gene_selection())
        selected_items.extend(self._cellular_list_view.get_gene_selection())

        # extract indices from items
        selected_ids = [item.get_value().gene_id() for item in selected_items]
        selected_indices = []
        for key, data in self._target_gene_data.iteritems():
            if data[0] in selected_ids:
                selected_indices.append(key)

        self.deselect_with_mapped()
        self.select_values(selected_indices, self._aspect)

    def _update_cutoff(self):
        listviews = [self._biological_list_view, self._molecular_list_view,
                     self._cellular_list_view]
        if self._pvalue_cutoff:
            for tree in listviews:
                tree.pvalue_cutoff(self._cutoff_value)
        elif self._depth_cutoff:
            for tree in listviews:
                tree.depth_cutoff(self._cutoff_value)
        else:
            for tree in listviews:
                tree.open_all_items()

    def name(self):
        """Name used for display when adding/showing plot mappings."""
        return "Gene Ontology browser"

    def source(self):
        """Used for display when adding plot mappings."""
        return ""

    def update_ports(self, port_list):
        """Inform browser about new data available in ports.

        Required to update database lookup.
        """
        self._target_data_selector.set_active_ports(port_list)
        self._master_data_selector.set_active_ports(port_list)

    def _lookup_db(self):
        """Perform database lookup based on selected data source and ID type."""
        self._target_gene_data = self._do_database_lookup(self._target_data_selector,
                                                          self._target_type_combo)
        if self._master_input_changed:
            self._master_gene_data = self._do_database_lookup(self._master_data_selector,
                                                              self._master_type_combo)
            self._master_input_changed = False
            
        self._update_available_genes()
        self._generate_master_graphs()

    def _do_database_lookup(self, data_selector, type_combo):
        gene_data = {}
        if not data_selector.get_list():
            return gene_data
        
        # make sure numbers can be looked up by converting to strings
        lookup_values = map(str, data_selector.get_list())
        data = self._lookup_function(str(type_combo.currentText()),
                                     lookup_values)
        for gene_id, gene_name, ids, llid in data:
            index = lookup_values.index(gene_id)
            gene_data[index] = gene_id, gene_name, ids, llid

        return gene_data

    def _update_available_genes(self):
        self._available_genes.setText('<p align="right"><b>%d</b></p>' \
                                      % len(self._target_gene_data))
        self._full_tree_button.setEnabled(bool(self._target_gene_data))

    def _generate_pvalue_inputs(self, target_bio_graph, target_mol_graph,
                                target_cel_graph):
        """Returns a list containing input for pvalue function.

        The result is a list of 3 tuples in a specific order. Each tuple
        contains:
         * list of goids
         * list of numbers of gene_counts found on and below each term for
           master
         * list of numbers of gene_counts found on and below each term for
           target
        """

        if not self._master_bio_subgraph.get_terms() or \
           not self._master_mol_subgraph.get_terms() or \
           not self._master_cel_subgraph.get_terms():
            return []


        results = []        

        for target, master in [(target_bio_graph, self._master_bio_subgraph),
                                (target_mol_graph, self._master_mol_subgraph),
                                (target_cel_graph, self._master_cel_subgraph)]:
            target_go_terms = list(Set(target.get_terms()))
            goids = [term.get_id() for term in target_go_terms]
            master_genes = []
            target_genes = []
            for term in target_go_terms:
                target_genes.append(term.get_gene_count())
                master_genes.append(master.lookup(term.get_id()).get_gene_count())

            results.append((goids, master_genes, target_genes))

        return results

    def _generate_master_graphs(self):
        allgenes = self._master_gene_data.values()

        graphs = self._generate_graphs(allgenes)
        
        self._master_bio_subgraph = graphs[0]
        self._master_mol_subgraph = graphs[1]
        self._master_cel_subgraph = graphs[2]

    def _generate_target_list_views(self, indices=None):
        """Generate trees based on indices and last lookup.

        If indices is None, all ids are looked up.
        """
        goids = Set()

        if indices == None:
            selected_genes = self._target_gene_data.values()
        else:
            selected_genes = []
            for i in indices:
                if self._target_gene_data.has_key(i):
                    selected_genes.append(self._target_gene_data[i])
        

        graphs = self._generate_graphs(selected_genes)

        inputs = self._generate_pvalue_inputs(*graphs)

        if inputs:
            master_gene_counts = [self._master_bio_subgraph.get_root().get_gene_count(),
                                  self._master_mol_subgraph.get_root().get_gene_count(),
                                  self._master_cel_subgraph.get_root().get_gene_count()]

            # add p-values to graphs
            for pvalue_input, graph, gene_count in izip(inputs, graphs, master_gene_counts):
                if graph.get_root():
                    goids, master_count, target_count = pvalue_input
                    pvalues = pvalue_function(gene_count,
                                              graph.get_root().get_gene_count(),
                                              master_count, target_count)
                    graph.add_pvalues(pvalues, goids)
        
        self._biological_list_view.populate(graphs[0])
        self._molecular_list_view.populate(graphs[1])
        self._cellular_list_view.populate(graphs[2])

        # number of genes in all trees should be same as number of
        # selected entries
        self._tree_genes.setText('<p align="right"><b>%d</b></p>' \
                                 % len(selected_genes))

    def _generate_graphs(self, genes):
        goids = Set()
        for gene_id, gene_name, ids, llid in genes:
            goids.union_update(ids)
        bio_subgraph = self._ontology.get_biological_process_subgraph(goids)
        mol_subgraph = self._ontology.get_molecular_function_subgraph(goids)
        cel_subgraph = self._ontology.get_cellular_component_subgraph(goids)
        
        self._add_genes_to_graph(genes, bio_subgraph)
        self._add_genes_to_graph(genes, mol_subgraph)
        self._add_genes_to_graph(genes, cel_subgraph)

        return bio_subgraph, mol_subgraph, cel_subgraph        

    def _add_genes_to_graph(self, gene_data, graph):
        for gene_id, gene_name, ids, llid in gene_data:
            gene = Gene(gene_id, gene_name, llid)
            for goid in ids:
                term = graph.lookup(goid)
                if term:
                    term.add_gene(gene)

    def show(self):
        self._widget.show()
        self._widget.raiseW()

    def _show_download_selector(self):
        self._dbwidget.show()
        self._dbwidget.raiseW()

    def _update_selected_genes(self):
        indices = self.get_highlighted_indices_by_aspect(self._aspect)
        self._selected_genes.setText('<p align="right"><b>%d</b></p>' \
                                     % len(indices))
        self._selection_tree_button.setEnabled(bool(indices) and
                                               bool(self._target_gene_data))

    def highlight_indices_by_aspect(self, indices, aspect):
        MappableSelection.highlight_indices_by_aspect(self, indices, aspect)
        self._update_selected_genes()
        self._highlight_gene_indices(indices)

    def _clear_selection(self):
        for gene_item in self._gene_items():
            gene_item.set_selected(False)

        self._biological_list_view.selectAll(False)
        self._cellular_list_view.selectAll(False)
        self._molecular_list_view.selectAll(False)

    def _highlight_gene_indices(self, indices):
        if not self._target_gene_data:
            # no genes has yet been looked up through _lookup_db()
            return

        self._clear_selection()
        
        # highlight genes from selection in mapped MappableSelection
        selected_ids = [self._target_gene_data[index][0] for index in indices]
        for item in self._gene_items():
            if item.get_value().gene_id() in selected_ids:
                item.set_selected(True)

    def deselect_self(self):
        MappableSelection.deselect_self(self)
        self._clear_selection()
        self._update_selected_genes()


class _GoBrowserListView(QListView):

    def __init__(self, parent, name='0'):
        QListView.__init__(self, parent, name)
        self.setSelectionMode(QListView.Extended)
        self.setRootIsDecorated(True)
        self.setResizeMode(QListView.LastColumn)
        self.addColumn('')
        # make sure columnheaders are hidden
        self.header().hide()

        self._gene_items = [] # all listviewitems holding Genes
        self._items = [] # all listviewitems in listview

        self._gene_icon = QPixmap(Icons.gene_icon)
        self._term_icon = QPixmap(Icons.term_icon)

        self._graph = None # the graph that this listview is populated from
        
        # tells whether or not the 'Ctrl' button was pressed during
        # the last mouse press event
        self._control_pressed = False 

    def get_go_terms(self):
        """Return all distinct _OntologyTerms in the listview tree."""
        # because of alternative ids we need to filter out duplicated terms
        return list(Set(self._graph.get_terms()))

    def get_selection(self):
        """Get the selected listviewitems of this listview."""
        selection = [item for item in self._items if item.isSelected()]
        return selection

    def get_gene_selection(self):
        """Get the selected listviewitems which are holding genes.

        If an internal node (i.e., a term listviewitem) is selected
        all of its descendant listviewitems which hold genes, are
        returned.
        """
        selection = []
        for item in self.get_selection():
            if item.childCount() or self._control_pressed:
                # do not add leafnodes which were in the previous
                # selection, unless 'Ctrl' was pressed at the same
                # time as selection was set
                selection.extend(self._find_all_leaf_nodes(item))

        return selection + [item for item in self._gene_items
                            if QListViewItem.isSelected(item)]

    def _find_all_leaf_nodes(self, item):
        if not item.childCount():
            return [item]

        leafs = []
        child = item.firstChild()
        while child:
            leafs.extend(self._find_all_leaf_nodes(child))
            child = child.nextSibling()

        return leafs

    def contentsMousePressEvent(self, mouseEvent):
        if mouseEvent.state() & Qt.ControlButton:
            self._control_pressed = True
            
        QListView.contentsMousePressEvent(self, mouseEvent)

    def contentsMouseReleaseEvent(self, mouseEvent):
        self._control_pressed = False
        QListView.contentsMouseReleaseEvent(self, mouseEvent)

    def populate(self, graph):
        """Populate listview with all terms in ontology graph."""
        # make sure sorting is off, so genes are placed below terms
        self.clear()
        self.setSorting(-1)
        self._gene_items = []
        self._items = []

        root = graph.get_root()
        if root:
            item_text = "%s (%d)" % (root.name(), root.get_gene_count())
            if root.get_p_value():
                item_text += " p=%0.4f" % root.get_p_value()
            root_item = ValueListViewItem(self, root, item_text)
            self._populate_from_term(root_item, root)
            root_item.setOpen(True)
            self._items.append(root_item)

        self._graph = graph
        
    def _populate_from_term(self, parent_item, term):
        """Create ListViewItems recursively for children of term."""
        last_child_item = None
        for child in term.children():
            item_text = "%s (%d)" % (child.name(), child.get_gene_count())
            if child.get_p_value():
                item_text += " p=%0.4f" % child.get_p_value()
            if last_child_item:
                item = ValueListViewItem(parent_item, child, last_child_item,
                                         item_text)
            else:
                item = ValueListViewItem(parent_item, child, item_text)
            self._populate_from_term(item, child)
            last_child_item = item
            # make sure term is open on creation, showing all subitems
            item.setOpen(True)
            item.setPixmap(0, self._term_icon)
            self._items.append(item)
            
        for gene in term.get_genes():
            if last_child_item:
                item = BoldTextValueListViewItem(parent_item, gene,
                                                 last_child_item, gene.name())
            else:
                item = BoldTextValueListViewItem(parent_item, gene, gene.name())
            item.setPixmap(0, self._gene_icon)
            self._gene_items.append(item)
            self._items.append(item)

    def depth_cutoff(self, depth):
        """Closes all items with item.depth() greater than or equal to depth."""
        for item in self._items:
            if item.depth() >= depth:
                item.setOpen(False)
            else:
                item.setOpen(True)

    def pvalue_cutoff(self, pvalue, item=None):
        """Hides items with p-value greater than pvalue.

        If all descendants of item has p-value greater than pvalue,
        the item will be hidden.
        """
        if not item:
            item = self.firstChild()

        if not item.childCount():
            # item is a gene item with no p-value
            return False

        item_pvalue = item.get_value().get_p_value()

        child = item.firstChild()
        children_visible = False
        while child:
            if self.pvalue_cutoff(pvalue, child):
                children_visible = True
            child = child.nextSibling()

        if children_visible or item_pvalue <= pvalue:
            return True
        else:
            item.setVisible(False)
            return False

    def open_all_items(self):
        for item in self._items:
            item.setOpen(True)
            item.setVisible(True)
        
            
            
class BoldTextValueListViewItem(ValueListViewItem):
    def __init__(self, parent, value, *text):
        ValueListViewItem.__init__(self, parent, value, *text)
        self._is_selected = False

    def set_selected(self, bool_):
        self._is_selected = bool_
        QListViewItem.repaint(self)

    def isSelected(self):
        return self._is_selected

    def paintCell(self, painter, *args):
        if self._is_selected:
            painter.font().setBold(True) 
        else:
            painter.font().setBold(False)

        ValueListViewItem.paintCell(self, painter, *args)
            

class Gene:
    def __init__(self, gene_id, name, locuslink_id):
        """Name is the name of the gene returned from eGOn, gene_id is
        the id that is looked up in eGOn"""
        self._gene_id = gene_id
        self._name = name
        self._locuslink_id = locuslink_id
        
    def name(self):
        return self._name

    def gene_id(self):
        return self._gene_id

    def locuslink_id(self):
        return self._locuslink_id

class CutOffDialog:
    def __init__(self, parent, enable_radio = False, pvalue_on = False, cutoff_value = 0.0):
        """Dialog widget used to enable/disable cutoff and enter the cutoff value

        enable_radio - bool, used to enable the buttongroup used for the radiobuttons
        pvalue_on - bool, if the pvalue radiobutton is on or not
        
        This dialog box is modal and exec_loop() will return 1 if OK is clicked
        and 0 if Cancel is clicked. This is defined in the ui file.
        """
        
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'gobrowser_cutoff.ui'),
                                             None, parent)

        self._widget.child('cutoff_check').setChecked(enable_radio)
        if enable_radio:
            self._widget.child('cutoff_radiobuttons_group').setEnabled(enable_radio)
            self._widget.child('cutoff_depth').setChecked(not pvalue_on)
            self._widget.child('cutoff_pvalue').setChecked(pvalue_on) 
            self._widget.child('cutoff_valueedit').setText(str(cutoff_value))

        connect(self._widget.child('cutoff_check'), SIGNAL('toggled(bool)'),
                self._slot_disable_radio)

    def _slot_disable_radio(self, bool_):
        if not bool_:
            self._widget.child('cutoff_depth').setChecked(False)
            self._widget.child('cutoff_pvalue').setChecked(False) 
                
    def show(self):
        return self._widget.exec_loop()
        
    def is_pvalue(self):
        return self._widget.child('cutoff_pvalue').isOn()

    def is_depth(self):
        return self._widget.child('cutoff_depth').isOn()
    
    def get_cutoff_value(self):
        try:
            if self.is_pvalue():
                return float(str(self._widget.child('cutoff_valueedit').text()))
            elif self.is_depth():
                # The string needs to be converted to a float before it can be
                # converted to an integer. 
                return int(float(str(self._widget.child('cutoff_valueedit').text())))
        except ValueError:
            return 0.0


_masterclones = {'21484': {'llid': '8577',
                           'goids': ['GO:0005309', 'GO:0043285', 'GO:0007548',
                                     'GO:0030034', 'GO:0009741', 'GO:0047295',
                                     'GO:0047282', 'GO:0020009'],
                           'name': '\xc5sa'},
                 '21908': {'llid': '7276',
                           'goids': ['GO:0046526', 'GO:0009205', 'GO:0001904',
                                     'GO:0018937', 'GO:0043001', 'GO:0008775'],
                           'name': 'Ulrikke'},
                 '21481': {'llid': '8603',
                           'goids': ['GO:0042930', 'GO:0005616', 'GO:0042291',
                                     'GO:0030934', 'GO:0010238', 'GO:0004623'],
                           'name': '\xc5se'},
                 '22467': {'llid': '5595',
                           'goids': ['GO:0015196'],
                           'name': 'Susann'},
                 '21649': {'llid': '8020',
                           'goids': ['GO:0004138', 'GO:0047126', 'GO:0012509',
                                     'GO:0042180', 'GO:0018514', 'GO:0018540',
                                     'GO:0003909', 'GO:0042296', 'GO:0009045',
                                     'GO:0019663', 'GO:0019098', 'GO:0015494',
                                     'GO:0047324', 'GO:0019204'],
                           'name': 'Vilma'},
                 '770027': {'llid': '3220',
                            'goids': ['GO:0045198', 'GO:0003955', 'GO:0005548',
                                      'GO:0018398', 'GO:0046481', 'GO:0048047',
                                      'GO:0031008', 'GO:0004924', 'GO:0018427'],
                            'name': 'Rebecca'},
                 '788183': {'llid': '3064',
                            'goids': ['GO:0042412', 'GO:0005139', 'GO:0006465',
                                      'GO:0030619', 'GO:0015599', 'GO:0042297',
                                      'GO:0043295'],
                            'name': 'Rania'},
                 '22177': {'llid': '6402',
                           'goids': ['GO:0009620', 'GO:0008097', 'GO:0047366',
                                     'GO:0019822', 'GO:0018089'],
                           'name': 'Thora'},
                 '22667': {'llid': '4814',
                           'goids': ['GO:0015294', 'GO:0019296', 'GO:0007083'],
                           'name': 'Siw'},
                 '22851': {'llid': '4472',
                           'goids': ['GO:0000154'],
                           'name': 'Silje'},
                 '22501': {'llid': '5179',
                           'goids': ['GO:0018460', 'GO:0019562', 'GO:0042876',
                                     'GO:0046994', 'GO:0017065', 'GO:0004712',
                                     'GO:0005110'],
                           'name': 'Sophia'},
                 '21764': {'llid': '7606',
                           'goids': ['GO:0031153', 'GO:0000818'],
                           'name': 'Veronika'},
                 '20075': {'llid': '8730',
                           'goids': ['GO:0009250', 'GO:0006670', 'GO:0045981',
                                     'GO:0051403', 'GO:0042410', 'GO:0006906',
                                     'GO:0015444'],
                           'name': 'Zainab'},
                 '2695870': {'llid': '2775',
                             'goids': ['GO:0005743', 'GO:0004001', 'GO:0047958',
                                       'GO:0018869', 'GO:0008066', 'GO:0000309'],
                             'name': 'Ragnhild'},
                 '21763': {'llid': '7702',
                           'goids': ['GO:0043314', 'GO:0051100', 'GO:0030547',
                                     'GO:0050475', 'GO:0019679', 'GO:0001576',
                                     'GO:0046340', 'GO:0042122', 'GO:0010197',
                                     'GO:0009545', 'GO:0000071', 'GO:0047647'],
                           'name': 'Veslem\xf8y'},
                 '725680': {'llid': '2597',
                            'goids': ['GO:0003745', 'GO:0003705'],
                            'name': 'Rachel'},
                 '20078': {'llid': '8687',
                           'goids': ['GO:0046437', 'GO:0051137', 'GO:0015517',
                                     'GO:0050079', 'GO:0008870', 'GO:0016004',
                                     'GO:0018602', 'GO:0001741', 'GO:0004921',
                                     'GO:0045622'],
                           'name': 'Zahra'},
                 '23125': {'llid': '3751',
                           'goids': ['GO:0045812', 'GO:0045664', 'GO:0050908',
                                     'GO:0008720', 'GO:0035124', 'GO:0046528',
                                     'GO:0042781', 'GO:0035006', 'GO:0001408',
                                     'GO:0030232', 'GO:0000003', 'GO:0046459',
                                     'GO:0047890'],
                           'name': 'Sabrina'},
                 '23116': {'llid': '3851',
                           'goids': ['GO:0030192', 'GO:0035125'],
                           'name': 'Sahra'},
                 '22264': {'llid': '6068',
                           'goids': ['GO:0001555', 'GO:0006627', 'GO:0019522',
                                     'GO:0003898', 'GO:0045928', 'GO:0047450',
                                     'GO:0017084', 'GO:0008511'],
                           'name': 'Thale'},
                 '376290': {'llid': '2326',
                            'goids': ['GO:0000115', 'GO:0050983', 'GO:0010095',
                                      'GO:0016737', 'GO:0003859', 'GO:0010181',
                                      'GO:0047700', 'GO:0018988', 'GO:0030319',
                                      'GO:0035010', 'GO:0007090'],
                            'name': 'Paula'},
                 '23119': {'llid': '3762',
                           'goids': ['GO:0045468'],
                           'name': 'Saga'},
                 '22478': {'llid': '5460',
                           'goids': ['GO:0015666'],
                           'name': 'Sunniva'},
                 '21919': {'llid': '7178',
                           'goids': ['GO:0006830', 'GO:0018767', 'GO:0047030',
                                     'GO:0047053', 'GO:0051102', 'GO:0008594',
                                     'GO:0031139'],
                           'name': 'Trude'},
                 '22765': {'llid': '4597',
                           'goids': ['GO:0030776', 'GO:0000758', 'GO:0015781',
                                     'GO:0007331', 'GO:0030744', 'GO:0000077',
                                     'GO:0051359', 'GO:0018352', 'GO:0045827',
                                     'GO:0047119', 'GO:0009178', 'GO:0046686',
                                     'GO:0005462', 'GO:0004834'],
                           'name': 'Siren'},
                 '21678': {'llid': '7852',
                           'goids': ['GO:0042149', 'GO:0005956', 'GO:0015318',
                                     'GO:0048394', 'GO:0050562', 'GO:0008040',
                                     'GO:0007448', 'GO:0009224', 'GO:0045004',
                                     'GO:0035033', 'GO:0003984'],
                           'name': 'Vilde'},
                 '22470': {'llid': '5506',
                           'goids': ['GO:0005621', 'GO:0043187', 'GO:0009815',
                                     'GO:0045853', 'GO:0015763', 'GO:0006846',
                                     'GO:0048474'],
                           'name': 'Susan'},
                 '21911': {'llid': '7259',
                           'goids': ['GO:0042308', 'GO:0042942', 'GO:0001608',
                                     'GO:0042624', 'GO:0004908', 'GO:0000300',
                                     'GO:0004655', 'GO:0030906', 'GO:0019783',
                                     'GO:0035281', 'GO:0051004', 'GO:0017138',
                                     'GO:0051275', 'GO:0000118'],
                           'name': 'Tyra'},
                 '21917': {'llid': '7239',
                           'goids': ['GO:0004048'],
                           'name': 'Tuva'},
                 '21670': {'llid': '7949',
                           'goids': ['GO:0018913', 'GO:0006353', 'GO:0009157',
                                     'GO:0035038', 'GO:0019077'],
                           'name': 'Vilja'},
                 '4152877': {'llid': '1919',
                             'goids': ['GO:0016559', 'GO:0046833', 'GO:0048227',
                                       'GO:0016892', 'GO:0030707', 'GO:0003737',
                                       'GO:0045406', 'GO:0016419', 'GO:0010171',
                                       'GO:0005008', 'GO:0009116', 'GO:0003815',
                                       'GO:0006932'],
                             'name': 'Ninni'},
                 '22927': {'llid': '4198',
                           'goids': ['GO:0042738', 'GO:0008715', 'GO:0018804',
                                     'GO:0018870', 'GO:0044249', 'GO:0000813',
                                     'GO:0050810', 'GO:0047213', 'GO:0008614',
                                     'GO:0030960', 'GO:0048351', 'GO:0008998'],
                           'name': 'Serina'},
                 '809892': {'llid': '1759',
                            'goids': ['GO:0015966', 'GO:0009889', 'GO:0006471',
                                      'GO:0031025', 'GO:0047895', 'GO:0042144',
                                      'GO:0018901', 'GO:0046899'],
                            'name': 'Nikoline'},
                 '22270': {'llid': '6029',
                           'goids': ['GO:0030171', 'GO:0048338', 'GO:0004143'],
                           'name': 'Terese'},
                 '22275': {'llid': '6019',
                           'goids': ['GO:0010023', 'GO:0006639', 'GO:0007213',
                                     'GO:0009373', 'GO:0031353', 'GO:0006671',
                                     'GO:0045420', 'GO:0004305', 'GO:0006454',
                                     'GO:0003984', 'GO:0045743', 'GO:0003841',
                                     'GO:0008971'],
                           'name': 'Tea'},
                 '811044': {'llid': '1074',
                            'goids': ['GO:0004736', 'GO:0004937', 'GO:0009994',
                                      'GO:0004900', 'GO:0018236', 'GO:0016891',
                                      'GO:0015862', 'GO:0046554'],
                            'name': 'Nadia'},
                 '22278': {'llid': '5978',
                           'goids': ['GO:0006076', 'GO:0009721', 'GO:0018062',
                                     'GO:0042906', 'GO:0042263', 'GO:0046402',
                                     'GO:0018563', 'GO:0046696', 'GO:0016127'],
                           'name': 'Taran'},
                 '128159': {'llid': '2138',
                            'goids': ['GO:0030792', 'GO:0016140', 'GO:0015565',
                                      'GO:0006496', 'GO:0004891', 'GO:0008988',
                                      'GO:0031386', 'GO:0007430', 'GO:0042313'],
                            'name': 'Oda'},
                 '22799': {'llid': '4485',
                           'goids': ['GO:0019956', 'GO:0004820', 'GO:0015497',
                                     'GO:0035157', 'GO:0043183', 'GO:0003832'],
                           'name': 'Simone'},
                 '21849': {'llid': '7315',
                           'goids': ['GO:0016527', 'GO:0006608', 'GO:0018219',
                                     'GO:0045818', 'GO:0035291', 'GO:0030475',
                                     'GO:0008478', 'GO:0008903', 'GO:0030966',
                                     'GO:0050208', 'GO:0000710', 'GO:0035035'],
                           'name': 'Una'},
                 '22794': {'llid': '4552',
                           'goids': ['GO:0001858', 'GO:0051227', 'GO:0009155',
                                     'GO:0031186', 'GO:0050307', 'GO:0044275'],
                           'name': 'Sina'},
                 '22568': {'llid': '4935',
                           'goids': ['GO:0030392', 'GO:0018047', 'GO:0015950',
                                     'GO:0045127', 'GO:0042879', 'GO:0035091',
                                     'GO:0045241', 'GO:0000313', 'GO:0018506',
                                     'GO:0030489', 'GO:0006381', 'GO:0042968',
                                     'GO:0005966'],
                           'name': 'Solveig'},
                 '22790': {'llid': '4576',
                           'goids': ['GO:0047514'],
                           'name': 'Sine'},
                 '21842': {'llid': '7325',
                           'goids': ['GO:0000742', 'GO:0048066', 'GO:0031096',
                                     'GO:0035252', 'GO:0030284', 'GO:0045485',
                                     'GO:0015293', 'GO:0018955', 'GO:0008681',
                                     'GO:0045425', 'GO:0042812'],
                           'name': 'Une'},
                 '22153': {'llid': '6529',
                           'goids': ['GO:0045439', 'GO:0048462', 'GO:0048387',
                                     'GO:0045186', 'GO:0000273', 'GO:0005622',
                                     'GO:0006243', 'GO:0043228', 'GO:0046004',
                                     'GO:0030172', 'GO:0051021', 'GO:0008126',
                                     'GO:0005848'],
                           'name': 'Tilde'},
                 '22018': {'llid': '6797',
                           'goids': ['GO:0019974'],
                           'name': 'Tone'},
                 '22156': {'llid': '6485',
                           'goids': ['GO:0009888', 'GO:0009087', 'GO:0047921',
                                     'GO:0048095', 'GO:0018624', 'GO:0015367',
                                     'GO:0035004', 'GO:0048079', 'GO:0017190',
                                     'GO:0048358', 'GO:0030081', 'GO:0017167'],
                           'name': 'Tia'},
                 '21584': {'llid': '8247',
                           'goids': ['GO:0030788', 'GO:0044265', 'GO:0009103',
                                     'GO:0009050', 'GO:0016563', 'GO:0045905',
                                     'GO:0046809', 'GO:0018676', 'GO:0016926',
                                     'GO:0035193', 'GO:0017066'],
                           'name': 'Wenche'},
                 '22010': {'llid': '6873',
                           'goids': ['GO:0030711', 'GO:0046134', 'GO:0030514',
                                     'GO:0008706', 'GO:0004981', 'GO:0004011',
                                     'GO:0045240', 'GO:0045092'],
                           'name': 'Tonje'},
                 '22400': {'llid': '5604',
                           'goids': ['GO:0030601', 'GO:0046163', 'GO:0010110',
                                     'GO:0031190', 'GO:0000350', 'GO:0003755',
                                     'GO:0042672', 'GO:0019087', 'GO:0000103',
                                     'GO:0008864'],
                           'name': 'Susanne'},
                 '22910': {'llid': '4332',
                           'goids': ['GO:0004766', 'GO:0003886', 'GO:0019643',
                                     'GO:0006051', 'GO:0043210'],
                           'name': 'Signe'},
                 '22484': {'llid': '5308',
                           'goids': ['GO:0004446', 'GO:0047386', 'GO:0051168',
                                     'GO:0007517', 'GO:0046126', 'GO:0004880',
                                     'GO:0047380', 'GO:0012502', 'GO:0018849',
                                     'GO:0001646', 'GO:0018143', 'GO:0006935',
                                     'GO:0019323', 'GO:0047052', 'GO:0046581'],
                           'name': 'Stephanie'},
                 '22487': {'llid': '5276',
                           'goids': ['GO:0019033', 'GO:0016124', 'GO:0007596',
                                     'GO:0045497', 'GO:0043234', 'GO:0046376',
                                     'GO:0008870', 'GO:0004064', 'GO:0042153',
                                     'GO:0009394'],
                           'name': 'Stella'},
                 '21740': {'llid': '7762',
                           'goids': ['GO:0046511'],
                           'name': 'Vibeke'},
                 '21500': {'llid': '8490',
                           'goids': ['GO:0019795', 'GO:0019468'],
                           'name': 'Yvonne'},
                 '21503': {'llid': '8393',
                           'goids': ['GO:0008372'],
                           'name': 'Ylva'},
                 '22482': {'llid': '5430',
                           'goids': ['GO:0009185', 'GO:0016605', 'GO:0045872',
                                     'GO:0018161', 'GO:0018710', 'GO:0048390',
                                     'GO:0004190', 'GO:0008637'],
                           'name': 'Stine'},
                 '26314': {'llid': '1148',
                           'goids': ['GO:0008188', 'GO:0019333', 'GO:0015279',
                                     'GO:0042857', 'GO:0006923', 'GO:0016598',
                                     'GO:0008858', 'GO:0019119'],
                           'name': 'Nadja'},
                 '21998': {'llid': '6946',
                           'goids': ['GO:0004423', 'GO:0008179', 'GO:0048130',
                                     'GO:0005585', 'GO:0019349', 'GO:0015399'],
                           'name': 'Tora'},
                 '47559': {'llid': '2072',
                           'goids': ['GO:0050656', 'GO:0019541', 'GO:0042126',
                                     'GO:0030479', 'GO:0050140', 'GO:0048442',
                                     'GO:0015651'],
                           'name': 'Norah'},
                 '21465': {'llid': '8617',
                           'goids': ['GO:0008497', 'GO:0019809', 'GO:0005024',
                                     'GO:0019154', 'GO:0004572', 'GO:0006176',
                                     'GO:0018112', 'GO:0015892', 'GO:0020029',
                                     'GO:0047330', 'GO:0046805', 'GO:0035240',
                                     'GO:0050238', 'GO:0019159', 'GO:0042760'],
                           'name': '\xc5shild'},
                 '812975': {'llid': '2491',
                            'goids': ['GO:0001897', 'GO:0047940', 'GO:0008110',
                                      'GO:0001727', 'GO:0048329', 'GO:0018696',
                                      'GO:0018361', 'GO:0051261', 'GO:0018983',
                                      'GO:0031066'],
                            'name': 'Petra'},
                 '23173': {'llid': '3518',
                           'goids': ['GO:0019987', 'GO:0030262', 'GO:0001634',
                                     'GO:0009327', 'GO:0045356', 'GO:0001783',
                                     'GO:0031237', 'GO:0051324', 'GO:0005350',
                                     'GO:0019869', 'GO:0042355', 'GO:0047472',
                                     'GO:0018769', 'GO:0016179'],
                           'name': 'Rita'},
                 '41898': {'llid': '2448',
                           'goids': ['GO:0000368', 'GO:0047450', 'GO:0009022',
                                     'GO:0044415', 'GO:0043078', 'GO:0048529'],
                           'name': 'Pernille'},
                 '755145': {'llid': '2987',
                            'goids': ['GO:0042389', 'GO:0004993', 'GO:0018410',
                                      'GO:0042547', 'GO:0046207'],
                            'name': 'Randi'},
                 '1509201': {'llid': '2855',
                             'goids': ['GO:0019175', 'GO:0007298', 'GO:0016015',
                                       'GO:0045326', 'GO:0048524', 'GO:0015770',
                                       'GO:0042585', 'GO:0007473', 'GO:0019546',
                                       'GO:0001937'],
                             'name': 'Rakel'},
                 '1570420': {'llid': '1979',
                             'goids': ['GO:0005599', 'GO:0006883', 'GO:0008624',
                                       'GO:0035074', 'GO:0030205', 'GO:0008418',
                                       'GO:0001516'],
                             'name': 'Nora'},
                 '22658': {'llid': '4821',
                           'goids': ['GO:0019083', 'GO:0030791', 'GO:0016995',
                                     'GO:0045511', 'GO:0047762'],
                           'name': 'Sofia'},
                 '21837': {'llid': '7422',
                           'goids': ['GO:0010232', 'GO:0048266', 'GO:0015595'],
                           'name': 'Vanessa'},
                 '22389': {'llid': '5640',
                           'goids': ['GO:0045584', 'GO:0030852'],
                           'name': 'Sylvia'},
                 '22148': {'llid': '6598',
                           'goids': ['GO:0015799', 'GO:0004238'],
                           'name': 'Tina'},
                 '21450': {'llid': '8629',
                           'goids': ['GO:0016500', 'GO:0048350', 'GO:0050156',
                                     'GO:0045893', 'GO:0046995', 'GO:0016480',
                                     'GO:0046359', 'GO:0000143', 'GO:0007383',
                                     'GO:0009295', 'GO:0050945', 'GO:0000735',
                                     'GO:0008978'],
                           'name': '\xc5sne'},
                 '22382': {'llid': '5696',
                           'goids': ['GO:0030810', 'GO:0017005', 'GO:0048326',
                                     'GO:0050293'],
                           'name': 'Synn\xf8ve'},
                 '23095': {'llid': '3931',
                           'goids': ['GO:0019642', 'GO:0009611', 'GO:0015522',
                                     'GO:0003924', 'GO:0009853', 'GO:0050871',
                                     'GO:0035122', 'GO:0018457', 'GO:0000298',
                                     'GO:0015913'],
                           'name': 'Salma'},
                 '21593': {'llid': '8161',
                           'goids': ['GO:0045611', 'GO:0019744', 'GO:0051051',
                                     'GO:0005369', 'GO:0000280', 'GO:0019628',
                                     'GO:0045655', 'GO:0005363', 'GO:0018180'],
                           'name': 'V\xe5rin'},
                 '22384': {'llid': '5643',
                           'goids': ['GO:0008105', 'GO:0031025', 'GO:0003834',
                                     'GO:0004349', 'GO:0000708', 'GO:0001543',
                                     'GO:0047352', 'GO:0007272', 'GO:0042771',
                                     'GO:0006357'],
                           'name': 'Synne'},
                 '241530': {'llid': '1605',
                            'goids': ['GO:0009113', 'GO:0016149'],
                            'name': 'Nicoline'},
                 '21514': {'llid': '8335',
                           'goids': ['GO:0006579', 'GO:0047461', 'GO:0042241',
                                     'GO:0004568', 'GO:0000395', 'GO:0018369',
                                     'GO:0008819', 'GO:0031219', 'GO:0007354',
                                     'GO:0043004', 'GO:0042773', 'GO:0035023'],
                           'name': 'Yasmin'},
                 '22984': {'llid': '4129',
                           'goids': ['GO:0018706', 'GO:0018114', 'GO:0015395',
                                     'GO:0015091', 'GO:0030944', 'GO:0006779',
                                     'GO:0047152', 'GO:0046574', 'GO:0048593',
                                     'GO:0042762', 'GO:0031050', 'GO:0004084',
                                     'GO:0016360'],
                           'name': 'Seline'},
                 '22985': {'llid': '4101',
                           'goids': ['GO:0050088', 'GO:0001836', 'GO:0019340',
                                     'GO:0047964', 'GO:0050910', 'GO:0006288',
                                     'GO:0008681', 'GO:0043080', 'GO:0005577',
                                     'GO:0051001', 'GO:0042458', 'GO:0016236',
                                     'GO:0006045', 'GO:0047692'],
                           'name': 'Selina'},
                 '22490': {'llid': '5248',
                           'goids': ['GO:0017075', 'GO:0042896', 'GO:0008505',
                                     'GO:0010047', 'GO:0009583', 'GO:0044269',
                                     'GO:0043339', 'GO:0001662', 'GO:0018643',
                                     'GO:0045729', 'GO:0030335', 'GO:0004765',
                                     'GO:0004777'],
                           'name': 'Sophie'},
                 '22863': {'llid': '4420',
                           'goids': ['GO:0018205', 'GO:0004162', 'GO:0019727',
                                     'GO:0043362', 'GO:0043144', 'GO:0051376',
                                     'GO:0047163', 'GO:0046879', 'GO:0045472',
                                     'GO:0035014', 'GO:0046179', 'GO:0051285',
                                     'GO:0015229', 'GO:0051100'],
                           'name': 'Silja'},
                 '823940': {'llid': '1565',
                            'goids': ['GO:0048504', 'GO:0019197', 'GO:0000766',
                                      'GO:0008290', 'GO:0005007', 'GO:0046851',
                                      'GO:0000302', 'GO:0007453', 'GO:0030199',
                                      'GO:0051038', 'GO:0010119', 'GO:0015696',
                                      'GO:0019414'],
                            'name': 'Nathalie'},
                 '22866': {'llid': '4364',
                           'goids': ['GO:0008688', 'GO:0051125', 'GO:0042717',
                                     'GO:0004329', 'GO:0006714', 'GO:0006687',
                                     'GO:0019870'],
                           'name': 'Sigrun'},
                 '50614': {'llid': '1321',
                           'goids': ['GO:0000335', 'GO:0004511', 'GO:0006386',
                                     'GO:0046525', 'GO:0005367', 'GO:0016061'],
                           'name': 'Natalia'},
                 '1461729': {'llid': '1294',
                             'goids': ['GO:0047255', 'GO:0048395', 'GO:0001411',
                                       'GO:0000032'],
                             'name': 'Naomi'},
                 '342721': {'llid': '2526',
                            'goids': ['GO:0046147', 'GO:0009115', 'GO:0009289',
                                      'GO:0030449'],
                            'name': 'Pia'},
                 '298963': {'llid': '1026',
                            'goids': ['GO:0019695', 'GO:0015775', 'GO:0030930',
                                      'GO:0008433', 'GO:0010108', 'GO:0051436',
                                      'GO:0030159', 'GO:0015957', 'GO:0046654',
                                      'GO:0016245', 'GO:0000333', 'GO:0045991',
                                      'GO:0050096', 'GO:0016089', 'GO:0047846'],
                            'name': 'Monika'},
                 '22253': {'llid': '6222',
                           'goids': ['GO:0047965', 'GO:0000921', 'GO:0016523',
                                     'GO:0050207', 'GO:0008986', 'GO:0030893',
                                     'GO:0045058', 'GO:0004017', 'GO:0050794',
                                     'GO:0018928', 'GO:0000389', 'GO:0030779'],
                           'name': 'Thelma'},
                 '22256': {'llid': '6162',
                           'goids': ['GO:0030012'],
                           'name': 'Thea'},
                 '22030': {'llid': '6740',
                           'goids': ['GO:0045763', 'GO:0046750', 'GO:0046682'],
                           'name': 'Tomine'},
                 '22629': {'llid': '4848',
                           'goids': ['GO:0043041', 'GO:0035222', 'GO:0050809',
                                     'GO:0019862'],
                           'name': 'Sofie'},
                 '22233': {'llid': '6314',
                           'goids': ['GO:0001720', 'GO:0001305', 'GO:0009268',
                                     'GO:0009544', 'GO:0046037', 'GO:0015936'],
                           'name': 'Theresa'},
                 '21825': {'llid': '7430',
                           'goids': ['GO:0016945', 'GO:0019413', 'GO:0047822',
                                     'GO:0043185', 'GO:0007331', 'GO:0048506'],
                           'name': 'Vanja'},
                 '22582': {'llid': '4929',
                           'goids': ['GO:0051358', 'GO:0009977', 'GO:0016015',
                                     'GO:0009736', 'GO:0006406', 'GO:0019126',
                                     'GO:0031328', 'GO:0016919', 'GO:0018528',
                                     'GO:0030357'],
                           'name': 'Sol'},
                 '21609': {'llid': '8079',
                           'goids': ['GO:0047769', 'GO:0009807', 'GO:0009915',
                                     'GO:0019339', 'GO:0050262', 'GO:0017126',
                                     'GO:0008536', 'GO:0015202', 'GO:0007091',
                                     'GO:0007028'],
                           'name': 'Vivian'},
                 '21606': {'llid': '8157',
                           'goids': ['GO:0000907', 'GO:0030026', 'GO:0015171',
                                     'GO:0030269', 'GO:0030609', 'GO:0015905'],
                           'name': 'V\xe5r'},
                 '22739': {'llid': '4633',
                           'goids': ['GO:0030316', 'GO:0004925', 'GO:0048452',
                                     'GO:0043196', 'GO:0043148', 'GO:0006218',
                                     'GO:0050844', 'GO:0016682', 'GO:0046811',
                                     'GO:0035165', 'GO:0043191', 'GO:0031110',
                                     'GO:0031485'],
                           'name': 'Siril'},
                 '23240': {'llid': '3410',
                           'goids': ['GO:0047257', 'GO:0016819', 'GO:0000271'],
                           'name': 'Renate'},
                 '23087': {'llid': '3947',
                           'goids': ['GO:0046030', 'GO:0050673', 'GO:0015485',
                                     'GO:0045707'],
                           'name': 'Samira'},
                 '756272': {'llid': '2626',
                            'goids': ['GO:0016521', 'GO:0044271', 'GO:0015827',
                                      'GO:0048519', 'GO:0018859', 'GO:0009852',
                                      'GO:0001706', 'GO:0016360', 'GO:0047632',
                                      'GO:0000932'],
                            'name': 'Ragna'},
                 '21721': {'llid': '7832',
                           'goids': ['GO:0019496', 'GO:0009239', 'GO:0045986',
                                     'GO:0008613'],
                           'name': 'Viktoria'},
                 '22974': {'llid': '4192',
                           'goids': ['GO:0047213', 'GO:0008544', 'GO:0019578',
                                     'GO:0019579', 'GO:0046049', 'GO:0019293',
                                     'GO:0047173', 'GO:0048561', 'GO:0030036',
                                     'GO:0001799', 'GO:0050839'],
                           'name': 'Selma'},
                 '300103': {'llid': '2348',
                            'goids': ['GO:0018863', 'GO:0019686', 'GO:0046482',
                                      'GO:0045713', 'GO:0008283', 'GO:0035122',
                                      'GO:0006612', 'GO:0016176', 'GO:0007608',
                                      'GO:0006780'],
                            'name': 'Pauline'},
                 '283954': {'llid': '2247',
                            'goids': ['GO:0006408', 'GO:0046816', 'GO:0051370',
                                      'GO:0045300'],
                            'name': 'Olivia'},
                 '23155': {'llid': '3585',
                           'goids': ['GO:0009175', 'GO:0012511'],
                           'name': 'Ronja'},
                 '21956': {'llid': '7034',
                           'goids': ['GO:0030942', 'GO:0031268', 'GO:0016267'],
                           'name': 'Torunn'},
                 '21951': {'llid': '7062',
                           'goids': ['GO:0042601', 'GO:0016580', 'GO:0050338',
                                     'GO:0018443', 'GO:0050166', 'GO:0050975',
                                     'GO:0019555', 'GO:0019359', 'GO:0018471',
                                     'GO:0019091', 'GO:0015653', 'GO:0051065',
                                     'GO:0047781', 'GO:0047648'],
                           'name': 'Tove'},
                 '23136': {'llid': '3706',
                           'goids': ['GO:0000031', 'GO:0015689', 'GO:0050753',
                                     'GO:0015112', 'GO:0009996', 'GO:0035078',
                                     'GO:0047588', 'GO:0003755', 'GO:0009377',
                                     'GO:0008440', 'GO:0008381', 'GO:0018871'],
                           'name': 'Ruth'},
                 '22040': {'llid': '6696',
                           'goids': ['GO:0004638', 'GO:0030629', 'GO:0045245',
                                     'GO:0042089', 'GO:0015671', 'GO:0018806',
                                     'GO:0019167', 'GO:0051372', 'GO:0030100',
                                     'GO:0048428'],
                           'name': 'Tirill'},
                 '23232': {'llid': '3429',
                           'goids': ['GO:0016889', 'GO:0007541', 'GO:0050898',
                                     'GO:0007078', 'GO:0015193', 'GO:0018036',
                                     'GO:0016714'],
                           'name': 'Rikke'},
                 '22049': {'llid': '6678',
                           'goids': ['GO:0042924', 'GO:0006163'],
                           'name': 'Tiril'},
                 '233199': {'llid': '1668',
                            'goids': ['GO:0019607'],
                            'name': 'Nikita'},
                 '143386': {'llid': '1596',
                            'goids': ['GO:0016677', 'GO:0006627', 'GO:0042981',
                                      'GO:0007379', 'GO:0001614', 'GO:0010033',
                                      'GO:0007067'],
                            'name': 'Nicole'},
                 '21538': {'llid': '8276',
                           'goids': ['GO:0031102', 'GO:0047864', 'GO:0001749',
                                     'GO:0006430', 'GO:0046685', 'GO:0035303',
                                     'GO:0009542', 'GO:0004382', 'GO:0035176',
                                     'GO:0008691', 'GO:0019336', 'GO:0050203'],
                           'name': 'Wilma'},
                 '21730': {'llid': '7804',
                           'goids': ['GO:0001502', 'GO:0004353', 'GO:0000176',
                                     'GO:0016200', 'GO:0016871', 'GO:0006528',
                                     'GO:0050307', 'GO:0018771'],
                           'name': 'Victoria'},
                 '547247': {'llid': '2226',
                            'goids': ['GO:0006737', 'GO:0019582'],
                            'name': 'Oline'},
                 '23004': {'llid': '4059',
                           'goids': ['GO:0047500', 'GO:0000505', 'GO:0047500',
                                     'GO:0009261'],
                           'name': 'Selin'},
                 '23005': {'llid': '4050',
                           'goids': ['GO:0045252', 'GO:0047267', 'GO:0015402'],
                           'name': 'Sarah'},
                 '22337': {'llid': '5891',
                           'goids': ['GO:0045939'],
                           'name': 'Tanja'},
                 '22332': {'llid': '5945',
                           'goids': ['GO:0010190', 'GO:0048528', 'GO:0046478'],
                           'name': 'Tara'},
                 '32609': {'llid': '3356',
                           'goids': ['GO:0005840', 'GO:0042253'],
                           'name': 'Reidun'},
                 '3506618': {'llid': '3336',
                             'goids': ['GO:0007140', 'GO:0018631', 'GO:0007092'],
                             'name': 'Regine'},
                 '128143': {'llid': '3304',
                            'goids': ['GO:0001657', 'GO:0018039', 'GO:0051349',
                                      'GO:0018904', 'GO:0047414', 'GO:0004708',
                                      'GO:0050239', 'GO:0000743', 'GO:0050247',
                                      'GO:0008360', 'GO:0043121'],
                            'name': 'Rebekka'},
                 '212429': {'llid': '3223',
                            'goids': ['GO:0030085', 'GO:0008282', 'GO:0016291',
                                      'GO:0047473', 'GO:0030985', 'GO:0004493',
                                      'GO:0006185', 'GO:0044270', 'GO:0000930',
                                      'GO:0015379'],
                            'name': 'Rebecka'},
                 '22350': {'llid': '5795',
                           'goids': ['GO:0043036', 'GO:0047042', 'GO:0008518'],
                           'name': 'Tale'},
                 '587010': {'llid': '3164',
                            'goids': ['GO:0051023'],
                            'name': 'Rannveig'},
                 '23141': {'llid': '3670',
                           'goids': ['GO:0009163', 'GO:0017040'],
                           'name': 'Runa'},
                 '22051': {'llid': '6657',
                           'goids': ['GO:0019950', 'GO:0035031', 'GO:0006405',
                                     'GO:0030030', 'GO:0015803', 'GO:0046706',
                                     'GO:0047048', 'GO:0019601', 'GO:0045793',
                                     'GO:0009254', 'GO:0050650', 'GO:0003874',
                                     'GO:0005825', 'GO:0008910', 'GO:0042661'],
                           'name': 'Tine'},
                 '22751': {'llid': '4603',
                           'goids': ['GO:0018935', 'GO:0019987', 'GO:0009430',
                                     'GO:0031085', 'GO:0005771', 'GO:0045616',
                                     'GO:0004629', 'GO:0008706', 'GO:0030189',
                                     'GO:0016711', 'GO:0018298', 'GO:0005780',
                                     'GO:0047079', 'GO:0016132'],
                           'name': 'Siri'},
                 '2309073': {'llid': '2908',
                             'goids': ['GO:0047178', 'GO:0043378', 'GO:0047161',
                                       'GO:0009897', 'GO:0047750', 'GO:0050804',
                                       'GO:0018785', 'GO:0005856', 'GO:0015689',
                                       'GO:0050688', 'GO:0008404', 'GO:0005801',
                                       'GO:0042308', 'GO:0019339'],
                             'name': 'Ramona'},
                 '210887': {'llid': '2676',
                            'goids': ['GO:0048600', 'GO:0042856', 'GO:0006440',
                                      'GO:0030274', 'GO:0015482', 'GO:0006787',
                                      'GO:0006846'],
                            'name': 'Ragne'},
                 '22912': {'llid': '4246',
                           'goids': ['GO:0017050', 'GO:0006923', 'GO:0009931',
                                     'GO:0046482', 'GO:0045916', 'GO:0046326'],
                           'name': 'Serine'},
                 '21815': {'llid': '7511',
                           'goids': ['GO:0048468', 'GO:0030854', 'GO:0018906',
                                     'GO:0042820', 'GO:0018061'],
                           'name': 'Vendela'},
                 '471200': {'llid': '1227',
                            'goids': ['GO:0005771', 'GO:0018019', 'GO:0007077',
                                      'GO:0018329', 'GO:0043034', 'GO:0007066',
                                      'GO:0004838', 'GO:0005315', 'GO:0019599',
                                      'GO:0048103', 'GO:0015422', 'GO:0007302'],
                            'name': 'Nanna'},
                 '22897': {'llid': '4353',
                           'goids': ['GO:0006799', 'GO:0045908'],
                           'name': 'Sigrid'},
                 '23031': {'llid': '4010',
                           'goids': ['GO:0007205', 'GO:0043155', 'GO:0031292'],
                           'name': 'Sanna'},
                 '21798': {'llid': '7583',
                           'goids': ['GO:0001785', 'GO:0000021', 'GO:0044276',
                                     'GO:0007083', 'GO:0045367', 'GO:0030469',
                                     'GO:0030650', 'GO:0017184', 'GO:0047603',
                                     'GO:0006843', 'GO:0050642', 'GO:0047673',
                                     'GO:0050870', 'GO:0007585', 'GO:0016797'],
                           'name': 'Veronica'},
                 '22453': {'llid': '5598',
                           'goids': ['GO:0050050', 'GO:0019229', 'GO:0009562',
                                     'GO:0016681', 'GO:0005890', 'GO:0018227',
                                     'GO:0050856', 'GO:0006946', 'GO:0030395',
                                     'GO:0006734', 'GO:0010037', 'GO:0004187'],
                           'name': 'Susanna'},
                 '23059': {'llid': '3998',
                           'goids': ['GO:0019050', 'GO:0018698', 'GO:0045142',
                                     'GO:0046456', 'GO:0030004', 'GO:0046982',
                                     'GO:0000072', 'GO:0019557', 'GO:0006461'],
                           'name': 'Sandra'},
                 '22690': {'llid': '4787',
                           'goids': ['GO:0048436', 'GO:0010111', 'GO:0004855',
                                     'GO:0047698', 'GO:0019176', 'GO:0010211'],
                           'name': 'Siv'},
                 '22519': {'llid': '5015',
                           'goids': ['GO:0004677', 'GO:0048337', 'GO:0047818',
                                     'GO:0046382', 'GO:0047934'],
                           'name': 'Solv\xe5r'},
                 '2229386': {'llid': '1825',
                             'goids': ['GO:0015166'],
                             'name': 'Nina'},
                 '22697': {'llid': '4702',
                           'goids': ['GO:0042698', 'GO:0004415', 'GO:0007422',
                                     'GO:0005965', 'GO:0015100', 'GO:0018161',
                                     'GO:0006057', 'GO:0016266', 'GO:0042791',
                                     'GO:0019885', 'GO:0009508', 'GO:0042669',
                                     'GO:0010034', 'GO:0008475'],
                           'name': 'Sissel'},
                 '22515': {'llid': '5099',
                           'goids': ['GO:0046472', 'GO:0050115'],
                           'name': 'Sonja'},
                 '22483': {'llid': '5396',
                           'goids': ['GO:0015280', 'GO:0015512', 'GO:0046708',
                                     'GO:0016261', 'GO:0046161', 'GO:0031473',
                                     'GO:0019932', 'GO:0001750', 'GO:0048241',
                                     'GO:0031116', 'GO:0004136', 'GO:0048223',
                                     'GO:0019945'],
                           'name': 'Stina'},
                 '292996': {'llid': '1498',
                            'goids': ['GO:0047276', 'GO:0043095', 'GO:0043164',
                                      'GO:0005315', 'GO:0005046', 'GO:0016422',
                                      'GO:0042982', 'GO:0051138', 'GO:0046345',
                                      'GO:0050148', 'GO:0048446', 'GO:0015202',
                                      'GO:0016493', 'GO:0004730'],
                            'name': 'Natasha'},
                 '212704': {'llid': '1400',
                            'goids': ['GO:0006742'],
                            'name': 'Natalie'},
                 '2466108': {'llid': '2275',
                             'goids': ['GO:0006641', 'GO:0048211'],
                             'name': 'Othilie'},
                 '21929': {'llid': '7152',
                           'goids': ['GO:0006040', 'GO:0019763'],
                           'name': 'Trine'},
                 '21806': {'llid': '7521',
                           'goids': ['GO:0031340', 'GO:0008715', 'GO:0006107',
                                     'GO:0009832', 'GO:0007098', 'GO:0001588',
                                     'GO:0045549'],
                           'name': 'Vera'},
                 '23020': {'llid': '4034',
                           'goids': ['GO:0004901', 'GO:0031495', 'GO:0005770',
                                     'GO:0004088', 'GO:0006700', 'GO:0040021'],
                           'name': 'Sara'},
                 '23021': {'llid': '4032',
                           'goids': ['GO:0004775', 'GO:0045290', 'GO:0010074',
                                     'GO:0047256', 'GO:0042882', 'GO:0042997',
                                     'GO:0007197', 'GO:0001549', 'GO:0050625',
                                     'GO:0006039', 'GO:0042300', 'GO:0018037',
                                     'GO:0009283'],
                           'name': 'Sanne'},
                 '22195': {'llid': '6387',
                           'goids': ['GO:0050142', 'GO:0042014', 'GO:0015944',
                                     'GO:0015868', 'GO:0035200'],
                           'name': 'Therese'}}


class FakeGeneDb:
    def lookup(self, id_type, ids):
        """Lookup list of gene IDs in eGOn.
        
        Returns list of [(id, genename, [GO-term ID, ...]), ...] for all
        IDs. IDs not known to the database will not be present in the
        result.
        
        Current implementation is a dummy-lookup, using fake data for
        testing purposes.
        """
    
        result = []
        for gene_id in ids:
            try:
                gene = _masterclones[gene_id]
                result.append((gene_id, gene["name"], gene["goids"], gene['llid']))
            except KeyError:
                print "KeyError:", repr(gene_id)
        return result
    

def pvalue_function(master_number, target_number, master_genes, target_genes):
    eGOn_pvalue_R_function = """
    egonP <- function(Narray,Nsig,GenesInMaster,GenesInTarget){                     
    
    GenesInMaster <- as.vector(GenesInMaster)                                        
    GenesInTarget <- as.vector(GenesInTarget)                                       
    
    a <- vector(mode='numeric',length=length(GenesInTarget))                        
    for (gen in 1:length(GenesInTarget)){                                            
    
    M <- matrix(c(Narray,Nsig,GenesInMaster[gen],GenesInTarget[gen]),nrow =        
    2,ncol=2)                                                                       
    ta <- fisher.test(M)                                                           
    a[gen] <- ta$p.value                                                           
    
    }                                                                              
    
    return(list(pval=a))                                                            
    
    } 
    """
    
    rpy.r(eGOn_pvalue_R_function)
    result = rpy.r.egonP(master_number, target_number, master_genes, target_genes)
    
    return result['pval']


    
