#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Nodes.guiplot2dnodesettings import PlotSettings
from GUI.Plots.plot2d.scatterplot import ScatterPlot, MarkerSettingsDialog
from Library.Util.Config import Config
from Library.Util import lists, rangeparser
from Library.Util.rangeparser import InvalidRangeError
from qt import Qt, QColor, SIGNAL
from qtui import QWidgetFactory
from dprint import dprint, flags

import qt
import os

connect = qt.QObject.connect


class LinePlot(ScatterPlot):
    """Simple line plot class."""

    _type = "line_plot"
    _name = "Line Plot"

    line_styles = [Qt.SolidLine, Qt.DashLine, Qt.DotLine, Qt.DashDotLine,
                   Qt.DashDotDotLine]

    def __init__(self, parent_plot, node, config):
        ScatterPlot.__init__(self, parent_plot, node, config)
        self._marker_size = (2, 2)
        self.set_line_style(Qt.SolidLine)
        self.set_line_colour(0, 0, 255)
        self.set_line_width(2)

        defaults = {}

        # set defaults
        defaults['line_type_index'] = self.line_styles.index(Qt.SolidLine)
        defaults['line_colour'] = (0, 0, 255)
        defaults['line_width'] = 2

        # overwrite defaults with loaded values and store
        defaults.update(self.get_marker_settings())
        self.set_marker_settings(defaults)

    def create_settings(self, index, tab_widget, plot_widget):
        return LinePlotSettings(self._node, index, tab_widget, plot_widget,
                                self)
            
    def set_data(self, x=None, y=None, indices=None):
        """Set subset of x,y as data, as described by indices (0-base)."""
        if not y: return

        # If no x values are given, assume y as function of x
        if not x: 
            x = range(1, len(y)+1)
            # make two wrongs do a right, reverse x-values to select
            # correct indices if indice-parameters are high to low.
            if indices and indices[0] > indices[-1]:
                x.reverse()

        ScatterPlot.set_data(self, x, y, indices)
        self.get_curve().setStyle(Qt.SolidLine)

    def set_line_style(self, style):
        pen = self.get_curve().pen()
        pen.setStyle(style)

    def set_line_colour(self, r, g, b):
        pen = self.get_curve().pen()
        colour = QColor(r, g, b)
        pen.setColor(colour)

    def set_line_width(self, width):
        pen = self.get_curve().pen()
        pen.setWidth(width)

    def _update_markers(self):
        """Update all markers and curves after changing marker settings."""
        ScatterPlot._update_markers(self)
        
        self.set_line_style(self.line_styles[self.get_marker_setting('line_type_index')])
        self.set_line_colour(*self.get_marker_setting('line_colour'))
        self.set_line_width(self.get_marker_setting('line_width'))


class LinePlotSettings(PlotSettings):

    INDEXED_KEY = 'Indexed'
    
    def __init__(self, node, index, tab_widget, qwt_plot_widget, plot):
        """Constructor.

        Parameters:
        node            -- the node to retrieve data from
        index           -- the index in the listbox of the tab_widget
        tab_widget      -- the tab_widget the associated widget is drawn in
        qwt_plot_widget -- the plot_widget in which the curve we control is drawn
        """
        PlotSettings.__init__(self, node, index, tab_widget, qwt_plot_widget,
                              plot)

        self._marker_settings_dialog = MarkerSettingsDialog(tab_widget,
                                                            node,
                                                            qwt_plot_widget,
                                                            plot)
        
        self._marker_settings_dialog.init_dialog() # Show line settings.
        self._load_settings()
        self._init_widget()
        self._set_widget_values()
        self._connect_signals()
        self._load_tooltip_settings()

    def _init_widget(self):
        """Initiates the associated widget."""
        self._settings_widget = QWidgetFactory.create(
            os.path.join(Config.uidir, 'form_line_plot_setup.ui'),
            self._tab_widget, self._tab_widget)

        self._line_edit_plot_name = self._settings_widget.child('line_edit_plot_name')
        self._combo_box_x_axis_data = self._settings_widget.child('combo_box_x_axis_data')
        self._combo_box_y_axis_data = self._settings_widget.child('combo_box_y_axis_data')
        self._combo_box_x_col_row = self._settings_widget.child('combo_box_x_col_row')
        self._combo_box_y_col_row = self._settings_widget.child('combo_box_y_col_row')
        self._spin_box_x_index = self._settings_widget.child('spin_box_x_index')
        self._spin_box_y_index = self._settings_widget.child('spin_box_y_index')
        self._line_edit_range = self._settings_widget.child('line_edit_range')
        self._push_button_line_setup = self._settings_widget.child('push_button_line_setup')
        self._autoupdate_checkbox = self._settings_widget.child('autoupdate_checkbox')
        self._update_button = self._settings_widget.child('update_button')
        self._tooltips_button = self._settings_widget.child('push_button_tooltips')

        self.update_data_names_combo_boxes()
        
        self._tab_widget.layout().addWidget(self._settings_widget, 0, 1)

    def _slot_line_setup_clicked(self):
        self._marker_settings_dialog.display_dialog()

    def _connect_signals(self):
        connect(self._push_button_line_setup, SIGNAL('clicked()'),
                self._slot_line_setup_clicked)
        connect(self._combo_box_x_axis_data,
                SIGNAL('activated(const QString &)'),
                self._slot_combo_box_x_axis_data_changed)
        connect(self._combo_box_y_axis_data,
                SIGNAL('activated(const QString &)'),
                self._slot_combo_box_y_axis_data_changed)
        connect(self._combo_box_x_col_row, SIGNAL('activated(int)'),
                self._slot_combo_box_x_col_row_changed)
        connect(self._combo_box_y_col_row, SIGNAL('activated(int)'),
                self._slot_combo_box_y_col_row_changed)
        connect(self._spin_box_x_index, SIGNAL('valueChanged(int)'),
                self._slot_spin_box_x_index_changed)
        connect(self._spin_box_y_index, SIGNAL('valueChanged(int)'),
                self._slot_spin_box_y_index_changed)
        connect(self._line_edit_plot_name,
                SIGNAL('textChanged(const QString &)'),
                self._slot_plot_name_changed)
        connect(self._line_edit_range, SIGNAL('textChanged(const QString &)'),
                self._slot_line_edit_range_changed)
        connect(self._line_edit_range, SIGNAL('returnPressed()'),
                self._slot_range_return_pressed)                                      
        connect(self._autoupdate_checkbox, SIGNAL('toggled(bool)'),
                self._slot_checkbox_changed)
        connect(self._update_button, SIGNAL('clicked()'), self.update_data)
        connect(self._tooltips_button, SIGNAL('clicked()'),
                self._slot_set_tooltips) # inherited slot

        
    def _load_settings(self):
        # set defaults

        default_input_port = ''
        input_ports = self._node.get_all_input_ports()
        if input_ports:
            default_input_port = input_ports[0].getName()
        self.set_parameter('x_axis_data', LinePlotSettings.INDEXED_KEY)
        self.set_parameter('y_axis_data', default_input_port)
        self.set_parameter('x_col_row', '0')
        self.set_parameter('y_col_row', '0')
        self.set_parameter('x_index', '0')
        self.set_parameter('y_index', '0')
        self.set_parameter('range', '1:end')
        self.set_parameter('name', 'Line Plot')
        self.set_parameter('autoupdate', False)

        for key, value in self._plot.get_settings().items():
            self.set_parameter(key, value)

    def _set_widget_values(self):
        if self._combo_box_x_axis_data.count():
            self.set_current_item_x_data(self.get_parameter('x_axis_data'))
        if self._combo_box_y_axis_data.count():
            self.set_current_item_y_data(self.get_parameter('y_axis_data'))
        x_axis_data = self._info_pattern.sub('', str(self._combo_box_x_axis_data.currentText()))
        
        if x_axis_data == 'Indexed':
            self._combo_box_x_col_row.setEnabled(False)
            self._spin_box_x_index.setEnabled(False)
        else:
            self._combo_box_x_col_row.setEnabled(True)
            self._spin_box_x_index.setEnabled(True)

        self._combo_box_x_col_row.setCurrentItem(int(self.get_parameter('x_col_row')))
        self._combo_box_y_col_row.setCurrentItem(int(self.get_parameter('y_col_row')))
        self._spin_box_x_index.setValue(int(self.get_parameter('x_index'))+1)
        self._spin_box_y_index.setValue(int(self.get_parameter('y_index'))+1)
        self._line_edit_range.setText(self.get_parameter('range'))
        self._line_edit_plot_name.setText(self.get_parameter('name'))
        self._autoupdate_checkbox.setChecked(self.get_parameter('autoupdate'))        
        if self.get_parameter('autoupdate'):
            self.update_data()
        

    def _slot_combo_box_x_axis_data_changed(self, q_string):
        """Sets combo_box_x_col_row and spin_box_x_index disabled if string in
        combo_box_x_axis_data is 'Index in Y-Data'. Enables them otherwise.
        """
        q_string = self._info_pattern.sub('', str(q_string))
        if q_string == 'Indexed':
            self._combo_box_x_col_row.setEnabled(False)
            self._spin_box_x_index.setEnabled(False)
        else:
            self._combo_box_x_col_row.setEnabled(True)
            self._spin_box_x_index.setEnabled(True)
        self.set_parameter('x_axis_data', str(q_string))
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()        


    def update_data_names_combo_boxes(self):
        """updates the combo boxes that contains the names of the inputs.
        Overridden from super class.
        """
        self._set_inports_in_combo_box(self._combo_box_x_axis_data)
        self._set_inports_in_combo_box(self._combo_box_y_axis_data)
        self._combo_box_x_axis_data.insertItem(LinePlotSettings.INDEXED_KEY)
        try:
            self.set_current_item_x_data(self.get_parameter('x_axis_data'))
        except NoSuchParameterError:
            # not set yet
            pass # sets top item as current
        try:
            if self._combo_box_y_axis_data.count():
                self.set_current_item_y_data(self.get_parameter('y_axis_data'))
        except NoSuchParameterError:
            pass

    def update_data(self):
        """Updates the plot and sets the data in the plot_widget"""
        PlotSettings.update_data(self)

        # get bad cases done with
        if not self._combo_box_x_axis_data.count() or \
               not self._combo_box_y_axis_data.count():
            # there is no data ports
            dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] no data ports to update")
            self._qwt_plot_widget.set_data_by_plot(self._plot, [], [])
            self._qwt_plot_widget.replot()
            return
            
        x_data_name = self.get_parameter('x_axis_data')
        y_data_name = self.get_parameter('y_axis_data')

        x_port_name = self._get_port_name(x_data_name)
        y_port_name = self._get_port_name(y_data_name)

        x_is_column = not int(self.get_parameter('x_col_row'))
        y_is_column = not int(self.get_parameter('y_col_row'))
        x_data_idx  = int(self.get_parameter('x_index'))
        y_data_idx  = int(self.get_parameter('y_index'))
        data_range = self.get_parameter('range')

        # decide if we have to update the plot
        dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] updating plot")
        try: # to get the data
            try:
                if x_data_name == 'Indexed':
                    xdata = []
                else:
                    dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN]  " +\
                           "getting xdata (%s, %s, %s)" \
                           % (x_data_name, x_is_column, x_data_idx))
                    xdata = self._node.get_data(x_data_name,
                                                x_is_column,
                                                x_data_idx)
                    dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] " +\
                           "got xdata, len %d" % len(xdata))

            except IndexError:
                if x_is_column:
                    col_row = 'column'
                else:
                    col_row = 'row'
                raise IndexError, "Failed to retrieve X-data from '%s' in %s %d."\
                      % (x_data_name, col_row, x_data_idx+1)

            try:
                dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] getting ydata " +\
                       "(%s, %s, %s)" % (y_data_name, y_is_column, y_data_idx))
                ydata = self._node.get_data(y_data_name,
                                            y_is_column,
                                            y_data_idx)
                dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] got " +\
                       "ydata, len %d" % len(ydata))

            except IndexError:
                if y_is_column:
                    col_row = 'column'
                else:
                    col_row = 'row'
                raise IndexError, "Failed to retrieve Y-data from '%s' in %s %d."\
                      % (y_data_name, col_row, y_data_idx+1)

        except AttributeError, e:
            # diagram not run yet
            dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] AttributeError:", str(e))
            return
        except IndexError, e:
            qt.QMessageBox.information(self._tab_widget,
                                       'Plot - Index out of bounds',
                                       'One of the indexes on the data is out'+\
                                       ' of bounds.\n%s' % e)
            self._set_max_value_spin_boxes()
                # recursively call _update_plot since the slot on the spin
                # boxes (above line) sets the value on the X_data_idx var
                # to the max value if it is above max value and hence we
                # have valid data on the recursive call...
            return self.update_data()
        except Exception, e:
            if str(e) == '[Plot2DNode]: no port called ':
                dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] Empty portname")
            else:
                dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] erroneous" +\
                       "port(s) resetting...")
                # erroneous port, reset plot:
                self._qwt_plot_widget.set_data_by_plot(self._plot, [], [])
                self._qwt_plot_widget.replot()
            return
        else: # try --- else
            # we have data!
            if xdata and len(xdata) != len(ydata):
                qt.QMessageBox.information(
                    self._tab_widget,
                    'Plot - inconsistent data',
                    'The number of elements in the X-data is not the same\n' +\
                    'as the number of elements in the Y-data. Stripping\n' +\
                    'whichever is the longest.')
                xdata = xdata[:len(ydata)]
                ydata = ydata[:len(xdata)]

            # make 0-indexed list of indices into dataset using ydata
            # as template as xdata might be empty when 'Indexed'
            try:
                indices = rangeparser.matlab_parse_range(self.get_parameter('range'),
                                                         range(len(ydata)))
            except InvalidRangeError, e:
                qt.QMessageBox.warning(self._tab_widget,
                                       'Plot - range notation error:',
                                       str(e) + "\nShowing all values.")
                indices = None

            try:
                dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] setting data")
                # Another evil try/except in case we get an
                # "un-hash-able" data set
                try:
                    x_hash = hash(lists.to_tuple(xdata))
                    y_hash = hash(lists.to_tuple(ydata))
                    indices_hash = hash(lists.to_tuple(indices))
                    if x_hash != self._x_hash or y_hash != self._y_hash or \
                           indices_hash != self._indices_hash:
                        self._x_hash = x_hash
                        self._y_hash = y_hash
                        self._indices_hash = indices_hash
                except Exception, e:
                    dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] Exception", e,
                           type(xdata), type(ydata), type(indices))
                
                self._qwt_plot_widget.set_data_by_plot(self._plot, xdata, ydata,
                                               indices)
            except TypeError, e:
                self._qwt_plot_widget.replot()
                qt.QMessageBox.information(
                    self._tab_widget,
                    'Plot - Wrong type of data',
                    'The data on the input port(s) are invalid.')
                dprint(flags.GUI.PLOTS.RUN, "[GUI.PLOTS.RUN] TypeError,", e)

            self._set_max_value_spin_boxes()

            # need to reload the tooltip settings
            self._load_tooltip_settings()

            # updating content of combo_boxes, in order to be updated
            # if the dialog window is open when play is pressed.
            self.update_data_names_combo_boxes()

        # update spinners max/min-values
        self._marker_settings_dialog._set_values()
        self._marker_settings_dialog.apply_settings()
