# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.



from qt import *
from math import ceil, floor
import types

from GUI.HelpViewer import *
from GUI.Canvas.CanvasLine import *
from GUI.Canvas.CanvasPolygon import *
from GUI.contextmenus import LinkContext
from System.Data.ports import MultipleInputPort
from Library.Util.Config import *
from dprint import *

class LinkCreationError(Exception):
    pass

class Direction:
    """Class representing discrete directions."""
    def __init__(self, x, y):
        """Instantiating this class yourself probably doesn't make much sense.

        Use through the constants:
        Direction.NORTH
        Direction.SOUTH
        Direction.EAST
        Direction.WEST
        """
        self._x = x
        self._y = y


    def isOppositeOf(self, otherDirection):
        if self._x == otherDirection._x == 0 and self._y * otherDirection._y < 0:
            return True
        if self._y == otherDirection._y == 0 and self._x * otherDirection._x < 0:
            return True
        return False


    def isSameDirection(self, otherDirection):
        if self._x == otherDirection._x == 0 and self._y * otherDirection._y > 0:
            return True
        if self._y == otherDirection._y == 0 and self._x * otherDirection._x > 0:
            return True
        return False


    def isPerpendicularTo(self, otherDirection):
        if not self.isOppositeOf(otherDirection) and not self.isSameDirection(otherDirection):
            return True
        else:
            return False


    def __str__(self):
        return "(%s,%s)" % (self._x, self._y)

Direction.NORTH = Direction(0, 1)
Direction.SOUTH = Direction(0, -1)
Direction.EAST = Direction(1, 0)
Direction.WEST = Direction(-1, 0)


class Link:
    """This class is responsible for showing the connection between two nodes"""

    def __init__(self, canvas_view, canvas, start_node, end_node, start_point, end_point):
        """Set up the connection link
        start_node - GUI.guinode.GUINode, where the connection starts
        end_node - GUI.guinode.GUINode, where the connection ends
        start_line - (QPoint, QPoint)
        end_line - (QPoint, QPoint)
        start_point - QPoint, mouseposition (where on node user clicked)
        end_point - QPoint, mouseposition (where on node user clicked)

        """
        self._canvas = canvas
        self._start_node = start_node
        self._end_node = end_node
        self._canvas_view = canvas_view
        self._is_hidden = False
        self._is_selected = False
        self.start_line, self.start_index = self._start_node.get_closest_exit_line_with_index(start_point)
        # NB! End line is represented as p1(x,y) and p2(-x,y), which means the
        #     left-most point is stored in p2 (self.end_line[1]).
        self.end_line, self.end_index = self._end_node.get_closest_entry_line_with_index(end_point)

        if not self.start_line and not self.end_line:
            raise LinkCreationError(
                'These two nodes cannot be connected like that.\n' + \
                '\'%s\' cannot be a source node and \'%s\' cannot be a recevier node.' \
                % (self._start_node.getDiagramNode().getName(),
                   self._end_node.getDiagramNode().getName()))

        elif not self.start_line:
            raise LinkCreationError(
                'These two nodes cannot be connected like that.\n' + \
                '\'%s\' cannot be a source node.' \
                % self._start_node.getDiagramNode().getName())

        elif not self.end_line:
            raise LinkCreationError(
                'These two nodes cannot be connected like that.\n' + \
                '\'%s\' cannot be a receiver node.'\
                % self._end_node.getDiagramNode().getName())
            
        self._start_node.addStartLink(self) #this way the node will know of its start and end nodes
        self._end_node.addEndLink(self)
        
        self.old_colour = 0
        self.lines = [] # lines for this connection
        self.line_points = []

        self._init_route()
        self._auto_connect()
        
        # add the link to link list in module diagram; (from_id, to_id)
        md = self._canvas_view.get_module_diagram()
        md.add_node_link(self._start_node.getDiagramNode().getId(), 
                         self._end_node.getDiagramNode().getId())
        
        dprint(flags.GUI.CREATE.LINK,
               "Creating a new link between node " + self._start_node.getDiagramNode().getId() + \
               " and node " + self._end_node.getDiagramNode().getId())

    
    def _init_route(self):
        self.line_points = []
        
        # x coordinate for the start of the first link line segment.
        base_x_start = self._start_node.x() + self.start_line[1].x()
        base_y_start = self._start_node.y() + self.start_line[0].y()
       
        # x coordinate for the end of the last link line segment.
        base_x_end   = self._end_node.x() + self.end_line[1].x()
        base_y_end   = self._end_node.y() + self.end_line[0].y()
      
        # Add start point of start segment (short line out from node)
        self.line_points.append(QPoint(
            self.start_line[0].x() + self._start_node.x(),
            base_y_start))
        horizontal_distance = int(base_x_end - base_x_start)
        vertical_distance = int(base_y_end - base_y_start)
        if horizontal_distance == 0 or vertical_distance == 0:
            # Add only a single line (vertical or horizontal).
            self.line_points.append(QPoint(base_x_start, base_y_start))
            self.line_points.append(QPoint(base_x_end, base_y_end))
        
        else:
            p1 = QPoint(self._start_node.x() + self.start_line[1].x(),
                              self._start_node.y() + self.start_line[1].y())
            self.line_points.append(p1)
            if horizontal_distance > 0:
                # Add two horizontal lines and one vertical.
                breakpoint_x = int(base_x_start + (horizontal_distance /2))
                    
                p2 = QPoint(breakpoint_x, base_y_start) 
                p3 = QPoint(breakpoint_x, base_y_end) 
                self.line_points.append(p2)
                self.line_points.append(p3)

            else:
                # Add two vertical lines and one horizontal.
                start_line_length = abs(self.start_line[1].x() - self.start_line[0].x())
                end_line_length = abs(self.end_line[1].x() - self.start_line[0].x())
                breakpoint_y = int(base_y_start + int(vertical_distance /2))
                
                p2 = QPoint(base_x_start, breakpoint_y) 
                p3 = QPoint(base_x_end, breakpoint_y) 
                self.line_points.append(p2)
                self.line_points.append(p3)
                
            p4 = QPoint(self._end_node.x() + self.end_line[1].x(),
                        self._end_node.y() + self.end_line[1].y())
            self.line_points.append(p4)
        
        self.line_points.append(self._draw_adjusted_right_arrow(
            QPoint(base_x_end, base_y_end),
            QPoint(base_x_end + abs(self.end_line[1].x()), base_y_end)))
        self._draw_lines()

    def set_hidden(self, bool_):
        self._is_hidden = bool_
        if self._is_hidden:
            self.hide_hidden_links()

    def hide_hidden_links(self):
        if self._is_hidden:
            for line in self.lines:
                line.setCanvas(None)
            self.arrow.setCanvas(None)
        self._canvas.update()
        
    def show_hidden_links(self):
        for line in self.lines:
            line.setCanvas(self._canvas)
        self.arrow.setCanvas(self._canvas)
            
        self._canvas.update()
            
    def _draw_lines(self):
        self.lines = []
        
        # Draw start line with flat cap, or else the cap is likely to
        # "collide" with the node icon.
        if self._is_hidden:
            canvas = None
        else:
            canvas = self._canvas
        
        self.lines.append(CanvasLine(self, canvas, self.line_points[0], self.line_points[1], Qt.FlatCap))
        
        # Draw rest of lines with a round cap.
        for x in range(2, len(self.line_points)):
            self.lines.append(CanvasLine(self, canvas, self.line_points[x-1], self.line_points[x],Qt.RoundCap))
        

    def _draw_adjusted_right_arrow(self, start, end):
        arrow_points = []
        spacing = 2
        max_length = abs(end.x() - start.x()) - spacing # subtract for spacing 
        width = max(Config.lineWidth, 2) # enforce minimum size for arrow
        length = min(max_length, width*2)
        
        arrow_points.append(QPoint(end.x() - length, end.y() - length))
        arrow_points.append(QPoint(end.x() - length, end.y() + length))
        arrow_points.append(QPoint(end.x(), end.y()))

        if self._is_hidden:
            canvas = None
        else:
            canvas = self._canvas
            
        self.arrow = CanvasPolygon(self, arrow_points, QColor(Config.lineR, Config.lineB, Config.lineG),
                                   canvas)
        
        return QPoint(start.x() + (max_length - length) + spacing ,end.y())


    def delete(self):
        """removes a connection link"""

        self._start_node.deleteLink(self)
        self._end_node.deleteLink(self)
        
        startNode = self._start_node.getDiagramNode()
        endNode = self._end_node.getDiagramNode()

        endNode.removeConnectionsFromNode(startNode)
        md = self._canvas_view.get_module_diagram()
        md.remove_node_link(startNode.getId(), endNode.getId())

        for line in self.lines:
            line.delete()

        self._canvas_view.links.remove(self)
    
        self.arrow.delete()
        self._canvas.update()


    def hide(self):
        
        for line in self.lines:
            line.delete()

        line = []
        self.arrow.delete()
        
        

    def select(self):
        """Cange color on the lines when a lines is selected/clicked """

        for line in self.lines:
            line.changeColor(Qt.blue) #OBS: should invert linecolor here instead of useing Qt.blue

        self.arrow.changeColor(Qt.blue)
        self._is_selected = True
        

    def deSelect(self):
        """reverts back to the old color) """
        
        oldColor = QColor(Config.lineR, Config.lineG, Config.lineB)

        for line in self.lines:
            line.changeColor(oldColor)

        self.arrow.changeColor(oldColor)
        self._is_selected = False


    def isSelected(self):
        return self._is_selected

        
    def redraw(self):
        """Function used to move a connections """
        dprint(flags.GUI.MOVE.LINK, "Moving link between node " + self._start_node.getDiagramNode().getId() + " and node " + self._end_node.getDiagramNode().getId())
        self._init_route() 
        

        
    def update(self):
        """This is used when the user changes something in prefs in case he/she also changed the linecolor """
        
        for line in self.lines: 
            line.changeColor(QColor(Config.lineR, Config.lineG, Config.lineB)) 

        self.arrow.changeColor(QColor(Config.lineR, Config.lineG, Config.lineB))
        self.hide()
        self.redraw()


    def linkDialog(self):

        link_control = LinkControl(self._canvas_view,
                                   self._start_node.getDiagramNode(),
                                   self._end_node.getDiagramNode(),
                                   self.start_index, self.end_index)
        link_control.show()

    def _auto_connect(self):
        """method that will try to auto-connect ports"""

        startNode = self._start_node.getDiagramNode()
        endNode = self._end_node.getDiagramNode()

        outputPorts = startNode.getOutputPorts()
        inputPorts = endNode.getInputPorts()
        multi_input_ports = endNode.get_multi_input_ports()

        if (len(inputPorts) == 1) and (len(outputPorts) == 1) and\
               not multi_input_ports:
            inputPorts[0].connectToDataSource(outputPorts[0])
            outputPorts[0].setPosition(int(inputPorts[0].getId()),
                                       self.start_index, self.end_index)
            
        elif len(multi_input_ports) == 1 and not inputPorts:
            # connect all output nodes if input node is MultipleInputPort
            for output_port in outputPorts:
                multi_input_ports[0].add_data_source(output_port)

    def contextMenu(self, point):
        menu = LinkContext(self, self._canvas_view)
        menu.popup(point)

    def get_start_node(self):
        return self._start_node

    def get_end_node(self):
        return self._end_node

    def get_nodes(self):
        return (self._start_node, self._end_node)

    def is_hidden(self):
        return self._is_hidden

class LinkControl:
    """Dialog for editing a link in a module diagram."""
    
    def __init__(self, parent, startNode, endNode, startIndex, endIndex):
        self._connections = {}

        self._start_node = startNode
        self._end_node = endNode
        
        self._start_index = startIndex
        self._end_index = endIndex

        self._output_ports = []
        self._input_ports = []

        self._connections = {}
        self._new_connections = {}
        self._deleted_connections = {}
        
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir, "link_control.ui"), None, parent)
        
        self._output_box = self._widget.child('output_box')
        self._input_box = self._widget.child('input_box')

        frame = self._widget.child('line_drawer_frame')
        self._line_draw_widget = LineDrawWidget(frame, self._draw_links, self._widget)

        QObject.connect(self._output_box, SIGNAL("highlighted(int)"),
                     self._slot_clicked_output)
        QObject.connect(self._input_box, SIGNAL("highlighted(int)"),
                     self._slot_clicked_input)

        connect_button = self._widget.child('connect_button')
        disconnect_button = self._widget.child('disconnect_button')
        QObject.connect(connect_button, SIGNAL("clicked()"), self._slot_connect)
        QObject.connect(disconnect_button, SIGNAL("clicked()"),
                     self._slot_disconnect)

        self._name_edit = self._widget.child('name_edit')
        self._type_edit = self._widget.child('type_edit')
        self._description_edit = self._widget.child('description_edit')

        help_button = self._widget.child('help_button')
        ok_button = self._widget.child('ok_button')
        cancel_button = self._widget.child('cancel_button')

        QObject.connect(help_button, SIGNAL("clicked()"), self._slot_help)
        QObject.connect(ok_button, SIGNAL("clicked()"), self._slot_ok)
        QObject.connect(cancel_button, SIGNAL("clicked()"), self._widget.close)

        for port in startNode.getOutputPorts():
            # index in outputPorts equals index in outputBox
            self._output_ports.append(port)
            self._output_box.insertItem(str(port.getName()))

        for port in self._end_node.getInputPorts():
            # index in inputPorts equals index in inputBox
            self._input_ports.append(port)
            self._input_box.insertItem(str(port.getName()))

        for port in self._end_node.get_multi_input_ports():
            # index in inputPorts equals index in inputBox
            self._input_ports.append(port)
            self._input_box.insertItem(port.get_name())

            # for multiinputports we assign connections right away
            for output_port in port.get_data_sources():
                if output_port in self._output_ports:
                    inputIndex = self._input_ports.index(port)
                    outputIndex = self._output_ports.index(output_port)
                    self._connections[(outputIndex,inputIndex)] = (output_port, port)
            

        # generate a dict with connections
        # - using index-ids (from the QListBox-lists) as keys
        # - the element is a list [changed, connected/disconnected, outputID, inputID]
        for connection in self._start_node.getConnectionsToNode(self._end_node):
            outputIndex = self._output_ports.index(connection[0])
            inputIndex = self._input_ports.index(connection[1])
            self._connections[(outputIndex,inputIndex)] = connection

    def show(self):
        self._widget.show()

    def _draw_links(self, painter):
        for key in self._connections:
            output, input = self._connections[key]

            outputIndex = self._output_ports.index(output)
            inputIndex = self._input_ports.index(input)

            outputItem = self._output_box.item(outputIndex)
            inputItem = self._input_box.item(inputIndex)

            outputRect = self._output_box.itemRect(outputItem)
            inputRect = self._input_box.itemRect(inputItem)

            # possition of the output port
            if outputRect.isValid(): #returnes FALSE if the item is not visable
                outputPos = QPoint(outputRect.x() + self._output_box.x() + self._output_box.width() + 3,
                                   outputRect.y() + self._output_box.y() + 8)
            else:
                #need to find it if the item is bellow the bottom or over the top
                if (outputIndex < len(self._output_ports)/2):
                    # should the over the top
                    outputPos = QPoint(self._output_box.x() + self._output_box.width() + 3,
                                       self._output_box.y())
                else:
                    # should be bellow the bottom
                    outputPos = QPoint(self._output_box.x() + self._output_box.width() + 3,
                                       self._output_box.y() + self._output_box.height())
                    
            # possition of the input port
            if inputRect.isValid():
                inputPos = QPoint(inputRect.x() + self._input_box.x() - 3,
                                  inputRect.y() + self._input_box.y() + 8)
            else:
                #need to find it if the item is bellow the bottom or over the top
                if (inputIndex < len(self._input_ports)/2):
                    # should the over the top
                    inputPos = QPoint(self._input_box.x(), self._input_box.y())
                else:
                    # should be bellow the bottom
                    inputPos = QPoint(self._input_box.x(),
                                      self._input_box.y() + self._input_box.height())

            painter.drawLine(outputPos, inputPos)
            
    def _slot_clicked_output(self, index):
        port = self._output_ports[index]

        self._name_edit.setText(str(port.getName()))
        self._type_edit.setText(str(port.getDataType()))
        self._description_edit.setText(str(port.getDescription()))

    def _slot_clicked_input(self, index):
        port = self._input_ports[index]

        if isinstance(port, MultipleInputPort):
            self._name_edit.setText(port.get_name())
            self._type_edit.setText("Multi-input port")
            self._description_edit.setText("Several output-ports can be connected to this port")
        else:
            self._name_edit.setText(str(port.getName()))
            self._type_edit.setText(str(port.getDataType()))
            self._description_edit.setText(str(port.getDescription()))

    def _slot_connect(self):
        outputIndex = self._output_box.currentItem()
        inputIndex = self._input_box.currentItem()

        if inputIndex == -1 or outputIndex == -1:
            return
        
        key = (outputIndex,inputIndex)

        if self._deleted_connections.has_key(key):
            del self._deleted_connections[key]
        
        busy_connections = {}
        #Checks if target ports in other connections
        for busy_output_index, busy_input_index in self._connections:
            # ignore multiple_input_nodes (which can have more than
            # one input)
            if isinstance(self._input_ports[busy_input_index],
                          MultipleInputPort):
                continue
            
            if busy_input_index == inputIndex:
                busy_connections[(busy_output_index, busy_input_index)] = 1

        #Deletes conflicting connects found abouve
        for busykey in busy_connections:
            self._disconnect_key(busykey)

        # newConnections is a set, value isn't important
        self._new_connections[key] = 1
        
        if not self._connections.has_key(key):
            self._connections[key] = [self._output_ports[outputIndex],
                                      self._input_ports[inputIndex]]

        # update that sends a paintEvent
        self._line_draw_widget.update()

    def _slot_disconnect(self):
        outputIndex = self._output_box.currentItem()
        inputIndex = self._input_box.currentItem()

        if inputIndex == -1 or outputIndex == -1:
            return
        
        key = (outputIndex,inputIndex)

        self._disconnect_key(key)
        
        self._line_draw_widget.update()

    def _disconnect_key(self, key):
        if self._new_connections.has_key(key):
            del self._new_connections[key]
            
        if self._connections.has_key(key):
            del self._connections[key]
            
            if not self._deleted_connections.has_key(key):
                self._deleted_connections[key] = [self._output_ports[key[0]],
                                                  self._input_ports[key[1]]]
       
    def _slot_ok(self):

        for key in self._deleted_connections:
            output_port, input_port = self._deleted_connections[key]

            if isinstance(input_port, MultipleInputPort):
                input_port.remove_data_source(output_port)
            else:
                input_port.disconnectFromDataSource()

        for key in self._new_connections:
            output_port, input_port = self._connections[key]

            if isinstance(input_port, MultipleInputPort):
                input_port.add_data_source(output_port)
            else:
                input_port.connectToDataSource(output_port)
                output_port.setPosition(int(input_port.getId()), self._start_index,
                                        self._end_index)
        
        self._widget.close()
        del self._widget
        
    def _slot_help(self):
        HelpViewer(self._line_draw_widget, os.path.join(Config.htmldocpath,
                                      "node_User_interface.html"))



class LineDrawWidget(QWidget):
    """Display lines for connections."""
    def __init__(self, parent, draw_function, other_widget):
        QWidget.__init__(self, parent)
        self.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding)
        self._draw_function = draw_function
        self._other_widget = other_widget

    def paintEvent(self, event):
        painter = QPainter(self, True)
        pen = QPen(Qt.black, 2)
        painter.setPen(pen)

        other_origin = self._other_widget.mapToGlobal(QPoint(0,0))
        origin = self.mapToGlobal(QPoint(0,0))
        
        painter.translate(other_origin.x() - origin.x(),
                    other_origin.y() - origin.y())

        self._draw_function(painter)
