# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os

from qt import QMainWindow, Qt, QWidget, QGridLayout, QPopupMenu, QAction
from qt import QAccel, SIGNAL, QApplication, QMessageBox, QFileDialog, QString
from qt import QCursor
from qttable import QTable
from qtui import QWidgetFactory
from GUI.HelpViewer import HelpViewer
from Library.Nodes.FileHandlers.FileWriterNode import FileWriterNode
from Library.Util.Config import Config
from Library.Util.lists import get_repr_dim
from Numeric import array

class UnknownExtensionError(Exception):
    """Signalises that a filename extension is unknown."""
    pass

class SpreadSheetWindow:
    """This window contains the spreadsheet.

    This Window is a basic spreadsheet.
    """
    def __init__(self):
        self._window = QWidgetFactory.create(os.path.join(Config.uidir,
                                                    "spreadsheetwindow.ui"),
                                             None);
        self._table = self._window.child("table")

        self._columnwise_orientation = True
        self._data = None
        self._file_writer_node = FileWriterNode()
        self._statusbar_perm_msg = ""

        self._window.connect(self._window.child("save_action"),
                             SIGNAL("activated()"),
                             self._slot_save_data)
        self._window.connect(self._window.child("windowCloseAction"),
                             SIGNAL("activated()"),
                             self._slot_close)
        self._window.connect(self._window.child("editCopyAction"),
                             SIGNAL("activated()"),
                             self._slot_copy_to_clipboard)
        self._window.connect(self._window.child("editSelect_AllAction"),
                             SIGNAL("activated()"),
                             self._slot_select_all)
        self._window.connect(self._window.child("viewColumn_wiseAction"),
                             SIGNAL("toggled(bool)"), 
                             self._slot_view_orientation_col)
        self._window.connect(self._window.child("viewRow_wiseAction"),
                             SIGNAL("toggled(bool)"), 
                             self._slot_view_orientation_row)
        self._window.connect(self._window.statusBar(),
                             SIGNAL("messageChanged(const QString &)"),
                             self._slot_statusbar_msg_changed)
        self._window.connect(self._window.child("helpHelpAction"),
                             SIGNAL("activated()"),
                             self._slot_help)

    def set_view_orientation_columnwise(self, flag):
        """Sets the view orientation of the table.

        If a vector is shown columnwise, the whole vector is in one column,
        spanning multiple rows. If a vector is shown rowwise, the whole vector
        is in one row, spanning multiple columns. The flag is True if the data
        is to be displayed columnwise, else False.
        """
        self._columnwise_orientation = flag;

    def show(self):
        """Shows the window if data has been set."""
        if self._data:
            self._auto_adjust_cells()
            self._adjust_window()
            self._window.show()
            self._window.raiseW()
        else:
            QMessageBox.information(self._window, "No data in spreadsheet",
                                    "There is no data in the spreadsheet.\n" +\
                                    "Most likely this means you have not\n" +\
                                    "run the diagram yet.", QMessageBox.Ok)

    def set_data(self, data):
        """Sets the data that the spreadsheet shall show.

        The data provided is a list of System.Data.Data.Data-objects.
        """
        if data:
            self._data = data
            self._table.setNumRows(0)
            self._table.setNumCols(0)
            for data_object in data:
                raw_data = data_object.getData()
                if type(raw_data) == type(array(0)):
                    raw_data = raw_data.tolist()

                name = data_object.getName()
                if type(raw_data) == type(list()):
                    if get_repr_dim(raw_data) == 1:
                        self._add_vector_to_table(name, raw_data)
                    elif get_repr_dim(raw_data) == 0:
                        # we have a list with dimension 0, which means it is
                        # empty: we assert to rule out an "impossible situation"
                        assert len(raw_data) == 0
                        self._add_scalar_to_table(name, "")
                    else:
                        self._add_matrix_to_table(name, raw_data)
                else:
                    self._add_scalar_to_table(name, raw_data)
                    
    def _add_matrix_to_table(self, name, matrix):
        if self._columnwise_orientation:
            offset = self._table.numCols()
            if len(matrix) > self._table.numRows():
                self._table.setNumRows(len(matrix))

            longest_col = self._get_longest_row(matrix)
            
            self._table.setNumCols(offset + longest_col)
            for col in range(longest_col):
                self._table.horizontalHeader().setLabel(
                    offset + col, "%s %d" % (name, col + 1))

            for row_index in range(len(matrix)):
                for col_index, value in enumerate(matrix[row_index]):
                    self._table.setText(row_index, offset + col_index,
                                        str(value))

        else:
            offset = self._table.numRows()
            if len(matrix) > self._table.numCols():
                self._table.setNumCols(len(matrix))

            longest_col = self._get_longest_row(matrix)

            self._table.setNumRows(offset + longest_col)
            for row in range(longest_col):
                self._table.verticalHeader().setLabel(offset + row,
                                                      "%s %d" % (name, row + 1))

            for col_index in range(len(matrix)):
                for row_index, value in enumerate(matrix[col_index]):
                    self._table.setText(offset + row_index, col_index,
                                        str(value))
            

    def _add_vector_to_table(self, name, vector):
        if self._columnwise_orientation:
            offset = self._table.numCols()
            if len(vector) > self._table.numRows():
                self._table.setNumRows(len(vector))

            self._table.setNumCols(offset + 1)
            self._table.horizontalHeader().setLabel(offset, name)

            for row_index, value in enumerate(vector):
                if type(value) == type(list()):
                    value = value[0]
                self._table.setText(row_index, offset, str(value))
        else:
            offset = self._table.numRows()
            if len(vector) > self._table.numCols():
                self._table.setNumCols(len(vector))
                
            self._table.setNumRows(offset + 1)
            self._table.verticalHeader().setLabel(offset, name)

            for col_index, value in enumerate(vector):
                if type(value) == type(list()):
                    value = value[0]
                self._table.setText(offset, col_index, str(value))

    def _add_scalar_to_table(self, name, scalar):
        if self._columnwise_orientation:
            offset = self._table.numCols()
            if self._table.numRows() < 1:
                self._table.setNumRows(1)
            self._table.setNumCols(offset + 1)
            self._table.horizontalHeader().setLabel(offset, name)
            self._table.setText(0, offset, str(scalar))
        else:
            offset = self._table.numRows()
            if self._table.numCols() < 1:
                self._table.setNumCols(1)
            self._table.setNumRows(offset + 1)
            self._table.verticalHeader().setLabel(offset, name)
            self._table.setText(offset, 0, str(scalar))

    def _get_longest_row(self, vectors):
        longest_row = 0
        for row in vectors:
            if longest_row < len(row):
                longest_row = len(row)
        return longest_row

    def _adjust_window(self):
        """Make sure window is centered and sized reasonably."""
        self._window.adjustSize()
        new_x = (QApplication.desktop().width() - self._window.width()) / 2
        new_y = (QApplication.desktop().height() - self._window.height()) / 2
        self._window.move(new_x, new_y)

    def _auto_adjust_cells(self):
        """Shrink/expand cells to fit data."""

        #since QTable cannot update the correct size of it's own QHeader we have
        #to do some magic... Below QFontMetrics is used to find the bounding
        #rectangle on all sections in the header. Then the width of the
        #rectangle is measured up against the previous width set. initially, we
        #set the comparison width to zero. wenn all sections is compared up
        #against the width variable, we can set the left margin of the QTable.

        from qt import QFontMetrics

        font = self._table.verticalHeader().font()
        fm = QFontMetrics(font)

        width = 0

        for index in range(self._table.verticalHeader().count()):
            label = self._table.verticalHeader().label(index)
            width = max(fm.boundingRect(label).width(), width)

        # setting a border and padding since it currently isn't possible to
        # retrieve this from Qt.
        padding_and_border = 12

        self._table.setLeftMargin(width + padding_and_border)

        for y in range(self._table.numRows()):
            self._table.adjustRow(y)
        for x in range(self._table.numCols()):
            self._table.adjustColumn(x)

    def _save_data(self, filepath):
        """Saves the data to the given filepath."""
        writers = self._file_writer_node.get_writers_by_extension()
        path, ext = os.path.splitext(filepath)
        # find the writer
        if ext[1:] not in writers.keys():
            raise UnknownExtensionError, ext 

        writer = writers[ext[1:]] # not the dot
        writer.write_data(self._data, filepath) 

    def _copy_to_clipboard(self):
        """Copy the contents of the table to the clipboard."""
        COL_SEP = "\t"
        ROW_SEP = os.linesep
        contents = QString("")

        if self._table.numSelections() < 1:
            return

        # only one selection
        selection = self._table.selection(0)
        
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        for row in range(selection.topRow(), selection.bottomRow() + 1):
            for col in range(selection.leftCol(), selection.rightCol() + 1):
                if col > selection.leftCol():
                    contents.append(COL_SEP)
                contents.append(self._table.text(row, col))

            contents.append(ROW_SEP)

        cb = QApplication.clipboard()

        cb.setText(contents, cb.Clipboard)
        if cb.supportsSelection():
            cb.setText(contents, cb.Selection)

        QApplication.restoreOverrideCursor()

    def __reload_data_to_table(self):
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        self._table.setUpdatesEnabled(False)
        self.set_data(self._data)
        self._auto_adjust_cells()
        self._window.adjustSize()
        self._table.setUpdatesEnabled(True)
        QApplication.restoreOverrideCursor()
        # update to remove "ghosts" in cells not updated
        self._table.updateContents()

    # slots
    def _slot_view_orientation_col(self, on):
        self.set_view_orientation_columnwise(on)
        self.__reload_data_to_table()

    def _slot_view_orientation_row(self, on):
        self.set_view_orientation_columnwise(not on)
        self.__reload_data_to_table()

    def _slot_save_data(self):
        """Queries the user for a file name to save data to."""
        filters = self._file_writer_node.get_filters()
        filepath = QFileDialog.getSaveFileName("",
                                               filters,
                                               self._window,
                                               "save file dialog",
                                               "Choose a filename to save " +\
                                               "the data under")
        if filepath:
            try:
                self._save_data(str(filepath))
            except UnknownExtensionError, e:
                QMessageBox.information(self._window, "Unknown file type",
                                        "Could not recognize the file type " +\
                                        "from the extension %s." % e,
                                        QMessageBox.Ok)
            except Exception, e:
                QMessageBox.information(self._window, "Error writing file",
                                        str(e), QMessageBox.Ok)

    def _slot_close(self):
        self._window.close()

    def _slot_statusbar_msg_changed(self, message):
        if message == QString():
            self._statusbar_perm_msg = message
            # Redraw permanent statusbar message.
            self._window.statusBar().message(self._statusbar_perm_msg)

    def _slot_help(self):
        HelpViewer(self._window, os.path.join(Config.htmldocpath, 
                                              "node_SpreadSheet_Node.html"),
                   Config.htmldocpath)

    def _slot_copy_to_clipboard(self):
        self._copy_to_clipboard()

    def _slot_select_all(self):
        self._table.removeSelection(self._table.currentSelection())
        self._table.selectCells(0, 0, self._table.numRows(),
                                self._table.numCols())
        
