# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


import os

from System.Data import Parameter
from System.Data.Node import Node
from System.events import NodeEvent

from Library.Util import lists

from System.FileHandlers.Input.matfilereader import MatFileReader
from System.FileHandlers.Input.BASEFileReader import BASEFileReader
from System.FileHandlers.Input.OctaveFileReader import OctaveFileReader
from System.FileHandlers.Input.RFileReader import RFileReader
from System.FileHandlers.Input.MoleculeFileReader import MoleculeFileReader
from System.FileHandlers.Input.textfilereader import TextFileReader
from System.FileHandlers.Input.GaussDalFileReader import GaussDalFileReader
from System.FileHandlers.Input.experimentfilereader import ExperimentFileReader

# domain exception
class UnknownReaderError(Exception):
    """FileReader unset or unknown."""
    

class ReadError(Exception):
    """Error reading data from file."""
    

class InputNode(Node):
    """
    A node that can read data from file and store it on its NodeOutputPorts.
    To be able to do this, parameters[0] MUST contain a valid filename path so
    that the FileReader can find a file to read from.

    The execution method will read the file and try to match the received data
    with the required variable names on the different NodeOutputPorts. """
    
    
    def __init__(self, name="InputNode"):
        """
        Params:
               - name : String - a String with the name of this Node
               - filereader : FileReader - the file reader that should do the
               reading work for this node"""
        Node.__init__(self, name)
    
        self.description = "This node reads data from file."
        #self.variable_name_support = False
        self.setCanAddInputPorts(False)
        self.setCanAddOutputPorts(True)
        self.fileReader = None
        
        self.datalist = None
        osexample = os.path.join('..','data','myfile.ext')
        self.filename_parameter = self.addParameter("Filename", "Fileload", """The filename of the data file.

If you want path relative to module diagram file, type in a relative path manually (ex: %s)."""%osexample)
 
        self.fileReaderList = [
            MatFileReader,
            BASEFileReader,
            OctaveFileReader,
            RFileReader,
            MoleculeFileReader,
            TextFileReader,
            GaussDalFileReader,
            ExperimentFileReader
            ]

        self.file_reader_parameter = self.addParameter("FileReader", "Text", "Unknown")
    
        self.fileTypeList = ["Unknown"]
        for reader in self.fileReaderList:
            self.fileTypeList.append(reader.fileDescription)
        #oyle 
        self._separator = None

    def get_possible_output_ports(self, source_path, filetype):
        reader = self._get_reader(filetype)
        ports = []
        source_path = self._make_path_absolute(source_path)
        if reader and self.file_exists(source_path):
            variables = reader.get_variable_names_safe(source_path)
            filetype_ = self.file_reader_parameter.getValue() 
            filename = self.filename_parameter.getValue() 
            filename = self._make_path_absolute(filename)

            if filetype == filetype_ and filename == source_path:
                chosen_ports = map(lambda port:port.getName(), self.getOutputPorts())
                variables = lists.filter_list(variables, chosen_ports)

            if variables:
                info_dict = reader.get_variable_info(source_path)
                for variable in variables:
                    ports.append((variable, info_dict[variable]))

        return ports

    def get_all_variable_names_in_file(self, source_path, filetype):
        reader = self._get_reader(filetype)
        variable_names = []
        source_path = self._make_path_absolute(source_path)
        if reader and self.file_exists(source_path):
            return reader.get_variable_names_safe(source_path)

    def get_current_output_ports(self):
        outports = self.getOutputPorts()
        current_ports = []
        for port in outports:
            name = port.getName()
            info = port.getDescription()
            current_ports.append((name, info))

        return current_ports


    def get_chosen_output_ports(self, filename, filetype):
        # if the contens of the new and the old file is the same, return the current output ports
        if self.get_possible_output_ports(self.get_current_filename(), self.get_current_filetype()) == \
           self.get_possible_output_ports(filename, filetype):
            return self.get_current_output_ports()
        return []

    def get_default_output_ports(self, filetype):
        reader = self._get_reader(filetype)
        ports = []
        output_ports = self.get_current_output_ports()
        if reader:
            outports = reader.get_default_output_ports()
            for port in outports:                
                name, dummy, info = port
                if name not in map(lambda tuple_: tuple_[0],
                                   output_ports):
                    ports.append((name, info))

        return ports
        

    def get_filetype_description(self, filetype):
        reader = self._get_reader(filetype)
        if reader:
            return reader.get_description()

        return "Unknown filetype."


    def get_current_filetype(self):
        if self.fileReader:
            return self.fileReader.get_filetype()

        return self.fileTypeList[0]


    def get_compatible_filetype(self, filename):
        """This method returns a filetype compatible with the given filename.
        If none found, returns "Unknown"."""
        
        extension = filename.split(".")[-1].lower()
        
        for reader in self.fileReaderList:
            if extension in map(lambda ext:ext.lower(), reader.supportedFileExtensions):
                return reader.fileDescription        
       
        return self.fileTypeList[0]


    def get_current_filename(self):
        if self.filename_parameter:
            return self.filename_parameter.getValue()

        return ""


    def set_filename(self, filename):
        self.filename_parameter.setValue(filename)


    def set_filetype(self, filetype):
        self.file_reader_parameter.setValue(filetype)
        self._setFileReader(filetype)


    def delete_port(self, port_name):
        port = self.getOutputPortByName(port_name)
        if port:
            port.remove()


    def file_exists(self, filename):
        if not filename:
            return False

        filename = self._make_path_absolute(filename)
        
        return os.path.isfile(filename)


    def _get_reader(self, filetype):
        for reader in self.fileReaderList:
            if reader.fileDescription == filetype:
                return reader()

        return None


    def _make_path_absolute(self, path):
        if not path:
            return path

        path = str(path)
        
        if os.path.isabs(path):
            abspath = path
        else:
            md_path = self.get_module_diagram_filename()
            if md_path:
                abspath = os.path.abspath(os.path.join(os.path.dirname(md_path), path))
            else:
                abspath = os.path.abspath(path)

        return abspath

        
    def _setFileReader(self, filetype):
        self.fileReader = None
        self.clean()
        
        reader = self._get_reader(filetype)
        if reader:
            self.getParameterByName("Filereader").setValue(filetype)
            self.fileReader = reader
        
            
    def getFileTypeList(self):
        return self.fileTypeList
    
        
    def setFileHandler(self, fileReader):
        for reader in self.fileReaderList:
            if reader.name == fileReader:
                self.fileReader = reader()
                self.addParameter("FileReader", "Text", "", value=reader.fileDescription)
                return
    
        
    def getCompatibleFileType(self, filename):
        """
        This method returns a FileType compatible with the given filename.
        If none found, return 'Unknown'."""
        
        extension = filename.split(".")[-1].lower()
        
        for reader in self.fileReaderList:
            if extension in map(lambda ext:ext.lower(), reader.supportedFileExtensions):
                return reader.fileDescription        
       
        return self.fileTypeList[0] #"Unknown"


    def createFilter(self, reader):
        filter = reader.fileDescription
        extensions = reader.supportedFileExtensions
        if extensions:
            filter += " (*." + extensions[0]
        for extension in reader.supportedFileExtensions[1:]:
            filter += " *." + extension

        filter += ")"

        return filter


    def getFilters(self):
        filters = "All files (*.*)"
        readers = self.fileReaderList
        if readers:
            reader = readers[0]
            filter = self.createFilter(reader)
            filters += ";;" + filter

            for reader in readers[1:]:
                filter = self.createFilter(reader)
                filters += ";;" + filter

        return filters

        
    def add_input_port(self, name, validType, description):
        """
        Params:
               - name : String - the name of the port
               - validType : String - the data type this port should handle
               - description : String - a descriptive string of this port

        This method does nothing, as you cannot
        add a NodeInputPort to an InputNode"""
        pass 
    
    
    def execute(self):
        """Read file and place results on output ports.

        Exceptions:

        UnknownReaderError - Raised when reader is unset or unknown
        ReadError - Raised when error reading data
        """
        Node.execute(self)
        
        if self.fileReader == None:
            raise UnknownReaderError, "FileType in InputNode is not set or unknown"
        
        source = ""
        if not self.datalist:
            # Checks if path have been set, otherwise raise exception
            if self.getParameterByName("fileName").isValueSet():
                source = str(self.getParameterByName("fileName").getValue())
                if not os.path.isabs(source):
                    # treat as relative to modulediagram
                    mdpath = self.moduleDiagram.getFilename()
                    if mdpath:
                        source = os.path.join(os.path.dirname(mdpath), source)

                # Catch exceptions here, as the read_data() method is
                # implemented by "outsiders".
                try:
                    self.datalist = self.read_data(source)
                except Exception, e:
                    raise ReadError, "Error reading from file %s: %s" % (source, e)
            else:
                raise ReadError, "Filename not specified in inputnode."

        # Traverse ports
        for port in self.outPorts:
            found = False

            # Traverse data
            for data in self.datalist:
                # Check if matching data is found
                if port.getVariableName() == data.getName():
                    port.setData(data)
                    found = True
            
            # If no data is found for specified output port, send an event.
            if not found:
                raise ReadError, "Variable '%s' not found in file '%s'." % (port.getVariableName(), source)
   
        self.datalist = None # throwing out the unwanted data out of memory as soon as possible!


    def clean(self):
        self.datalist = None
        Node.clean(self)

                
    def read_data(self, source):
        if self.fileReader:
            if (self.fileReader.name == "TextFileReader") and self._separator:
                symbols = self._separator.get_symbols()
                row = symbols['row']
                col = None
                if symbols['col']:
                    col = symbols['col']
                return self.fileReader.read_data(source, split_on = row, col_split = col)
            else:
                return self.fileReader.read_data(source)

    
    def get_variable_info(self, sourcePath, variable_name = None):
        if self.fileReader:
            return self.fileReader.get_variable_info(sourcePath, variable_name)
        return ""
        
            
    def getFileReaderClassName(self):
        if self.fileReader:
            return self.fileReader.name
        else:
            return "Unknown"
        
        
    def addParameter(self, name, type, description, min=None, max=None, step=None, values=None, default="", value=None):
        if name == "FileReader" and value:
            self._setFileReader(value)
        if name == "FileReader" and not value:
            self._setFileReader("Unknown")
        return Node.addParameter(self, name, type, description, min, max, step, values, default, value)
        

    def get_module_diagram_filename(self):
        if self.moduleDiagram:
            return self.moduleDiagram.getFilename()

        return None

    def textfile_separator(self):
        if not self._separator:
            self._separator = lists.Separator()
        return self._separator

    def toXML(self):
        """Inserts xml of selections into Node's XML 
           (called from Node upon save)"""
        self.nodeRepName
        xml_list = Node.toXML(self)
        xml_swap = xml_list[:]
        if self.fileReader:
            if self.fileReader.name == "TextFileReader" and self._separator:
                xml_swap = xml_list[:-1]
                xml_swap.extend(self._separator.to_XML().split('\n'))
                xml_swap.append(xml_list[-1])
        return xml_swap

    def from_text_input_xml(self, all):
        if not self._separator:
            self._separator = lists.Separator()
        self._separator.set_selection(all)

        

