# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os
from System.Data import Parameter
from System.Data import Single
from System.Data.Node import Node
from System.events import NodeEvent
from System.FileHandlers.Output.FileWriter import FileWriter
from System.FileHandlers.Output.MoleculeFileWriter import MoleculeFileWriter
from System.FileHandlers.Output.OctaveFileWriter import OctaveFileWriter
from System.FileHandlers.Output.RFileWriter import RFileWriter

# domain exceptions
class OutputError(Exception):
    """Base class of all exceptions from OutputNode."""


class UnknownWriterError(OutputError):
    """FileWriter unset or unknown."""


class WriteError(OutputError):
    """Error when writing data to file."""


class OutputNode(Node):
    """A Node that can write data to file.
    
    This node has not defined any input ports, and no output ports can
    be defined at all.  The user has to define the input ports he/she
    needs. When the node executes, it will then traverse all input
    ports and write the variables contained there to the same data
    file.

    Deprecated in favour of FileWriterNode.
    """

    supportedFileExtensions = []
    
    def __init__(self, name="OutputNode"):
        """
        Params:
               - name : String - a String with the name of this node
               - writer : FileWriter - a FileWriter capable of writing data to a file.
               Must be a subclass of the class FileWriter!"""
        Node.__init__(self, name)
        self.setCanAddInputPorts(True)
        self.setCanAddOutputPorts(False)
        self.description = "This node will write data to file."
        osexample = os.path.join('..','data','myfile.ext')
        self._file_parameter = self.addParameter("Filename", "Filesave",
                                                 """The filename of the data file.

If you want path relative to module diagram file, type in a relative path manually (ex: %s)."""%osexample)

        fileTypeParameter = self.addParameter("Filetype", "Text",
                                              "The filetype of the data file.")
        fileTypeParameter.setValue("Unknown")

        self.fileWriterList = []
        self.fileTypeList = ["Unknown"]

        for fileWriter in [RFileWriter, MoleculeFileWriter, OctaveFileWriter]:
            self._addFileWriter(fileWriter)

        self.addParameter("FileWriter", "Text", "")

        self.fileWriter = None


    def _addFileWriter(self, fileWriter):
        self.fileWriterList.append(fileWriter)
        self.fileTypeList.append(fileWriter.fileDescription)

            
    def getFileTypeList(self):
        """This method returns the different filetypes this output node supports."""
        return self.fileTypeList


    def getCompatibleFileType(self,source):
        """This method takes as input a filename and returns the filetype of this file.

        If filetype is not supported in any of the FileWriter classes, the method returns 'Unknown'.
        """

        extension = source.split(".")[-1].lower()

        for writer in self.fileWriterList:
            if extension in map(lambda ext:ext.lower(), writer.supportedFileExtensions):
                return writer.fileDescription

        return self.fileTypeList[0] #Unknown


    def createFilter(self, writer):
        filter = writer.fileDescription
        extensions = writer.supportedFileExtensions
        if extensions:
            filter += " (*." + extensions[0]
        for extension in writer.supportedFileExtensions[1:]:
            filter += " *." + extension

        filter += ")"

        return filter


    def getFilters(self):
        filters = "All files (*.*)"
        writers = self.fileWriterList
        if writers:
            writer = writers[0]
            filter = self.createFilter(writer)
            filters += ";;" + filter

            for writer in writers[1:]:
                filter = self.createFilter(writer)
                filters += ";;" + filter

        return filters
            

    def getFileWriterClassName(self):
        classname = "Unknown"
        if self.fileWriter:
            classname = self.fileWriter.name

        return classname
    
    
    def setFileType(self, filetype):
        """Set filetype for this outputnode.

        The filtype for a outputnode states to which types of files the node can write"""
        
        #self._setFileWriter(filetype)
        self.addParameter("FileWriter", "Text", "", value=filetype)


    def getFileType(self):
        if self.fileWriter:
            return self.fileWriter.fileDescription
        
        return "Unknown"


    def getFileTypeDescription(self, fileType):
        for filewriter in self.fileWriterList:
            if filewriter.fileDescription == fileType:
                return filewriter.description

        return "Unknown file type"


    def _setFileWriter(self, filetype):
        self.fileWriter = None
        for writer in self.fileWriterList:
            if filetype ==  writer.fileDescription:
                self.fileWriter = writer()
                return


    def setFileHandler(self, fileWriter):
        self.fileWriter = None
        for writer in self.fileWriterList:
            if writer.name == fileWriter:
                self.fileWriter = writer()
                return

    
    def add_output_port(self, name, validType, description):
        """
        Params:
               - name : String - the name of the port you want to add
               - validType : String - the data type of the port
               - description : String - a description of the port
        The method does nothing, as you cannot add a
        NodeOutputPort to an OutputNode"""
        pass
       
       
    def write_data(self, dataList, destinationPath):
        self.fileWriter.write_data(dataList, destinationPath)


    def get_current_filetype(self):
        if self.fileWriter:
            return self.fileWriter.get_filetype()

        return self.fileTypeList[0]


    def execute(self):
        """Store data on input ports to file.

        Takes all data on the input ports, concatenates them and sends
        them to a FileWriter, specifying a path to where the data are
        to be written.
        """
        Node.execute(self)

        if not self.fileWriter:
            raise UnknownWriterError, "File type in OutputNode not specified or unknown"

        datalist = []

        # Traverses all ports and gets the data
        for port in self.inPorts:
            
            # Fetches the data of the port
            data = port.getData()
            
            # Sets the name of the data to the name of the port
            data.name = port.getVariableName()  
            
            #Appends to the list that is to be sent to the writer
            datalist.append(data)


        destinationPath = ""

        if self._file_parameter.isValueSet() and self._file_parameter.getValue():
            destinationPath = self._file_parameter.getValue()
        else:
            raise WriteError, 'Path to file not specified'

        if not os.path.isabs(destinationPath):
            # treat as relative to modulediagram
            mdpath = self.moduleDiagram.getFilename()
            if mdpath:
                destinationPath = os.path.join(os.path.dirname(mdpath), destinationPath)
               
        try:
            self.write_data(datalist, destinationPath)
        except Exception, e:
            raise WriteError, "Error while writing to %s: %s" % (destinationPath, e)

    def addParameter(self, name, type, description, min=None, max=None, step=None, values=None, default = "", value=None):
        if name == "FileWriter" and value:
            self._setFileWriter(value)
        if name == "FileWriter" and not value:
            self._setFileWriter("Unknown")
        return Node.addParameter(self, name, type, description, min,  max, step, values, default, value)
                      
