# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


from System.Data.Node import Node
import copy
from Library.Util import lists
from Library.plot.plotmap import PlotMapHandler, PlotMapConfig

class Plot2DNode(Node):
    """
    A MultiPlotNode sends it data from the input port to the GUI for plotting, and
    puts command data suitable for an EditNode on its (only) output port."""

    def __init__(self, name):
        """Params:
            - name : String - the name of this node object
        """
        Node.__init__(self, name)
        self.setCanAddInputPorts(True)
        self.setCanAddOutputPorts(True)
        self._plot_map_handler = PlotMapHandler(self)
        self._multi_input = self.add_multi_input_port("plot input")

        # plot node settings
        self._widgets = {}
        self._horisontal_widgets = 0
        self._vertical_widgets = 0
        
        self._plot_maps = {}
        self._go_browser_maps = []

        # set one by one widgets as default
        self.update_widgets(1, 1)

    def execute(self):
        """Executes this node"""
        Node.execute(self)
        # a plot node currently does no processing of data

    def get_data_size(self, port_name, is_column):
        """Retrieves the number of columns/rows of the data.

        Parameters:
        port_name -- the name of the port the data is on
        is_column -- True if data is supposed to be retrieved from a column in data,
                     False if data is supposed to be retrieved from a row in data.
        """
        port = self.getInputPortByName(port_name)
        if not port:
            raise Exception, '[Plot2DNode]: no port called %s' % port_name

        data = port.getData()
        return data.get_data_size(is_column)
            
    def get_data(self, port_name, is_column, index):
        """Retrieves data from input port data (which is a matrix).

        Parameters:
        port_name -- the name of the port to get the data from
        is_column -- True if data is supposed to be retrieved from a column in data,
                     False if data is supposed to be retrieved from a row in data.
        index     -- column/row index in data
        """

        port = self.getInputPortByName(port_name)
        if not port:                       
            raise Exception, '[Plot2DNode]: no port called %s' % port_name
        
        data = port.getData()
        if data:
            return data.get_sub_data(is_column, index)
        else:
            return []

    def get_all_data_by_name(self, port_name):
        port = self.getInputPortByName(port_name)
        if not port:
            return []

        data = port.getData()
        if data:
            return port.getData().getData()
        else:
            return []

    def toXML(self):
        self.nodeRepName = "PlotNodes.Plot2DNode"
        xml_list = Node.toXML(self)
        xml_swap = xml_list[:-1]
        xml_swap.extend(self.to_xml())
        xml_swap.append(xml_list[-1])
        return xml_swap

    def to_xml(self):
        """This is method is called from node whenever the user desides to save
        the module diagram.

        Returns:
            - list : contains xml tags and values.
        """
        xml_list = []
        xml_list.append("\t<plotnode_config>")
        xml_list.append("\t\t<horisontal_widgets>%s</horisontal_widgets>" % self._horisontal_widgets)
        xml_list.append("\t\t<vertical_widgets>%s</vertical_widgets>" % self._vertical_widgets)

        for widget in self._widgets.values():
            xml_list.extend(widget.to_xml())

        handler = self.get_plot_map_handler()
        xml_list.append("\t\t<plot_map_config>")
        xml_list.append("\t\t\t<auto_map_enabled>%d</auto_map_enabled>" % handler.get_auto_map_enabled())
        xml_list.append("\t\t\t<enabled>%d</enabled>" % handler.get_enabled())

        for index in self._plot_maps:
            xml_list.extend(self._get_plot_map_config(index).to_xml())

        for settings in self.get_go_browser_map_settings():
            xml_list.append("\t\t\t<go_browser_map>")
            for key, value in settings.items():
                xml_list.append("\t\t\t\t<%s>%s</%s>" % (key, value, key))
            xml_list.append("\t\t\t</go_browser_map>")

        xml_list.append("\t\t</plot_map_config>")    
        xml_list.append("\t</plotnode_config>")
        return xml_list

    def custom_plot_to_xml(self):
        xml_list = self.to_xml()
        xml_swap = xml_list[:-1]
        extend_list = []
        
        ip_dict = self.get_input_port_descriptions()
        if ip_dict:
            extend_list.append("\t\t<input_ports_used>")
            for key in ip_dict:
                extend_list.append("\t\t\t<ip>")
                extend_list.append("\t\t\t\t<name>" + key + "</name>")
                extend_list.append("\t\t\t\t<type>" + ip_dict[key][0] + "</type>")
                extend_list.append("\t\t\t\t<description>" + ip_dict[key][1] + "</description>")                
                extend_list.append("\t\t\t</ip>")
            extend_list.append("\t\t</input_ports_used>")
            
        op_dict = self.get_output_port_descriptions()
        if op_dict:
            extend_list.append("\t\t<output_ports_used>")
            for key in op_dict:
                extend_list.append("\t\t\t<op>")
                extend_list.append("\t\t\t\t<name>" + key + "</name>")
                extend_list.append("\t\t\t\t<type>" + op_dict[key][0] + "</type>")
                extend_list.append("\t\t\t\t<description>" + op_dict[key][1] + "</description>")
                extend_list.append("\t\t\t</op>")
            extend_list.append("\t\t</output_ports_used>")
        
        xml_swap.extend(extend_list)
        xml_swap.append(xml_list[-1])
        
        return xml_swap
        
    def remove_input_port(self, port_name):
        self.remove_port(self.getInputPortByName(port_name))

    def get_input_port_descriptions(self):
        """Returns the input ports represented in a dictionary.

        example of a possible dictionary returned:

        ports_in_dictionary = {'portname1':('matrix', 'matrix data'),
                               'portname2':('scalar', 'color of marks')} 
        """
        
        ports = self.getInputPorts()
        ports_in_dictionary = {}

        for port in ports:
            ports_in_dictionary[port.getVariableName()] = (port.getDataType(), port.getDescription())

        return ports_in_dictionary

    def get_output_port_descriptions(self):
        """Returns the output ports represented in a dictionary.

        example of a possible dictionary returned:

        ports_in_dictionary = {'portname1':('matrix', 'matrix data'),
                               'portname2':('scalar', 'color of marks')} 
        """
        
        ports = self.getOutputPorts()
        ports_in_dictionary = {}

        for port in ports:
            ports_in_dictionary[port.getVariableName()] = (port.getDataType(), port.getDescription())

        return ports_in_dictionary

    def get_all_emulated_input_ports(self):
        """Gather all emulated ports from multiinputports."""
        return self._multi_input.get_emulated_input_ports()

    def get_plot_map_handler(self):
        return self._plot_map_handler

    def set_widget_settings(self, widget_x_coord, widget_y_coord,
                             widget_settings):
        """Saves settings concerning a specific widgets:
        
        Params:
            - widget_x_coord  : int  - the x coordinate of the widget to save
                                       settings in.
            - widget_y_coord  : int  - the y coordinate of the widget to save
                                       settings in.
            - widget_settings : dict - see class docstring.
        """
        self.get_widget_config(widget_x_coord,
                                widget_y_coord).set_settings(widget_settings)

    def get_widget_settings(self, widget_x_coord, widget_y_coord):
        """Retrieves settings concerning a specific widget.

        Params:
            - widget_x_coord : int - the x coordinate of the widget from which
                                     the settings should be retrieved.
            - widget_y_coord : int - the y coordinate of the widget from which
                                     the settings should be retrieved.

        Returns:
            - dict : the settings for the specified widget
        """
        return self.get_widget_config(widget_x_coord,
                                       widget_y_coord).get_settings()
    
    def update_plots(self, widget_x_coord, widget_y_coord, number_of_plots):
        """Creates structure for the number of plots wished in the given widget:
        
        Params:
            - widget_x_coord  : int - the x coordinate of the widget where the
                                      number of plots should be updated.
            - widget_y_coord  : int - the y coordinate of the widget where the
                                      number of plots should be updated.
            - number_of_plots : int - the number of plots within the specified
                                      widget.

        IMPORTANT!!!

        This method must be called every time the gui wishes to add/remove a
        plot to/from a specific widget so that the class can ready the data
        structures needed to save future settings.
        """
        self.get_widget_config(widget_x_coord,
                                widget_y_coord).update_plots(number_of_plots)

    def get_number_of_plots(self, widget_x_coord, widget_y_coord):
        """
        Params:
            - widget_x_coord : int - the x coordinate of the widget.
            - widget_y_coord : int - the y coordinate of the widget.

        Returns:
            - int : the number of plots in the given widget.
        """
        
        return self.get_widget_config(widget_x_coord,
                                       widget_y_coord).get_number_of_plots()

    def update_widgets(self, horisontal_widgets, vertical_widgets):
        """Creates structure for the number of widgets wished in the plotnode:
        
        Params:
            - horisontal_widgets : int - the number of horisontal widgets.
            - vertical_widgets   : int - the number of vertical widgets.

        IMPORTANT!!!

        This method must be called every time the gui wishes to add/remove one
        or more widgets so that the class can ready the data structures needed
        to save future settings.
        """
        self._horisontal_widgets = horisontal_widgets
        self._vertical_widgets = vertical_widgets
        
        temp_dict = {}
        for x in range(self._horisontal_widgets):
            for y in range(self._vertical_widgets):
                if not self._widgets.has_key((x, y)):
                    temp_dict[(x, y)] = _WidgetConfig(x, y)
                else:
                    temp_dict[(x, y)] = self._widgets[(x, y)]
        self._widgets = temp_dict.copy()

    def set_marker_settings(self, widget_x_coord, widget_y_coord, index,
                             marker_settings):
        """Saves settings concerning markers and lines for a specific plot within the given widget:
        
        Params:
            - widget_x_coord     : int  - the x coordinate of the widget where the
                                          plot is located.
            - widget_y_coord     : int  - the y coordinate of the widget where the
                                          plot is located.
            - index              : int  - the index of the plot to save settings in.
            - marker_settings : dict - see class docstring.
        """
        plot_config = self.get_plot_config(widget_x_coord, widget_y_coord,
                                            index)
        plot_config.set_marker_settings(marker_settings)

    def get_marker_settings(self, widget_x_coord, widget_y_coord, index):
        plot_config = self.get_plot_config(widget_x_coord, widget_y_coord,
                                            index)
        return plot_config.get_marker_settings()

    def get_horisontal_widgets(self):
        """
        Returns:
            - int : the number of widgets in x-direction (horisontal-direction).
        """
        return self._horisontal_widgets    

    def get_vertical_widgets(self):
        """
        Returns:
            - int : the number of widgets in y-direction (vertical-direction).
        """
        return self._vertical_widgets

    def update_plot_maps(self, number_of_plot_maps):
        """Creates structure for the number of plot maps:
                
        Params:
            - number_of_plot_maps : int
        """
        temp_dict = {}
        for index in range(number_of_plot_maps):
            if not self._plot_maps.has_key(index):
                temp_dict[index] = PlotMapConfig(index)
            else:
                temp_dict[index] = self._plot_maps[index]
        self._plot_maps = temp_dict.copy()
        
    def set_plot_map_settings(self, index, settings):
        """Saves settings concerning a single plot map:
        
        Params:
            - index    : index of the plot map
            - settings : dict with plot map settings
        """
        plot_map_config = self._get_plot_map_config(index)
        plot_map_config.set_settings(settings)

    def get_plot_map_settings(self):
        """Retrieves list of dictionaries with settings for all plot maps."""
        return map(PlotMapConfig.get_settings, self._plot_maps.values())

    def get_widget_config(self, *key):
        """Get config for widget in zero-indexed position (x, y)."""
        return self._widgets[key]

    def get_plot_config(self, widget_x_coord, widget_y_coord, index):
        _widget_config = self.get_widget_config(widget_x_coord, widget_y_coord)
        return _widget_config.get_plot_config(index)

    def _get_plot_map_config(self, index):
        return self._plot_maps[index]

    def get_go_browser_map_settings(self):
        """Get list with settings/dict for each mapping to a GO browser."""
        return self._go_browser_maps

    def set_go_browser_map_settings(self, maps):
        """Set list with settings/dict for each mapping to a GO browser."""
        self._go_browser_maps = maps
    

class _WidgetConfig:
    def __init__(self, x_coord, y_coord):
        self._x_coord = x_coord
        self._y_coord = y_coord

        self._settings = {}
        self._plots = {}
        self._number_of_plots = 0

    def set_settings(self, widget_settings):
        self._settings = widget_settings.copy()

    def get_settings(self):
        return self._settings

    def update_plots(self, number_of_plots):
        """Updates self._plots dictionary and self._number_of_plots"""
        temp_dict = {}
        for index in range(number_of_plots):
            if not self._plots.has_key(index):
                temp_dict[index] = PlotConfig(index)
            else:
                temp_dict[index] = self._plots[index]
        self._plots = temp_dict.copy()

        self._number_of_plots = len(self._plots)

    def get_plot_config(self, index):
        if self._plots.has_key(index):
            return self._plots[index]
        return None

    def to_xml(self):
        self.__update_standard_parameters()

        xml_list = []
        xml_list.append("\t\t<widget_config>")

        for setting, value in self._settings.items():
            xml_list.append("\t\t\t<%s>%s</%s>" % (str(setting), str(value),
                                                   str(setting)))

        for plot in self._plots.values():
            xml_list.extend(plot.to_xml())

        xml_list.append("\t\t</widget_config>")
        return xml_list

    def __update_standard_parameters(self):
        self._settings["x_coord"] = str(self._x_coord)
        self._settings["y_coord"] = str(self._y_coord)

    def get_number_of_plots(self):
        return self._number_of_plots
    
    def get_widget_x_coord(self):
        return self._x_coord

    def get_widget_y_coord(self):
        return self._y_coord


class PlotConfig:
    def __init__(self, index):
        self._settings = {}
        self._marker_settings = {}
        self._index = index

    def set_settings(self, plot_settings):
        self._settings = plot_settings.copy()

    def get_settings(self):
        return self._settings

    def set_setting(self, key, value):
        self._settings[key] = value

    def get_setting(self, key):
        return self._settings[key]

    def set_marker_setting(self, key, value):
        self._marker_settings[key] = value

    def get_marker_setting(self, key):
        return self._marker_settings[key]

    def set_marker_settings(self, marker_settings):
        self._marker_settings = marker_settings.copy()

    def get_marker_settings(self):
        return self._marker_settings

    def to_xml(self):
        if self._settings:
            self.__update_standard_parameters()
        
            xml_list = []
            xml_list.append("\t\t\t<plot_config>")

            settings_keys = self._settings.keys()
            settings_keys.sort()

            for setting in settings_keys:
                xml_list.append("\t\t\t\t<" + str(setting) + ">" +
                                str(self._settings[setting]) +
                                "</" + str(setting) + ">")

            if self._marker_settings:
                xml_list.append("\t\t\t\t<plot_marker_settings>")
                marker_setting_keys = self._marker_settings.keys()
                marker_setting_keys.sort()
                for setting in marker_setting_keys:
                    xml_list.append("\t\t\t\t\t<" + str(setting) + ">" +
                                    str(self._marker_settings[setting]) +
                                    "</" + str(setting) + ">")
                xml_list.append("\t\t\t\t</plot_marker_settings>")
            xml_list.append("\t\t\t</plot_config>")

            return xml_list
        else:
            return []

    def __update_standard_parameters(self):
        self._settings["index"] = str(self._index)


    def get_index(self):
        return self._index
