# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import cPickle
import os
import re
import sys
import time
import types
import thread

from dprint import dprint, flags
from Library.Util.InterruptibleCommand import InterruptibleCommand
from System.Plugin import Plugin, PluginInterrupted, get_python_command
from System.Data import Data
from System.Exceptions import ZPluginError

dflag  = flags.LIBRARY.PLUGINS.PYTHONPLUGIN
dstart = "[LIBRARY.PLUGINS.PYTHONPLUGIN]:"

class PythonPlugin(Plugin):

    def __init__(self):
        """Constructor which creates a PythonPlugin-object"""
        Plugin.__init__(self)
        self._start_time = 0
        
        # holds what was printed to stdout/err during last run
        self._last_stdout = "Not run"
        self._last_stderr = "Not run"

    def clean(self):
        Plugin.clean(self)

    def _clean_dprint_and_raise(self, function_name, current_action):
        self.clean()
        dprint(dflag, dstart,
               "Interrupted before executing (function %s)" % function_name)
        dprint(dflag, dstart, "Total time spent in plugin:",
               time.time() - self._start_time)
        raise PluginInterrupted, "PythonPlugin interrupted - " + current_action

    def get_stdout_from_last_run(self):
        """Returns what the python plugin wrote to stdout during last run."""
        return self._last_stdout

    def get_stderr_from_last_run(self):
        """Returns what the python plugin wrote to stderr during last run."""
        return self._last_stderr

    def execute(self, functionName, functionFileName, nodeInputs,
                nodeOutputs, nodeParams, interruptCallback):
        """Execute python script and return result.

        Return:
            - List[Data] - a List with Data-objects which are the results from the python-script.
                  
        General usage: When making a python-script for use as a
        function within SciCraft, you can do pretty much anything you
        want. However, if your function returns more than one
        variable, you have to return them as elements in a
        dictionary. If your function returns one variable,
        PythonPlugin will assume it is supposed to have the same name
        as the output-port. In case there are more than one
        output-port, and the result from the function is not a
        dictionary, the plugin will raise an error, claiming the given
        function and xml-declaration are inconsistent.
        
        Also note that objects outside the function-shell cannot be
        referenced from whithin the function.  This means that
        e.g. imports need to be done within the function.
        """

        variables = []
        for inp in nodeInputs:
            variables.append(inp.getData().getData())
        
        for par in nodeParams:
            variables.append(par.getValue())

        self._start_time = time.time()
        dprint(dflag, dstart,
               "Started (function %s - thread %s)" % (functionName,
                                                      thread.get_ident()))

        if interruptCallback():
            self._clean_dprint_and_raise(functionName, "before retrieving data")

        func = functionName

        if interruptCallback():
            self._clean_dprint_and_raise(functionName, "before executing")

        try:
            cmd = get_python_command()
            dprint(dflag, dstart, "Starting new python instance")
            cmd.run_interruptible()
            dprint(dflag, dstart, "Creating pickle-string")
            startToken = '### START OF PICKLEDUMP ###'
            endToken   = '###  END OF PICKLEDUMP  ###'
            
            callString = _compose_call_string(variables, func,
                                              functionFileName, startToken,
                                              endToken)
            cmd.write_stdin(callString)
            
            dprint(dflag, dstart,
                   "Closing input to python instance - triggering processing")
            cmd.close_stdin()
            dprint(dflag, dstart, "Reading result")

            stdout = cmd.read_stdout()
            stderr = cmd.read_stderr()
            while not stdout and not stderr:
                time.sleep(0.1)
                if interruptCallback():
                    cmd.interrupt()
                    self._clean_dprint_and_raise(functionName,
                                                 "while executing")
                stderr = cmd.read_stderr()
                stdout = cmd.read_stdout()
                
            # if user has made the plugin we want the user to be able
            # to see his own error messages
            if stderr:
                print stderr

            # keep stderr output for later use
            self._last_stderr = stderr
            # parse with regexp to find tokens and the cPickle dumpstring
            dprint(dflag, dstart, "Parsing result")
            start_match = re.search(startToken, stdout)
            end_match   = re.search(endToken,   stdout)
            if start_match and end_match:
                res = cPickle.loads(stdout[start_match.end():end_match.start()])
            else:
                raise Exception, 'No values returned.\n\nError messages from script:\n' + stderr
            # if no output to stdout is generated in plugin, start_match will be first
            if start_match.start() != 0: # if user has made the plugin we want
                # the user to be able to see his own printouts
                self._last_stdout = stdout[:start_match.start() - 1]
                print self._last_stdout # TODO remove this line, not needed when
                                        #      we can retrieve stdout in gui
                dprint(dflag, dstart,
                       "stdout follows:\n" + self._last_stdout)
        except PluginInterrupted, e:
            raise e
        except Exception, e:
            print "exception in pythonplugin", e
            dprint(dflag, dstart, "Total time spent in plugin:",
                   time.time() - self._start_time)
            raise ZPluginError.ZPluginError, "Error when running python-script:\n%s" % str(e)
             
        data = []
        if not isinstance(res, dict):
            numOfOutputs = len(nodeOutputs)            
            if numOfOutputs > 1:
                print "Exception in pythonplugin. Method declaration inconsistent with actual method behavior. No dictionary found!"
                dprint(dflag, dstart, "Total time spent in plugin:",
                       time.time() - self._start_time)
                raise ZPluginError, 'Inconsistency between returned data from function and xml-declaration.'
            elif numOfOutputs == 1:
                data.append(Data.Data(nodeOutputs[0].getName(), res))
                return data
        else:
            for key in res.keys():
                data.append(Data.Data(key, res[key]))
        dprint(dflag, dstart, "Total time spent in plugin:",
               time.time() - self._start_time)
        dprint(dflag, dstart, "Returning data", map(lambda d : d.getData(),
                                                    data))
        return data

def _compose_call_string(variables, func, functionFileName, startToken,
                         endToken):
    callString  = "import cPickle\n"
    callString += "import sys\n"
    if variables:
        callString += "input_ = cPickle.loads(%s)\n" % repr(cPickle.dumps(variables))
    # make sure filename doesn't make string unicode
    callString += "execfile(r'%s')\n" % str(functionFileName)
    if variables:
        callString += "res = locals()['%s'](*input_)\n" % func
    else:
        callString += "res = locals()['%s']()\n" % func
    # print tokens to allow plugin to print
    callString += "sys.stdout.write('%s')\n" % startToken
    callString += "sys.stdout.write(cPickle.dumps(res))\n"
    callString += "sys.stdout.write('%s')\n" %endToken
    dprint(dflag, dstart, "Dumping data to python instance")
    return callString
