# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


import cPickle
import Numeric
import os
import re
import sys
import time
import types
import tempfile
import mutex

from dprint import dprint, flags
from Library.Util.InterruptibleCommand import InterruptibleCommand
from Library.Util.file import StreamHolder
from Library.Util.rtester import RTester
from Library.Util.Config import Config
from System.Plugin import Plugin, PluginInterrupted, get_python_command
from System import Tools
from System.Data.Data import Data
from System.Data.Parameter import Parameter
from System.Exceptions.ZPluginError import ZPluginError
from System.FileHandlers.Input.RFileReader import RFileReader

function_mutex = mutex.mutex()

    
class RPlugin(Plugin):

    def __init__(self):
        """Constructor which creates an RPlugin-object"""
        Plugin.__init__(self)
        # normal level debug messages
        self.DEBUG   = flags.LIBRARY.PLUGINS.RPLUGIN
        # verbose debug messages
        self.DEBUG_V = 0
        # string to put in front of debug messages
        self.DMESG   = '[LIBRARY.PLUGINS.RPLUGIN]:'
        # (advanced) user messages
        self.DEVINFO = False

        RTester.r_install_test()
        import rpy
        from Library.Plugins.RPluginFiles.RObjectHandler import RObjectHandler

        self.robject_handler = RObjectHandler()
        self.input_path = None
        self.output_path = None

        # holds stdout from last run
        self._last_stdout = "Not run"
        self._last_stderr = "Not run"

    def execute(self, function_name, function_filename, inputPorts,
                outputPorts, parameters, interruptCallback):
        """Executes R-function in an external python process.

        We need to execute this plugin in another process to be able
        to stop it.  For this to be possible we need to transfer the
        parameters over to the other process (which was chosen to be
        done through pickle).

        """
        function_filename = str(function_filename) # need this to avoid unicode...
        outputPortNames = [port.getName() for port in outputPorts]

        return self._execute_function(function_name, function_filename,
                                          inputPorts, outputPortNames, parameters,
                                          interruptCallback)
        
    def _execute_function(self, function_name, function_filename, variables,
                          output_port_names, parameters, interrupt_callback):

        RTester.r_install_test()
        import rpy

        # Need a mutex when running as debug since rpy will then
        # run use the same environment for both functions.
        if Config.debugR: 
            while not function_mutex.testandset():
                time.sleep(0.2)
        try:
            data_list = []
            input_data = variables[:]
            input_data.extend(parameters)
            if input_data:
                (fd1, self.input_path) = tempfile.mkstemp()
                os.close(fd1)
                if os.name == "nt": self.input_path = self.input_path.replace("\\", "/")
            if output_port_names:
                (fd2, self.output_path) = tempfile.mkstemp()
                os.close(fd2)
                if os.name == "nt": self.output_path = self.output_path.replace("\\", "/")

            # Get unique name for the output data
            output_name = self.robject_handler.getUniqueVarname()

            # Save input data to file, returnes dict with name as key and unique names as value
            saved_names = self._save_r_input_data(input_data)
            # Generate call string for rpy 
            call_string = self._generate_call_string(saved_names, input_data,
                                                     output_name, function_name)
            # Test if the function can be loaded in R
            self._test_loading_of_function(function_name, function_filename)
            # Generate string used for external python process

            if not Config.debugR:
                # Run the function in an external process
                self._run_in_process(output_name, input_data, call_string, function_filename,
                                     interrupt_callback)
            else:
                # Use this to be able to get exceptions from rpy
                self._execute_process(call_string, self.input_path, self.output_path,
                                      output_name, function_filename)

            data_list = []
            if self.output_path:
                reader = RFileReader()
                try:
                    data_list = reader.read_data(self.output_path)
                except:
                    raise ZPluginError, "These was no output from the function:" + \
                          str(function_filename)

        finally:
            self.clean()
            function_mutex.unlock()

        return data_list

    def _run_in_process(self, output_name, input_data, call_string, function_filename,
                        interrupt_callback):
        
        
        start_time = time.time()
        cmd = get_python_command()
        cmd.run_interruptible() # start the command
        dprint(self.DEBUG, self.DMESG, "External python proc initiated")        

        process_string = self._generate_process_string(call_string, output_name,
                                                       function_filename)
        
        # Write to the process
        cmd.write_stdin(process_string)
        cmd.close_stdin() # This will execute the process
        stdout = cmd.read_stdout()
        stderr = cmd.read_stderr()
        # wait for the process to finish
        while cmd.is_running():
            time.sleep(0.1)
            if interrupt_callback():
                cmd.interrupt()
                self.clean()
                raise PluginInterrupted, "RPlugin interrupted during execution"
            
        self._last_stdout = cmd.read_stdout()
        dprint(self.DEBUG, self.DMESG, "stdout follows:\n" + self._last_stdout)
        stderr = cmd.read_stderr()
        self._last_stderr = stderr
        if stderr:
            dprint(self.DEVINFO, "stderr from external python process:\n",
                   stderr)
        dprint(self.DEBUG, self.DMESG, "external python proc finished. " + \
               "Time spent: %s" % (time.time() - start_time))
                                  
    def _save_r_input_data(self, input_data):
        RTester.r_install_test()
        import rpy
        saved_names = {}
        if not input_data:
            return saved_names

        for var in input_data:
            if isinstance(var, Parameter):
                data = Data(var.getName(), var.getValue())
            else:
                data = var.getData() #lalal
            saved_names[data.getName()] = self.robject_handler.setRVar(data, use_unique = True)

        # now we want to save these variables to file
        rpy.r("save(list=c('%s'), file = '%s')" % ("','".join(saved_names.values()),
                                                   self.input_path))
        # Lets clean up these variables
        rpy.r('rm(' + ",".join(saved_names.values()) + ')')
        
        return saved_names 


    def _generate_call_string(self, saved_names, input_data, output_name, function_name):

        input_data_names = []
        for data in input_data:
            if isinstance(data, Parameter):
                input_data_names.append(data.getName())
            else:
                input_data_names.append(data.getData().getName())

        names = []
        for name in input_data_names:
            names.append(saved_names[name])            
        
        call_string = output_name + "<-" + function_name + '(' + ",".join(names) + ')'
        return  call_string


    def _generate_process_string(self, call_string,
                                 output_name, functionFileName):
        
        execute_parameters = (call_string, self.input_path, self.output_path, output_name,
                              functionFileName)
        process_string = """
import sys
import os
import qt

sys.path = %s
from Library.Plugins.RPlugin import RPlugin
from System.Plugin.PluginRepository import PluginRepository
execute_parameters = %s
rplugin = RPlugin()
rplugin._execute_process(*execute_parameters)
"""% (sys.path, execute_parameters)

        return process_string    

    def _execute_process(self, call_string, input_path, output_path, output_name,
                         function_filename):
        RTester.r_install_test()
        import rpy
        old_path = Config.scicrafthome
        try:    
            # capture stdout
            stdout_stream = StreamHolder()
            original_stdout = sys.stdout # to allow tests to capture stdout too...
            sys.stdout = stdout_stream
            # to be able to capture stderr from rpy in debugR mode, we need to 
            # use the sink function (see Wiki)
            if Config.debugR:
                # reroutes stderr to file errorfile
                fd, errorfile = tempfile.mkstemp(text=True)
                os.close(fd)
                errorfile = errorfile.replace("\\", "/")
                error_id = rpy.r("errorfile = file('%s', 'w')" % errorfile)                
                rpy.r("sink(file=errorfile, type='message')")

            new_path = os.path.dirname(str(function_filename)).replace(os.sep, '/')
            old_path = os.getcwd().replace('\\', '/')
            # change working directory
            self._change_wd_on_r(new_path)
            try:
                # Load input data if we have any
                if input_path:
                    rpy.r("load('" + input_path + "')")
                # source the function file
                rpy.r("source('%s')" % os.path.basename(function_filename))
                # Execute the call string
                rpy.r(call_string)
                # Magic to be able to save the variable names
                names = Tools.assertList(rpy.r("names(%s)" % output_name))
                for name in names:
                    rpy.r('%s<-%s$%s' %(name, output_name, name))
                if output_path:
                    rpy.r("save(list=names(%s), file='%s')" %(output_name, output_path))
                # read stdout from our StreamHolder
                self._last_stdout = stdout_stream.read()
            except rpy.RException, e:
                
                raise ZPluginError, "Got the following error message when executing: \n" + \
                      "Function name: " + os.path.basename(function_filename) + "\n" + \
                      "Error: " + str(e)
            except Exception, e:
                raise ZPluginError, "Got the following python error: \n" + str(e)
        finally:
            sys.stdout = original_stdout # not sys.__stdout__ because then tests can't capture stdout..
            sys.stdout.write(stdout_stream.read()) # has always been printed

            if Config.debugR:
                # reset stderr in rpy
                rpy.r('sink(file=NULL, type="message")')                
                rpy.r('close(errorfile)')
                fd = open(errorfile, "r")
                self._last_stderr = fd.read()
                # stderr should still be written
                sys.stderr.write(self._last_stderr) 
                fd.close()
                os.unlink(errorfile)
            try: 
                self._change_wd_on_r(old_path)
            except Exception, e:
                print e
                self._change_wd_on_r(Config.scicrafthome)
                
            
    def _test_loading_of_function(self, function_name, function_filename):
        RTester.r_install_test()
        import rpy
        try: # test if the function can be loaded
            new_path = os.path.dirname(str(function_filename)).replace(os.sep, '/')
            old_path = os.getcwd().replace('\\', '/')
            self._change_wd_on_r(new_path)
            rpy.r("source('%s')" % os.path.basename(function_filename))
            # remove the function since it is not needed
            rpy.r('rm(' + function_name + ')')
            self._change_wd_on_r(old_path)
        except rpy.RException, e:
            raise ZPluginError, \
                  "The following function file contains the following errors:\n" + \
                  "Function file: " + function_filename + "\n" + \
                  "Error reported: " + str(e)

    def _change_wd_on_r(self, path):
        # No need to catch exceptions here since they will be cought where
        # this function is used
        RTester.r_install_test()
        import rpy
        rpy.r("setwd('%s')" % str(path))
        
    
    def clean(self): # Overrides super's empty function
        Plugin.clean(self)
        dprint(self.DEBUG, self.DMESG, "cleaning plugin..")
        if self.input_path and os.path.exists(self.input_path):
            try:
                os.remove(self.input_path)
            except:
                pass
        if self.output_path and os.path.exists(self.output_path):
            try:
                os.remove(self.output_path)
            except:
                pass

        #rpy.r('rm(list=ls())') # Removing all R-side variables

    def get_stdout_from_last_run(self):
        return self._last_stdout

    def get_stderr_from_last_run(self):
        return self._last_stderr
