#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from sys import maxint
from string import strip
import Numeric
import math

# These functions checks if a number is odd or even
odd = lambda x:bool(x % 2)
even = lambda x:not bool(x % 2)

# Checks if something is a string or not. This is the recommended way.
is_string = lambda x:isinstance(x, (str, unicode))

def create_indexed_list(count, prefix=""):
    """Create count indexed string values, optionally with prefix.

    create_indexed_list(3) => ["1", "2", "3"]
    create_indexed_list(2, "Object ") => ["Object 1", "Object 2"]
    """
    return [prefix + str(num) for num in xrange(1, count + 1)]

def get_dim(something=[], check_element_nr=1):
    """Returns the dimensions of the given data.
       If check_element_nr is 1 (or True), it's the "mathematical" version.
       If check_element_nr is 0 (or False), it's the functional "repr" version.
    """
    for dim in xrange(maxint):
        try:
            # Try to go a step deeper every round
            if is_string(something):
                break
            something = something[check_element_nr]
        except Exception:
            break
    return dim

def get_repr_dim(list=[]):
    """Returns the dimensions of a list- or array-structure.
       The functional "repr" version.
    """
    return get_dim(list, 0)

def get_cols(list=[]):
    """Returns the number of columns in a matrix.
        Columns are one step deeper than rows.
    """
    if get_repr_dim(list) == 1:
        return len(list)
    else:
        try:
            return len(list[0]) 
        except IndexError:
            return 0
        except TypeError:
            return -1
        except KeyError:
            return -1

def get_rows(list=[]):
    """Returns the number of rows in a matrix.
        Rows are one step shallower than columns.
    """
    if get_repr_dim(list) == 1:
        return 1
    else:
        try:
            return len(list)
        except TypeError:
            return -1

def describe_variable(variable):
    """Describe a scalar, string, or list or multidimensional matrix.

    describe_variable(5) => "scalar"
    describe_variable("hi") => "String"
    describe_variable([1,2,3]) => "1x3"
    describe_variable(["hi", "hello"]) => "2xString"
    """

    sizes = []

    try:
        while True:
            if is_string(variable):
                sizes.append("String")
                break
            sizes.append(len(variable))
            variable = variable[0]
    except TypeError:
        if len(sizes) == 1:
            sizes.insert(0, 1)
    except IndexError:
        sizes.insert(0, 0)

    if sizes:
        return "x".join(map(str, sizes))
    else:
        return "Scalar"

def filter_list(list_, remove_list):
    """Removes elements in remove_list from list and returns
        this filtered list.
    """
    return [item for item in list_ if item not in remove_list]


def remove_duplicates(list_):
    for x in list_:
        while list_.count(x) != 1:
            list_.remove(x) #list.remove() only removes one of the duplicated items
    

def merge_lists(list1, list2):
    """Merges two lists and removes duplicates"""
    list1.extend(list2)
    remove_duplicates(list1)

def count_dulicates(list1, list2):
    """Counts the number of ojects that can be found in both lists"""
    count = 0
    for x in list1:
        count += list2.count(x)

    return count
        
def to_string(list_):
    return str(list_).replace('\'','')

def try_number_coerce(text):
    """Try to change text into number, or return text."""
    try:
        return int(text)
    except:
        try:
            return float(text)
        except:
            return text

def try_strip(text_or_number):
    """Strip whitespace from text, or return unmodified."""
    try:
        return strip(text_or_number)
    except:
        return text_or_number

def flatten(nested_list):
    """Creates a list of atoms from a structured/nested list."""
    result = []
    for element in nested_list:
        try:
            result.extend(flatten(element))
        except TypeError:
            try:
                result.extend(element)
            except:
                result.append(element)

    return result

def to_tuple(data):
    """Converts list and Numeric.array to tuple"""
    if type(data) == Numeric.ArrayType: 
        return to_tuple(data.tolist())
    elif not isinstance(data, list):
        return data     
    for k in range(len(data)): 
        if isinstance(data[k], list): 
            data[k] = to_tuple(data[k])

    return tuple(data)
    
def count_interval(list_, bottom, top):
    """Counts the number of elements with value between bottom and top.

    This function will alther the list.
    """

    result = 0
    for element in list_[:]:
        if element >= bottom and element <= top:            
            result += 1
            if len(list_) > 0 and type(list_) == Numeric.ArrayType:                
                temp = list_.tolist()
                temp.remove(element)
                list_ = Numeric.array(temp)
            else:
                list_.remove(element)
    return result


class TokenFile(list):

    def __init__(self, filename, split_on=" \t\n", col_split = None,
                 parse_numbers=False):
        self.filename = filename
        self.string = open(filename).read()
        reslist = [] 
        
        if col_split:
            for key in col_split:
                self.string = self.string.replace(key, "har_DU_husket_KJOTKAKENE?")
            list = self.string.split("har_DU_husket_KJOTKAKENE?")
            [list.remove(f) for f in list[:] if not f]
            for element in list:
                for key in split_on:
                    element = element.replace(key, "har_DU_husket_KJOTKAKENE?")
                element = element.split("har_DU_husket_KJOTKAKENE?")
                [element.remove(f) for f in element[:] if not f]
                element = map(strip, element)
                if parse_numbers:
                    element = map(try_number_coerce, element)
                reslist.append(element)
        else:
            for key in split_on:
                self.string = self.string.replace(key, "har_DU_husket_KJOTKAKENE?")
            reslist = self.string.split("har_DU_husket_KJOTKAKENE?")
            if parse_numbers:
                reslist = map(try_number_coerce, reslist)

        # Remove all empty tokens
        reslist = filter(bool, reslist)
        reslist = map(try_strip, reslist)

        # I am a list, so this is possible
        self.extend(reslist)

class Separator:
    """
        Class to keep track of how we are to read
        text files. i.e. What separates rows, and 
        columbs
    """
    def __init__(self):
        self._def_row = [0, 0, 0, 0, 0, 0, 0, 1]
        self._def_col = [0, 0, 0, 0, 0, 0, 0, 0]

        self._symbols = ["ALARM", '\t', ',', ';', 'CuStOm', ' ', ':', '\n']
        
        self.custom_row = None
        self.custom_col = None
        self.row = self._def_row[:]
        self.col = self._def_col[:]

    def get_selection(self):
        return {"row": self.row, "col": self.col}

    def set_selection(self, all):
        self.row = all['lists'][0]
        self.col = all['lists'][1]
        if all['custom'][0]:
            self.custom_row = all['custom'][0]
        else:
            self.custom_row = None
        if all['custom'][1]:
            self.custom_col = all['custom'][1]
        else:
            self.custom_row = None
        self._def_row = self.row[:]
        self._def_col = self.col[:]
     

    def goto_default(self):
        self.row = self._def_row[:]
        self.col = self._def_col[:]
            
    def set_row_select(self, symb, remove = False):
        """Sets or removes a symbol for row """
        index = None
        for i,j in enumerate(self._symbols):
            if j == symb:
                index = i
        if remove:
            self.row[index] = 0
        else:
            self.row[index] = 1
        
    def set_col_select(self, symb, remove = False):
        """Sets or removes a symbol for col """
        index = None
        for i,j in enumerate(self._symbols):
            if j == symb:
                index = i
        if remove:
            self.col[index] = 0
        else:
            self.col[index] = 1

    
    def set_custom_row(self, text):
        """Sets the custom separator for row """
        if text:
            if text != self.custom_col:
                self.custom_row = text
            else:
                pass # Sett inn en exception.
    
    def set_custom_col(self, text):
        """Sets the custom separator for col """
        if text:
            if text != self.custom_row:
                self.custom_col = text
            else:
                pass # Sett inn en exception.

    def get_custom_row(self):
        return self.custom_row

    def get_custom_col(self):
        return self.custom_col
        
                
    def to_XML(self):
        """Writes itself to XML """
        alfa = ['a','b','c','d','e','f','g','h']
        cust_row = ""
        cust_col = ""
        s = ""
        s += '\t<separator>\n'
        s += '\t\t<row'
        for i,f in enumerate(self.row):
             s += ' ' + alfa[i] + ' = "' + str(f) + '"'
        if self.custom_row:
            cust_row = str(self.custom_row)
        s += ' custom = "' + cust_row + '" />\n' 
        s += "\t\t<col"
        for i,f in enumerate(self.col):
             s += ' ' + alfa[i] + ' = "' + str(f) +'"'
        if self.custom_col:
            cust_col = str(self.custom_col)
        s += ' custom = "' + cust_col + '" />\n'
        s += '\t</separator>\n'
        return s
    
    def get_symbols(self):
        """Returns a dictionary with keys 'row' and 'col'
            and lists of symbols as values """
        # filter(None, liste), tar bort falske/tomme elementer
        # map gir symbolset om self.row[i] er sann eller False
        row = filter(None, map(lambda i : self.row[i] and self._symbols[i], range(len(self._symbols))))
        if 'CuStOm' in row:
            row.remove('CuStOm')
            if self.custom_row:
                row.append(self.custom_row)
                
        col = filter(None, map(lambda i : self.col[i] and self._symbols[i], range(len(self._symbols))))
        if 'CuStOm' in col:
            col.remove('CuStOm')
            if self.custom_col:
                row.append(self.custom_col)
        
        if 'ALARM' in row or 'ALARM' in col:
            print 'ALARM'

        return {'row':row, 'col':col}

        
