# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import re 
import types
from Numeric import array
from Library.Util import lists

from System.FileHandlers.Input.FileReader import data_info


#TODO: Discuss use of varType and metadata. Should there be standard ways to describe them?
#      Some ways to solve the problem:
#         * Have a "created by" field on every data-object, so the information in metadata,
#           varType etc. can be set in the correct context. This would, however, not alone 
#           solve the problem. In addition, some of the below points should be considered:
#               1: 'standard' and 'specific'-fields in metadata and varType so each handler
#                  know where to find information it knows how to use (ie: ignore 'special'-
#                  field if "created by" does not match.
#               2: Distribute handling to a handler able to use the metadata (based on "created
#                  by").
#               3: ...

class Data:
    """This class is a representation of a data object.  It can be
    passed around in SciCraft, and usually contains metadata.
    """
   
    def __init__(self, name, data):
        """
        Params:
                - name : string - a string with the variablename for the data object
                - data : the actual data that the object will contain

        Constructor which will create a data object
        """
        self.nameFilter = re.compile("^[^\ \n\r]+.*$")
        #^(([a-zA-Z]+[0-9]*)+)(([\$\.\@]{1})[a-zA-Z]+[0-9]*)*$|[0-9]")
        
        if type(name) is not types.StringType:
            raise TypeError, "Attribute 'name' has to be a string. Recieved " + str(type(name))
        if self.nameFilter.match(name):
            self.name = name
        else:
            raise AttributeError, "[Data.__init__()#001] Invalid name. Must match the following regex: " + str(self.nameFilter.pattern) \
                + "\nBad name: " + str(name)
        self.data      = data 
        self.varType   = {'internal':None, 'special':None} 
        self.metadata  = {'internal':None, 'special':None}


    def setData(self, data):
        """
        Params:
                - data : Object - the data stored in this Data-object

        A method that sets the data in the data object
        """
        self.data = data  
        

    def setName(self, name):
        """
        Params:
                - name : String - a string with the name
            A method that sets the name on the data object.
        """
        if type(name) is not types.StringType:
            raise TypeError, "[Data.setName()#001] Name has to be a string. Received " + str(type(name))
        if not self.nameFilter.match(name):
            raise AttributeError, "[Data.setName()#002] Invalid name. Must match the following regex: " + str(self.nameFilter.pattern) \
                + "\nBad name: " + str(name)
        self.name = name
   
        
    def getName(self):
        """
        Return:
                String - the name of this data-object"""

        return self.name

        
    def getData(self):
        """
        Return:
               - Object  - Returns the actual data the Data-object contains.

        Returns the data this Data-object is holding"""

        return self.data


    def setType(self, typ):
        """
        Params:
            typ : Dictionary / String - The selected type of this data object.

        Sets the varType-variable of this data-object. If input is a string, it sets
        the 'internal'-field in the varType-dictionary. An input of type dictionary
        overwrites varType completely. To fulfil the requirements of this dictonary,
        the 'internal' and 'special' fields must always be in the dictionary.
        """

        if type(typ) is types.StringType:
            self.varType['internal'] = typ
        elif type(typ) is types.DictType:
            if 'internal' in typ.keys() and 'special' in typ.keys():
                self.varType = typ
            else:
                raise ValueError, "The parameter does not meet the demands set (dictionary with 'internal' and 'special' fields)."
        else:
            raise TypeError, "Invalid parameter. Expected string or dictionary, got %s." % str(type(typ))


    def setTypeInternal(self, typ):
        """
        Params:
            typ : String - The selected type of this data object.

        Sets the SciCraft-internal type declaration of this data object.
        """
        if type(typ) is not types.StringType:
            raise TypeError, "Internal type declaration must be a string. Got " + str(type(typ))
        self.varType['internal'] = typ


    def setTypeSpecial(self, typ):
        """
        Params:
            typ : Object - The selected type fo this data-object.

        Could be any python-oject. Declares specific type-information set by the creator of this data-object.
        This information could be vital for building the variable in some foreign environment, eg by plugins.
        """
        self.varType['special'] = typ
        
        
    def getType(self):
        """
        Return:
               - Dictionary - A list with the actual data the Data-object contains

        Returns the data this Data-object is holding"""
        return self.varType
        
   
    def getTypeInternal(self):
        """
        Return:
               - String - A list with the actual data the Data-object contains

        Returns the data this Data-object is holding"""
        return self.varType['internal']


    def getTypeSpecial(self):
        """
        Return:
               - Dictionary - A list with the actual data the Data-object contains

        Returns the data this Data-object is holding"""
        return self.varType['special']
        
        
    def setMetadataInternal(self,internalMeta):
        """
        Params:
           internalMeta : Object/module - The generic metadata of this data object.

        Sets the internal (SciCraft-related) meta of this data-object. Used for generic rebuilding of this data-object (mostly done
        when rebuilding it in its original environment by a plugin).
        """
        self.metadata['internal'] = internalMeta


    def setMetadataSpecial(self, specialMeta):
        """
        Params:
            specialMeta : Object/module - The metadata-dictionary of this data object.

       Sets the generic meta of this data-object. Used for generic rebuilding of this data-object (mostly done
        when rebuilding it in its original environment by a plugin).
        """
        self.metadata['special'] = specialMeta
        

    def getMetadata(self):
        """
        Return:
               - Dictionary - A list with the actual data the Data-object contains

        Returns the metadata this Data-object is holding"""
        return self.metadata


    def getMetadataSpecial(self):
        """
        Return:
               - Object - A python-object of some sort - holding the special metadata for this data-object.

        Returns the special metadata this Data-object is holding"""
        return self.metadata['special']
    
        
    def getMetadataInternal(self):
        """
        Return:
               - Object - A python-object of some sort - holding the internal (SciCraft-related) metadata for this data-object.

        Returns the internal metadata this Data-object is holding"""
        return self.metadata['internal']
    
        
    def toString(self):
        pass

    def get_data_info(self):
        """Returns the name and type/size of the data object.
        
        Examples: 'X - 6x4', 'MyX - molecule', 
                  'D - Dictionary with keys: a, b'
        """
        return "%s - %s" % (self.name, self.get_data_size_str())

    def get_data_size_str(self):
        """
        Returns:
            A string describing the type and/or size of the data.
            Examples: 'molecule', '63xString', '4x8'
        """
        # TODO Make sure the "else -> molecule assumtion holds.
        # TODO Make sure BASE-dataobjects are handeled.
        # TODO Never mind these problems; subclass Data and get rid of them!

        data = self.data
        try:
            if isinstance(data, dict):
                key_list = ", ".join(data.keys())
                dim_string = "Dictionary with keys: %s" %(key_list)
            else:
                
                if hasattr(data, "index") and len(data) > 0 and hasattr(data[0], "molecule"):
                    dim_string = "molecule"
                elif isinstance(data, list) and len(data) and isinstance(data[0], dict):
                    dim_string = Data(self.name, data[0]).get_data_size_str()
                else:
                    dim_string = data_info(data)
        except TypeError, e:
            dim_string = "n/a"
        
        return dim_string
        
    def get_sub_data(self, is_column, index):
        dim  = lists.get_repr_dim(self.data)
        if is_column:
            if dim == 2:
                return [row[index] for row in self.data]
            else:
                return [self.data[index]]
        else: #is row
            if dim == 2:
                return self.data[index]
            else:
                if index > 0:
                    raise IndexError, 'Index %d out of range' % index
                return self.data

    def get_data_size(self, is_column):
        dim = lists.get_repr_dim(self.data)
        if is_column:
            # retrieve number of cols
            if dim == 2:
                return len(self.data[0])
            elif dim == 1:
                return len(self.data)
            else:
                return 1
            
        else:
            # retrieve number of rows
            if dim == 2:
                return len(self.data)
            else:
                return 1
