# Should be run with main scicraft directory as working directory
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.Data.filters import *
from Numeric import *

class FilteredMatrix:
    """Represents a filtered data set. This means that the data
    can come from several MatrixDataSets, and that subsets (selection of rows)
    can be extracted from the data set. Supports data labelling in both rows
    and columns.
    """
    def __init__(self):
        """Does nothing special
        """
        # List of filters
        self._current_data = []

        # List of indices
        self._range = []

        # Gui plot
        self._plot = None

    def set_plot(self, plot):
        """Assigns a plot to this dataset.
        """
        self._plot = plot

    def add_matrix_column(self, source, column):
        """Add a column from a MatrixDataSet to this data set
        """
        self._current_data.append(ColumnFilter(source, column))
        source.add_listener(self)

    def add_matrix_row(self, source, row):
        """Add a row from a MatrixDataSet to this data set as a column.
        """
        self._current_data.append(RowFilter(source, row))
        source.add_listener(self)

    def get_sources(self):
        """Return a list of all the MatrixDataSet-objects that provide this
        dataset with data
        """
        return_list = []
        for filter in self._current_data:
            if not filter.get_source() in return_list:
                return_list.append(filter.get_source())
        
    def get_subset(self, rows = []):
        """Return a subset of data specified with a list of row indices.
        If an index is out of range for this data set, the row is simply
        ignored.
        """
        full_data = self.get_data()
        result_data = []
        error_string = ""
        if not rows:
            return full_data

        for row in rows:
            try:
                result_data.append(full_data[row])
            except IndexError:
                error_string += str(row)

        return result_data

    def get_data_in_range(self):
        return self.get_subset(self._range)

    def get_data(self):
        """Returns an Numeric array of the data contained
        in this filtered matrix
        """
        result_data = []        

        for filter in self._current_data:
            result_data.append(filter.get_values())

        # Until now we have been dealing with columns of data.
        # We transpose to make the data "row"-compatible with the rest of SC
        # Note that the resulting array object is NOT converted back to
        # a python list.        
        result_data = transpose(result_data)                
        return result_data

    def clear_filters(self):
        """Reset this dataset by deleting all filters
        """
        for filter in self._current_data:
            data_set = filter.get_source()
            data_set.remove_listener(self)
        
        self._current_data = []

    def set_range(self, rangelist = []):
        """Set the range of this data set, i.e. which
        rows that are going to be used.
        """
        self._range = rangelist

    def get_axis_labels(self):
        """Returns a list with a label for each column of thsi data set.
        """
        return_list = []

        for filter in self._current_data:
            label = filter.get_axis_label()
            if label:
                return_list.append(label)
                
        return return_list

    def get_object_labels(self):
        """Returns a list with a label for each row of this data set.
        If this filtered matrix has several sources, the first source
        with row labels will be used to denote the names for all objects.
        """
        for filter in self._current_data:

            object_list = filter.get_object_labels()
            if object_list:
                return object_list
            
        return []

    def rows_changed(self, marked_rows = [], unmarked_rows=[]):
        """Method called when something has been picked in a plot.
        Translates the specified index into a real index if a range subset
        of the data has been provided to the plot, and notifies all the
        filters about the event. This method does not manipulate data
        or plots in any way. 
        """

        real_marked = []
        real_unmarked = []
        for marked_row in marked_rows:
            real_index = marked_row
            if self._range:
                real_index = self._range[marked_row]

            real_marked.append(real_index)
                

        for unmarked_row in unmarked_rows:
            real_index = unmarked_row
            if self._range:
                real_index = self._range[unmarked_row]

            real_unmarked.append(real_index)
        
        for filter in self._current_data:
            filter.rows_changed(real_marked, real_unmarked)

    def rows_deleted(self):

        for filter in self._current_data:
            filter.get_source().rows_deleted()

    def data_updated_event(self, source):

        for filter in self._current_data:
            if filter.get_source() == source and not filter.index_within_bounds():
                self._current_data.remove(filter)
                # print "Error, index is not in bounds"

    def rows_changed_event(self, marked_rows = [], unmarked_rows = []):
        """Method fired by a listener server (like MatrixDataSet) when
        a row is marked. Checks if the index is within the range of this
        data set, and informs the plot (if any) to pick the actor represented
        by the row.
        """

        if self._plot:
            self._plot.pick_actors(marked_rows)
            self._plot.unpick_actors(unmarked_rows)

        
    def rows_deleted_event(self, indices):

        if self._plot:
            self._plot.delete_actors(indices)
        
        

    def row_undeleted_event(self, index):
        pass
    
