# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import types
import traceback
import sys
import os
from dprint import *
from System.Data.Node import Node
from System.Data.Data import Data
from System.events import NodeEvent
from System.Plugin import Plugin, PluginInterrupted
from System.Exceptions.ZPluginError import ZPluginError
from Library.Util import lists
# domain exceptions
class MissingOutputError(Exception):
    """Plugin did not provide output for all ports."""
    pass

class FailedSubObjectError(Exception):
    """Plugin did not return a result that we managed to decompose"""
    pass

class FunctionNode(Node):
    def __init__(self, name, pluginRepository):
        """
        Params:
            - name : string - a String with this node's name
            - pluginRepository : PluginRepository - Used to fetch plugin on execution
        """
        Node.__init__(self, name)
        self.plugin = None # Plugin-instance
        self.pluginType = None # Plugin-type (e.g "OctavePlugin")
        self.pluginRep = pluginRepository
        self.functionfile = ""
        self.functionName = name
        self._runner = None
        self._group = []
        self.sub_objects = {}
        self.icon = ""

        
    def setPluginType(self, pluginType):
        """Set what type of plugin this node should use when executing."""
        self.pluginType = pluginType

    def get_plugin_type(self):
        return self.pluginType
    
    def getPlugin(self):
        """Get last plugin instance used."""
        if self._runner:
            return self._runner.get_plugin()
        return self.plugin

    def setFileName(self, file_):
        """Set source path for the executing function."""
        self.functionfile = file_

    def getFunctionFileName(self):
        """Get absolute path to function script."""
        return self.functionfile

    def getFunctionName(self):
        """Get name of function to execute."""
        return self.functionName

    def setFunctionName(self,name):
        """Set name of function to execute."""
        self.functionName = str(name)

    def set_sub_objects(self, sub_objects):
        self.sub_objects = sub_objects
        
    def set_icon(self, icon):
        self.icon = icon
        
    def get_icon(self):
        if self.icon:
            # path given in zml file is absolute
            full_path = self.icon
            # path given in zml file is relative, merge with path to zml-file
            if not os.path.isabs(full_path):
                full_path = os.path.normpath(os.path.join(os.path.split(self.functionfile)[0], full_path))

            if os.path.isfile(full_path):
                return full_path

        return None
            
    def execute(self):
        """Execute this function node.

        Exception:
                - ValueError - if one of the output ports cannot be filled with data
                from the results of the function

        Runs function through plugin, getting function parameters from
        input ports and node parameters. Results are collected and
        sent to output ports.

        Names of output ports are used to map results.
        """
        if not self._runner:
            self.plugin = self.pluginRep.getPlugin(self.pluginType)
            self.plugin.setLocked()
        

        result = []
        try:
            try:
                if self._runner:
                    result = self._runner.execute(self)
                else:
                    result = self.plugin.execute(self.getFunctionName(),
                                                 self.getFunctionFileName(),
                                                 self.getInputPorts(),
                                                 self.getOutputPorts(),
                                                 self.getParameters(),
                                                 self.isInterrupted)
            except PluginInterrupted, e:
                dprint(flags.SYSTEM.DATA.NODE, "[FunctionNode] " + str(e))
                raise
            except Exception, e:
                trace_string = ""
                for level in traceback.format_tb(sys.exc_info()[2]):
                    trace_string += level
                    
                dprint(flags.SYSTEM.DATA.NODE,
                       "[FunctionNode] Error in node %s:" % self.name,
                       trace_string, e)
                raise
        finally:
            if not self._runner:
                self.plugin.setUnlocked()
        
        data_results = self._handle_sub_objects(result)
        
        if len(data_results) != len(self.outPorts): # use a method for sendign EXCEPTION - event
            raise MissingOutputError, \
                  "%s: Expected results for %s ports, but plugin returned %s." % \
                  (self.getFunctionName(), len(self.outPorts), len(result))

        data_by_name = {}

        for data in data_results:
            data_by_name[data.getName()] = data

        for port in self.outPorts:
            try:
                # Check if matching data is found
                port.setData(data_by_name[port.getVariableName()])
            except KeyError, e:
                # This probably happens because the script differs
                # from the zml-file (return-statements/outputports)
                raise MissingOutputError, '%s: Variable %s not returned by plugin.' % (self.getFunctionName(), e)


    def toXML(self):
        xml = Node.toXML(self)
        xml.insert(len(xml)-1,"<functionName>"+self.functionName+"</functionName>")
        return xml


    def set_runner(self, runner, group):        
        self._runner = runner
        self._group = group

    def remove_runner(self):
        self._runner = None
        self._group = []

    def isReady(self):
        if not self._runner:
            return Node.isReady(self)
        
        for port in self.getInputPorts():
            owner = port.getDataSource().getOwnerNode()
            if owner not in self._group:
                if port.isOptional():
                    if port.getDataSource() and not port.getData():
                        return False
                elif not port.getData():
                    return False
        return True

    
    def getNextNodes(self):
        """Returnes next nodes that must be executed.

        This method is a modification of the implementation in the
        superclass since all nodes in the group will already be in the
        waiting list.
        """
        next_nodes = Node.getNextNodes(self)
        next_nodes = lists.filter_list(next_nodes, self._group)
        return next_nodes


    def _handle_sub_objects(self, data_set):
        """Method responsible for extracting sub objects.

        The input will be a dictionary on the form
        {'out': {'appname': 'scicraft', 'version': '0.9.0'}}.
        the function then uses the R syntax to extract values.
        self.sub_object specifices which sub object to extract and
        which port to place it on. 
        """

        if not self.sub_objects:
            return data_set

        result = []
        for (port_name, object_names) in self.sub_objects.items():
            new_data = Data(port_name, None)
            data = None

            for data_res in data_set:
                if data_res.getName() == object_names[0] and \
                   type(data_res.getData()) == types.DictType:
                    data = data_res.getData()

            if len(object_names) > 1:
                for name in object_names[1:]:
                    if type(data) == types.DictType:
                        try:
                            data = data[name]
                        except KeyError:
                            data = None

            if data:
                new_data.setData(data)
                result.append(new_data)
            else:
                raise FailedSubObjectError, "Failed to extract sub object with the " + \
                      "search string: " + '$'.join(object_names)
        return result

    def get_formatted_output(self):
        """Retrieves a formatted output containing both stderr and stdout.

        The text string returned is formatted in the following manner:

        Output from last run:
        ---------------------
        ...

        Error output from last run:
        ---------------------------
        ...
        
        """
        stdout = self.get_standard_output()
        stderr = self.get_error_output()

        text = ""
        if stdout:
            text = "Output from last run:\n---------------------\n%s" % stdout
            
        if stderr:
            if text:
                text += "\n\n"
            text += "Error output from last run:\n---------------------------\n%s" % stderr

        return text
    
    def get_error_output(self):
        """Retrieves what was printed on stderr in the plugin during last run."""
        if self._runner:
            return self._runner.get_plugin().get_stderr_from_last_run()
        elif self.plugin:
            return self.plugin.get_stderr_from_last_run()
        else:
            return ""

    def get_standard_output(self):
        """Retrieves what was printed on stdout in the plugin during last run."""
        if self._runner:
            return self._runner.get_plugin().get_stdout_from_last_run()
        elif self.plugin:
            return self.plugin.get_stdout_from_last_run()
        else:
            return ""
