# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

class MatrixDataSet:
    """DataSet representing a two-dimensional matrix."""
    def __init__(self, data = []):
        # Data in form of lists or
        # Numeric arrays
        self._data = data

        # Indices of marked rows
        self._marked_rows = []

        # Indices of removed rows
        self._deleted_rows = []

        self._listeners = []        
        self._row_labels = []
        self._column_labels = []
    
    def get_data(self):
        return self._data

    def get_row_labels(self):
        return self._row_labels

    def get_column_labels(self):
        return self._column_labels

    def set_row_labels(self, labels):
        self._row_labels = labels

    def set_column_labels(self, labels):
        self._column_labels = labels

    def get_num_rows(self):
        return len(self._data)

    def get_num_columns(self):
        row = self._data[0]
        return len(row)
    
    def update_data(self, data):
        self._data = data
        self.notify_data_updated()

    def get_marked_rows(self):
        return self._marked_rows

    def get_deleted_rows(self):
        return self._deleted_rows

    # Listener server methods
    def add_listener(self, listener):
        if not listener in self._listeners:
            self._listeners.append(listener)

    def notify_data_updated(self):
        for listener in self._listeners:
            listener.data_updated_event(self)

    def rows_changed(self, marked_rows = [], unmarked_rows = []):
        """Marks a row in the dataset and notifies all the
        listeners about the event.
        """

        for marked_row in marked_rows:            
            if not marked_row in self._marked_rows and marked_row < self.get_num_rows():
                self._marked_rows.append(marked_row)
            
        for unmarked_row in unmarked_rows:
            if unmarked_row in self._marked_rows and unmarked_row < self.get_num_rows():
                self._marked_rows.remove(unmarked_row)            

        for listener in self._listeners:
            listener.rows_changed_event(marked_rows, unmarked_rows)     
            
    def rows_deleted(self):
        """Untested
        """
        for listener in self._listeners:
            listener.rows_deleted_event(self._marked_rows)

        for row_index in self._marked_rows:            
                       
            if not row_index in self._deleted_rows:
                self._deleted_rows.append(row_index)
        self._marked_rows = []
            
    def rows_undeleted(self):
        pass    

    def remove_listener(self, listener):
        try:
            index = self._listeners.index(listener)
            return self._listeners.pop(index)
        except ValueError:
            # Listener was not in list
            return

    def get_listeners(self):
        return self._listeners

    def clear_listeners(self):
        self._listeners = []
