# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.



import thread
import time
import traceback

from random import *

from dprint import *
from System.Data.ModuleObject import ModuleObject
from System.Data.ports import NodeInputPort, NodeOutputPort, MultipleInputPort
from System.Data.Parameter import Parameter
from System.Plugin import PluginInterrupted
from System.events import NodeEvent
from Library.Util.unrepr import unrepr
from qt import QThread


class Node(ModuleObject):
    """Base class for all nodes."""

    def __init__(self, name):
        """Create a new do-nothing node.

        Subclass Node and reimplement Node.execute() to define
        behavior.

        Params:
        name - the name of this node
        """
        ModuleObject.__init__(self)
        self.setName(name)
        
        # Link to any documentation
        # Set from XML in functionNodes. Set in subclasses in other
        # classes.
        self._doc_url = None
        # Used by XML for storing filenames
        self.nodeRepName = ""
        self._parameters = []
        self.inPorts = []
        self.outPorts = []
        # can in / out ports be added to this Node?
        self.canAddInputPorts = True
        self.canAddOutputPorts = True
        # If given, will override what the client reports as its
        # placements
        self.menuAddress = None
        self.moduleDiagram = None

        # states
        # True when node is running
        self._alive = False
        self._started = False
        # True when node is interrupted
        self._interrupted = False
        self._crashed = False
        
        self._multi_ports = []

        # default to keep output after data has been transmitted to
        # connected nodes
        self.setVariable("keep_output", True)

    def getConnectionsToNode(self, toNode):     
        """Get all OutputPort->InputPort connections from this
            node to another node.
        """
        connections = []    

        for inputPort in toNode.getInputPorts():
            connected = inputPort.getDataSource()
            if connected in self.getOutputPorts():
                pair = (connected, inputPort)
                connections.append(pair)

        return connections

    def get_all_input_ports(self):
        """Get all input ports available, emulated and normal."""
        ports = self.getInputPorts()[:]
        for multi_port in self.get_multi_input_ports():
            ports.extend(multi_port.get_emulated_input_ports())
        return ports

    def get_successing_nodes(self):
        """Returnes nodes that recives data from this node"""
        nodes = {}                
        for output in self.getOutputPorts():
            connected_nodes = output.get_connected_nodes()
            for node in connected_nodes:
                nodes[node] = 0

        return nodes.keys()
            
    def get_module_diagram(self):
        return self.moduleDiagram

    def set_module_diagram(self, moduleDiagram):
        self.moduleDiagram = moduleDiagram

    def start_blocking(self, single=False):
        """Execute node and wait.

        Returns when execution is finished. If flag 'single' is True,
        next nodes will not be added to module diagram.
        """
        self._run_and_ship(single)

    def get_providing_nodes(self):
        """Get list of all nodes providing direct input to this node.

        Returns all nodes connected to this node's input ports.
        """
        nodes = {} # used as set

        for input_port in self.getInputPorts():
            connected_node = input_port.getDataSource().getOwnerNode()
            nodes[connected_node] = 0 # any value

        for multi_port in self.get_multi_input_ports():
            connected_ports = multi_port.get_data_sources()
            for port in connected_ports:
                nodes[port.getOwnerNode()] = 0

        return nodes.keys()

    def set_keep_output_in_providing_nodes(self):
        """Make sources of data to this node keep output after execution."""
        def set_keep_output(node):
            node.set_keep_output()

        map(set_keep_output, self.get_providing_nodes())
    
    def setNodeRepName(self, text):
        self.nodeRepName = text

    def getParameterByName(self, name):
        """Gets a parameter by case-insensitive name."""
        for param in self._parameters:
            if param.getName().lower() == name.lower():
                return param

    def addParameter(self, name, type, description, min = None, max = None, 
                     step = None, values = None, default = "", value = None):
        parameter = self.getParameterByName(name)
        if not parameter:
            parameter = Parameter()
            self._parameters.append(parameter)

        parameter.setName(name)
        parameter.setType(type)
        if value:
            parameter.setValue(value)
        parameter.setDescription(description)
        parameter.setRestrictions(min = min, max = max, values = values, 
            default = default, step = step)

        return parameter

    def getParameters(self):
        return self._parameters

    def _run_and_ship(self, single):
        """Execute node, send data to successive nodes.

        If flag 'single' is True, next nodes will not be added to
        module diagram.
        """
        self._alive = True
        self._interrupted = False
        self._crashed = False
        # must be set after alive
        self._started = True

        try:
            self.execute()
        except PluginInterrupted:
            # just swallowed
            self._interrupted = True
        except Exception, e:
            self._crashed = True
            trace_string = ""
            for level in traceback.format_tb(sys.exc_info()[2]):
                trace_string += level
            dprint(flags.SYSTEM.DATA.NODE,
                   "[Node] Error in node %s:" % self.name, trace_string, e)
            self.moduleDiagram.sendEvent(NodeEvent(self.moduleDiagram,
                                                   self, "EXCEPTION", str(e)))
        else:
            #Adds the nodes following this one to the waiting
            if not single:
                # must be done before setting oneself finished
                self.moduleDiagram.addNextNodes(self.getNextNodes())
                dprint(flags.SYSTEM.DATA.NODE, "[Node] Next nodes:",
                       self.getNextNodes())
                self._send_output()
                dprint(flags.SYSTEM.DATA.NODE, "[Node] Output sent from",
                       self.getName())

        self._alive = False
        self.moduleDiagram.setFinished(self)
        self.moduleDiagram.sendEvent(NodeEvent(self.moduleDiagram, self, 
            "FINISH", "The node " + self.name + " is finished."))
        dprint(flags.SYSTEM.DATA.NODE, "[Node] Node finished running:", 
            self.name)
        dprint(flags.THREADING, "[Threading:Node] Stopping thread (ID",
                str(thread.get_ident()) + ") for node", self.name)

    def has_data_waiting(self):
        ports = self.getOutputPorts()
        if ports:
            for port in ports:
                if not port.getData():
                    return False
            return True
        else:
            return False
            
    def execute(self):
        """Execute node.

        Should be overridden in subclasses. Should read data from
        input ports, perform an operation on the data, then put the
        results on the output ports.
        """
        pass

    def _send_output(self):
        """Send output to ports of connected nodes.
        
        Propagate output to correct nodes in module diagram. If some
        of the input nodes are busy, this method sleeps for a while
        and tries again.
        """
        # gather list of all connected input ports
        for port in self.outPorts:
            if port.getData():
                port.ship_data()
                
                # Remove sent data from this node
                if not self.get_keep_output():
                    port.setData(None)

    def set_keep_output(self, should_keep=True):
        """If set keeps all data on output ports after executing node.

        Default behavior is to delete all data on output ports after
        shipping to connected nodes. Setting keep_output True makes
        this node retain all data. This is primarily useful for the
        ability to run a module diagram from a selected spot.
        """
        self.setVariable("keep_output", should_keep)

    def get_keep_output(self):
        """Check if node keeps data on output ports after execution."""
        return unrepr(self.getVariable("keep_output"))

    def collect_data(self):
        """Collect data from nodes connected to input ports.

        Depends on nodes having data available.

        Throws NodeInputPort.NoDataToCollectError when a port has no
        available data.
        """
        for input_port in self.getInputPorts():
            input_port.collect_data()

        for multi_input in self.get_multi_input_ports():
            multi_input.collect_data()

    def toXML(self):
        """Create and return a list of strings, one for each line of xml."""
        returnlist = []
        returnlist.append("<node>")
        returnlist.append("\t<class>" + str(self.getNodeRepName()) + "</class>")

        for line in ModuleObject.toXML(self):
            returnlist.append("\t"+line)

        if self.getParameters() or self.getInputPorts() or \
               self.get_multi_input_ports():
            returnlist.append("\t<input>")
            for pa in self.getParameters():
                for line in pa.toXML():
                    returnlist.append("\t\t" + line)
            for ip in self.getInputPorts():
                for line in ip.toXML():
                    returnlist.append("\t\t" + line)
            for mip in self.get_multi_input_ports():
                for line in mip.to_xml():
                    returnlist.append("\t\t" + line)
            returnlist.append("\t</input>")

        if len(self.getOutputPorts()):
            returnlist.append("\t<output>")
            for op in self.getOutputPorts():
                for line in op.toXML():
                    returnlist.append("\t\t" + line)
            returnlist.append("\t</output>")

        returnlist.append("</node>")
        return returnlist

    def add_input_port(self, variableName, validType="", description="",
                       options="", optional=False):
        """Create and return a named input port.

        The port is optionally created with a type and a
        description. Available options used are FORCE and
        isNamed. Marking the port optional allows the diagram to run
        without connecting to the port. After adding a connection to
        the port, the port's functionality is equal to a non-optional
        one.
        
        """
        if self.isAllowedAddInputPorts() or options.find("FORCE") != -1:

            if not description:
                description = "Variable " + variableName
            for input in self.inPorts:
                if input.getName().lower() == variableName.lower():
                    #reflect possible change in case
                    input.setName(variableName)
                    input.setDescription(description)
                    input.setOptional(optional)
                    return input
                
            port = NodeInputPort(self, variableName, validType, description,
                                 optional)
            if options.find("isNamed=1") != -1:
                port.setIsNamed(1)

            self.inPorts.append(port)
            port.setOwnerNode(self)
            return port

        return False

    def getInputPorts(self):
        """
        Return:
            - List[NodeInputPort] - a list of all NodeInputPorts of this Node

        Return all the node's input ports"""

        return self.inPorts

    def getInputPortByName(self, name):
        for port in self.inPorts:
            if port.getName() == name:
                return port

        for multi_port in self.get_multi_input_ports():
            for port in multi_port.get_emulated_input_ports():
                if port.getName() == name:
                    return port
        return False

    def getOutputPortByName(self, name):
        for port in self.outPorts:
            if port.getName() == name:
                return port

        return False

    def getNonvalidInputPorts(self):
        """
        Return:
            - List[NodeInputPort] - a list of all NodeInputPorts which aren't
            connected to NodeOutputPorts """

        tempList = []
        for port in self.inPorts:
            if not port.getDataSource() and not port.isOptional():
                tempList.append(port)

        return tempList
    
    def isAllowedAddInputPorts(self):
        return self.canAddInputPorts

    def isAllowedAddOutputPorts(self):
        return self.canAddOutputPorts
    
    def setCanAddInputPorts(self, canAdd):
        self.canAddInputPorts = canAdd
    
    def setCanAddOutputPorts(self, canAdd):
        self.canAddOutputPorts = canAdd

    def add_output_port(self, variableName, validType="", description="",
                        options=""):
        """Create and return a named output port.

        The port is optionally created with a type and a
        description. Available options used are FORCE and
        isNamed.
        
        """
        if self.isAllowedAddOutputPorts() or options == "FORCE":

            for output in self.outPorts:
                if output.getVariableName().lower() == variableName.lower():
                    # Make sure new case is used
                    output.setName(variableName)
                    output.setDescription(description)
                    return output

            port = NodeOutputPort(self, variableName, validType, description)
            self.outPorts.append(port)
            port.setOwnerNode(self)
            return port
        return False

    def getNodeOutputPort(self, name):
        """
        Params:
            - name : String - the variableName of the NodeOutputPort

        Return:
            - NodeOutputPort - The NodeOutputPort expected

        Exception:
            - ValueError - when a port cannot be found

        The method return a NodeOutputPort specified by
        its variableName"""

        for cu in self.outPorts:
            if cu.getVariableName() == name:
                return cu

        raise ValueError, 'NodeOutputPort not found'

    def getOutputPorts(self):
        """
        Return:
            - List[NodeOutputPort] - a list of all NodeOutputPorts

        Returns a list of all this Node's output ports"""

        return self.outPorts

    def removeConnectionsFromNode(self, fromNode):
        """Remove all connections from otherNode to this node."""
        outPorts = fromNode.getOutputPorts()

        for inPort in self.getInputPorts():
            if inPort.getDataSource() in outPorts:
                inPort.disconnectFromDataSource()

        for multi_port in self.get_multi_input_ports():
            for output_port in multi_port.get_data_sources()[:]:
                if output_port in outPorts:
                    multi_port.remove_data_source(output_port)

    def isParametersReady(self):
        """Returns 'True' if all parameters has been set with a value."""
        for param in self._parameters:
            if not param.isValueSet():
                return False
        return True

    def getNextNodes(self):
        """
        Return:
            - List[Node] - retruns a list of the "next" nodes following this node
            in a module diagram, provided that the data on the corresponding
            NodeOuputPort of this node is set. """
        nextNodes = []
        for outPort in self.outPorts:
            if outPort.getData():
                for newport in outPort.getConnectedInputPorts():
                    newnode = newport.getOwnerNode()
                    if nextNodes.count(newnode) == 0:
                        nextNodes.append(newnode)
        return nextNodes

    def isReady(self):
        """Returns 'True' if node is ready to start. 'False' otherwise."""
        for multi_port in self.get_multi_input_ports():
            for emulated_port in multi_port.get_emulated_input_ports():
                if not emulated_port.getData():
                    return False
            
        for port in self.inPorts:
            if port.isOptional():
                if port.getDataSource() and not port.getData():
                    return False
            elif not port.getData():
                return False

        return True

    def isStartNode(self):
        """Returns 'True' if node is independent of other nodes, and
        therefore can be started immediately.  Returns 'False' if node
        need data from other nodes.
        """
        for multi_port in self.get_multi_input_ports():
            if multi_port.get_emulated_input_ports():
                return False

        if self.inPorts:
            return False
        
        return True

    def is_started(self):
        """Find out if node has been started since last clean()."""
        return self._started
        
    def is_alive(self):
        """Returns 'True' if node is running."""
        return self._alive

    def get_status_string(self):
        """ Returns a string describing the nodes status. """
        if not self.is_started():
            return "Not started"
        if self.is_alive():
            return "Running"
        if self._interrupted:
            return "Interrupted"
        if self._crashed:
            return "Crashed"
        return "Finished" 

    def clean(self):
        """Cleans the input and output ports of this object"""
        dprint(flags.SYSTEM.DATA.NODE, "[Node] Cleaning node:", self.getName())
        self._alive = False
        self._started = False
        self._interrupted = False
        self._crashed = False
        for inPort in self.inPorts:
            inPort.clean()

        for outPort in self.outPorts:
            outPort.clean()

    def getNodeRepName(self):
        return self.nodeRepName

    def remove_port(self, unwanted_port):
        if unwanted_port in self.inPorts:
            self.inPorts.remove(unwanted_port)
            unwanted_port.disconnectFromDataSource()
            unwanted_port.clean()
        if unwanted_port in self.outPorts:
            self.outPorts.remove(unwanted_port)
            for outport in unwanted_port.getConnectedInputPorts():
                outport.disconnectFromDataSource()
            unwanted_port.clean()
    
    def setName(self, name):
        """Overwrites method in superclass ModuleObject. 
            'None' is not allowed as name. If encountered, the last part
            of the node class is used. If node class i unavailable, 
            name is set to "unnamed".
        """
        if name is None:
            if (self.nodeRepName is None or self.nodeRepName == ""):
                name = "unnamed"
            else:
                # Use last part of node class/noderepname. 
                name = self.nodeRepName.replace(":", ".")
                name = self.name.split(".")[-1]
        ModuleObject.setName(self, name)
    
    def setPosition(self, x, y):
        """Set position of node in GUI."""
        # store as variable, to be stored by XML
        self.setVariable("position", "[%s, %s]"%(x,y))
    
    def getPosition(self):
        """Get position of node in GUI."""
        var = self.getVariable("position")

        if not var:
            var = "[0, 0]"
            self.setVariable("position", var)

        # position is text in format [xpos-float, ypos-float]
        # example: "[324.0, 234.0]"

        varlist = var[1:-1].split(",")

        var = (float(varlist[0]), float(varlist[1]))
        return var

    def getNodeType(self):
        """Gets the actual classname of a Object. Objects may be
        inherited though, so this should *not* be used for any
        critical identification of objects
        """
        return self.__class__.__name__

    def _postEvent(self, type, description):
        event = NodeEvent(self.moduleDiagram, self, type, description)
        self.moduleDiagram.sendEvent(event)    

    def getHtmlLink(self):
        return self._doc_url

    def setHtmlLink(self, link):
        self._doc_url = link

    def getMenuAddress(self):
        return self.menuAddress

    def setMenuAddress(self, address):
        self.menuAddress = address

    def isInterrupted(self):
        return self._interrupted

    def interrupt(self):
        self._interrupted = True

    def get_multi_input_ports(self):
        """Get all MultipleInputPorts used by this node."""
        return self._multi_ports

    def add_multi_input_port(self, port_name):
        """Create a new MultipleInputPort for this node."""
        if self.isAllowedAddOutputPorts():
            named_ports = {}
            for port in self.get_multi_input_ports():
                named_ports[port.get_name()] = port

            # if port already exists, return that
            if port_name in named_ports:
                return named_ports[port_name]

            port = MultipleInputPort(port_name, self)
            self._multi_ports.append(port)
            return port
        return False

    def get_port_data(self, input_node):
        """Returnes data that is sent between this node and input_node"""
        data = []
        output = self.getOutputPorts()
        for port in output:
            for input_ in port.getConnectedInputPorts():
                if input_ in input_node.get_all_input_ports():
                    if port.getData():
                        data.append(port.getData())

        return data
