# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os
import types

import Numeric
import time
from Library.Util import RVariableNameIdentifier
from Library.Util.rtester import RTester
from Library.Util.Config import *
from System.Data import Data
from System.FileHandlers.Input.FileReader import FileReader, data_info
from System.Exceptions import ZPluginError
from Numeric import array
from Library.Util import lists
from mutex import mutex



class RFileReader(FileReader):
    """This class is responsible for reading R-files. It inherits its methods from
        FileReader, but overwrites all methods."""

    supportedFileExtensions = ["RData"]
    variable_name_support = True
    description = "GNU R file format. Supports all R data file formats supported " +\
        "by R itself." 
    fileDescription = "R"
    name = "RFileReader"
        
    def __init__(self):
        """Constructor for creating an RFileWriter-object"""
        FileReader.__init__(self)
        self.current_filename = None
        self.data = []

        
    def get_variable_names(self, sourcePath):
        while not self._mutex.testandset():
            time.sleep(0.5)
        try:
            returnValue = self._getVariableNamesCriticalSection(sourcePath)
        finally:
            self._mutex.unlock()
        return returnValue
        

    def _getVariableNamesCriticalSection(self, sourcePath):
        """
            Params:
                - sourcePath : String - The path of the file containing R-variables.
            Return:
                - res : List[String] - List of variables names.
            Exception:
                - RException - Something went wrong in the R-session.
        """
        RTester.r_install_test()
        import rpy
        if not os.path.exists(sourcePath):
            raise ZPluginError.ZPluginError, "File %s does not exist." % sourcePath


        if os.name == "windows" or os.name == "nt":
            # [TODO] Really should implement simple caching here as in BASEFileReader.
            rVarID = RVariableNameIdentifier.RVariableNameIdentifier()
            return rVarID.findRVars(sourcePath)
        else:
            try:
                res = rpy.r.load(sourcePath)
            except rpy.RException, e:
                raise ZPluginError.ZPluginError, str(e)
            if type(res) is types.StringType: # Only one variable-name returned
                res = [res]
            return res


    def read_data(self, sourcePath):


        while not self._mutex.testandset():
            time.sleep(0.5)
        try:
            returnValue = self._readDataCriticalSection(sourcePath)
        finally:
            self._mutex.unlock()
        return returnValue

        
    def _readDataCriticalSection(self, sourcePath):
        """
            Params:
                - sourcePath : String  - The path and name of the file being read.
        Return:
        - data : List[Data]  - A list of Data-objects containing the data
                       read from file.          

            Exception:
                - IOError - Cannot read. Possible causes:
                                                    * No read priveleges at selected path.
                                                    * Path/file does not exist.

            This is a method for reading R ascii-files."""
        
        RTester.r_install_test()
        import rpy
        if type(sourcePath) is not types.StringType:
            raise TypeError, "Invalid parameter. Expected string, got %s" % type(sourcePath)

        if not os.path.exists(sourcePath):
            raise ZPluginError.ZPluginError, "Cannot find file %s." % sourcePath
    
        variables = [] # Array holding the variable-names returned by RPy.
        self.data = [] # Array holding the data-objects retrieved from the R-file.
        res = 0        # Variable holding the raw information returned by RPy.


        try:                    
            res = rpy.r.load(sourcePath) # Raw result from RPy. Either a string or a list of strings returned.
                           # The strings are the names of the variables loaded into R.
        except IOError:
            raise ZPluginError.ZPluginError, "R could not read from file %s. Check path and permissions." % sourcePath
             
        except rpy.RException:
            raise ZPluginError.ZPluginError, "[RFileReader] ERROR: The file '%s' is either corrupted or not a R data file. No data loaded." % str(sourcePath)
            # [NOTE] This will now fail in InputNode.py, complaining about variable X not found. I think this is OK, as the
            #        variable X actually is not found in the file :) If thie behavior is unacceptable, raise an exception here.

        if type(res) is types.StringType:
        # Only one variable returned from R.
            variables.append(res)
        else:
        # Has to be a list of strings.
            variables = res
        
        from Library.Plugins.RPluginFiles import RObjectHandler
        
        robjHandler = RObjectHandler.RObjectHandler()
        for x in variables:
            self.data.append(robjHandler.getRVar(x))

        return self.data


    def get_variable_info(self, filename, variable_name = None, node = None):

        variable_info = {}
        # RPy makes SciCraft crash when read_data() is called from within this class.
        # This needs to be fixed!
        if os.name == "windows" or os.name == "nt":
            if variable_name:
                return "RData"
            else:
                var_names = self.get_variable_names(filename)
                for var in var_names:
                    variable_info[var] = "RData"
                return variable_info
        else:
            if node:
                filename = self.convert_to_absolute_filepath(filename, node.get_module_diagram().getFilename())
                
            if not os.access(filename, os.R_OK) or os.path.isdir(filename):
                return "clear"
            try:
                data_array = self.read_data(filename)
            except Exception, e:
                return "clear"

            for data in data_array:
                if isinstance(data.getData(), dict):
                    variable_info[data.getName()] = "dictionary"                   
                else:
                    variable_info[data.getName()] = data_info(data.getData())


            if variable_name:
                if variable_info.has_key(variable_name):
                    return variable_info[variable_name]
                else:
                    return ""
            else:
                return variable_info
                
RFileReader._mutex = mutex()
