# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Nodes import EditNodes, PlotNodes, FileHandlers, TestNodes
from Library.Nodes import UserInputNodes

from dprint import *
from Library.Util.Config import *
from System.Plugin.PluginRepository import PluginRepository
from System.XMLReader import handle_plotnode_xml, convert_from_string_to_dom
from System import XMLFunction
from System.Data.FunctionNode import FunctionNode
from System.Data.PlotNode import PlotNode
from System.events import InfoEvent
from Library.Nodes.PlotNodes.Plot2DNode import Plot2DNode
from xml.parsers.expat import ExpatError
from System.XMLFunction import MissingTagsError

from System.Data.Node import Node

# used when checking if a class is a subclass of Node
from System.Data.Node import Node as super_class

import copy
import os
import re
import types

# domain exceptions

class UnknownNodeError(Exception):
    """Tried to create node not known to repository."""
    pass

class NodeEntry:
    def __init__(self):
        pass
    def get_full_name(self):
        return ".".join(self.get_categories() + [self.get_short_name()])

class FunctionNodeEntry(NodeEntry):
    """Represents an available function node."""
    def __init__(self, short_name, filename, categories):
        NodeEntry.__init__(self)
        self._short_name = short_name
        self._categories = categories
        self._filename = filename

    def get_short_name(self):
        return self._short_name

    def get_categories(self):
        function = XMLFunction.build(self._filename)
        address = function.getMenuAddress()
        if address:
            return address.split(".")
        else:
            return self._categories

    def create_node(self):
        """Create the FunctionNode instance represented by this entry."""
        function = XMLFunction.build(self._filename)
        function.setNodeRepName(self.get_full_name())
        function_name = function.getFileName()
        function.setFileName(os.path.join(os.path.split(self._filename)[0],
                                          function_name))
        return function.createFunctionNode(PluginRepository())

class PythonNodeEntry(NodeEntry):
    """Represents an available python based node."""
    def __init__(self, short_name, categories, class_):
        NodeEntry.__init__(self)
        self._short_name = short_name
        self._categories = categories
        self._class = class_

    def get_short_name(self):
        return self._short_name

    def get_categories(self):
        return self._categories

    def create_node(self):
        node = self._class(self.get_short_name())
        node.setNodeRepName(self.get_full_name())
        return node

class PythonNodeSource:
    """Gives repository access to all nodes in a set of modules."""
    def __init__(self, modules):
        """Generate a list of entries from all nodes in modules.

        The modules parameter is a list of modules containing class
        modules within. Each class module should contain a class of
        the same name.

        It is a requirement that all modules contain the __all__
        attribute showing the class modules within, or that all class
        modules are already available in the modules' namespace.
        """
        self._modules = modules

    def get_entries(self):
        entries = []
        for module in self._modules:
            # does the following line have any function at all?
            #attributes = map(lambda a : getattr(module, a), dir(module))

            class_modules = {}
            classes = {}

            category = module.__name__.split('.')[-1]

            attributes = {}
            for a in dir(module):
                attributes[a] = None
            
            if hasattr(module, "__all__"):
                for a in module.__all__:
                    if not attributes.has_key(a):
                        __import__(module.__name__ + "." + a)
                        
            for name in dir(module):
                sub_module = getattr(module, name)

                if not isinstance(sub_module, types.ModuleType):
                    continue

                if not hasattr(sub_module, name):
                    continue

                node_class = getattr(sub_module, name)

                if not isinstance(node_class, types.ClassType):
                    continue

                if not issubclass(node_class, Node):
                    continue

                entries.append(PythonNodeEntry(name, [category], node_class))

        return entries

class FunctionNodeSource:
    """Gives repository access to function nodes in a directory."""
    def __init__(self, path):
        self._entries = []

        for root, dirs, files in os.walk(path):
            zml_files = filter(lambda entry : entry.endswith('.zml'), files)
            names = map(lambda entry : os.path.splitext(entry)[0], zml_files)
            
            # split path into categories and remove empty entries
            categories = filter(None, root[len(path):].split(os.sep)[1:])
            
            for name in names:
                zml_file = os.path.join(root, name + '.zml')
                self._entries.append(FunctionNodeEntry(name, zml_file, categories))

            if '.svn' in dirs:
                dirs.remove('.svn') # don't visit SubVersion directories
    
    def get_entries(self):
        """Return all entries accessible through this source."""
        return self._entries

class NodeRepository:
    """Repository of available nodes in SciCraft.

    All available nodes, both subclasses of System.Data.Node.Node and function
    nodes are made available through the NodeRepository interface.
    """

    DEFAULT_MODULES = TestNodes, EditNodes, PlotNodes, FileHandlers, UserInputNodes
                      
    
    def __init__(self, pluginRepository=PluginRepository(),
                 node_modules=None, loadPath=None, functionPath=None,
                 event_server=None):
        """Constructor for NodeRepository. Makes a NodeRepository.

        pluginRepository - store of all plugins
        node_modules - modules containing class modules (modules with
                       same-named classes)
        loadPath - path to user-defined .zml-files (functions)
        functionPath - path to included .zml-files (functions)
        """
        self._pluginRepository = pluginRepository
        self._event_server = event_server

        # remember paths for reloading
        self._node_modules = None
        self._load_path = None
        self._function_path = None
        self.updatePaths(node_modules, loadPath, functionPath)

    def reload_functions(self):
        """Reload all functions from paths loaded earlier."""
        self.updatePaths(self._node_modules, self._load_path,
                         self._function_path)

    def updatePaths(self, node_modules=None, loadPath=None, functionPath=None):
        """Update this repository with nodes/functions in new paths."""

        self._node_modules = node_modules
        self._load_path = loadPath
        self._function_path = functionPath

        # clear settings
        self._sources = []
        self._entries = {}

        if loadPath or functionPath:
            for path in self.__create_load_path_list(loadPath, functionPath):
                self._sources.append(FunctionNodeSource(path))

        if node_modules:
            self._sources.append(PythonNodeSource(node_modules))

        duplicates = {}

        for source in self._sources:
            for entry in source.get_entries():
                if not self._entries.has_key(entry.get_short_name()):
                    self._entries[entry.get_short_name()] = entry
                else:
                    duplicates.setdefault(entry.get_short_name(), []).append(entry)

        # to ensure backwards compatibilty in ZML-files
        if self._entries.has_key('VectorSpreadsheetNode'):
            self._entries['VectorSpreadsheetNode'] = self._entries['SpreadSheetNode']

        self._report_duplicates(duplicates)

    def _report_duplicates(self, dupes):
        for name, duplicates in dupes.items():
            original = self._entries[name]
            message = "Duplicate function-entries\n"
            full_names = map(NodeEntry.get_full_name, [original] + duplicates)
            
            message += "\n".join(full_names)
            message += "\n\nOnly one node will use the name '%s':\n" % name
            message += original.get_full_name() + "\n"
            message += "Other nodes with same name will be ignored."
            if self._event_server:
                self._event_server.sendEvent(InfoEvent(None, "INFO", message))
            dprint(flags.SYSTEM.NODEREPOSITORY, message)
            
            
    def createNodeByName(self, name):
        """Returns a single node as requested with the name:string.

        The name is the fully qualified node name as returned by
        getNodeList(). This name is also the same as read from
        XML.
        """
        plotvalue = False
        isCustomPlot = False
        oldname = name

        base_name = name.split(":")[0].split(".")[-1]

        if name.count(":"):
            first, last = name.split(':', 1)

            if base_name == "PlotNode":
                name = "Library.Nodes." + first
                plotvalue = last.split(".", 1)[-1]

                path = name.split(".")
                plotnode = PlotNode(name, plotvalue)
                plotnode.nodeRepName = oldname
                plotnode.setName(oldname.split(":", 1)[0])
                return plotnode
                    
            # checking if the node to be created is a custom plot
            if base_name == "Customplots":
                #in the case of a custom Plot2D node
                customPlotNode = Plot2DNode(name)
                customPlotNode.nodeRepName = name
                customPlotNode.setName(name.split(":")[1])

                #setting the custom settings
                path = os.path.join(Config.userconfigdir, "customplots", name.split(":")[1] + ".zpl")
                stream = open(path, 'r')
                xml_string = stream.read()
                dom = convert_from_string_to_dom(xml_string)
                plotnode_xml = dom.getElementsByTagName("plotnode_config").item(0)
                handle_plotnode_xml(plotnode_xml, customPlotNode)

                return customPlotNode
        else:
            if self._entries.has_key(base_name):
                return self._entries[base_name].create_node()

            raise UnknownNodeError, "Can not find the node " + name + " in the repository."

    def get_doc_path_by_node_name(self, name):
        if self._entries.has_key(name):
            entry = self._entries[name]
            filename = entry._filename
            function = XMLFunction.build(filename)
            doc_path = function.get_doc_path()
            if doc_path:
                full_path = os.path.join(os.path.split(filename)[0], doc_path)
                if os.path.exists(full_path) and os.path.isfile(full_path):                
                    return full_path

        return None
        
    def get_icon_path_by_node_name(self, name):
        if self._entries.has_key(name):
            entry = self._entries[name]
            filename = entry._filename
            function = XMLFunction.build(filename)
            icon_path = function.get_icon()
            if icon_path:
                # path given in zml file is absolute
                if os.path.isabs(icon_path):
                    full_path = icon_path
                # path given in zml file is relative, merge with path to zml-file
                else:
                    full_path = os.path.normpath(os.path.join(os.path.split(filename)[0], icon_path))

                if os.path.isfile(full_path):
                    return full_path

        return None
        
    def get_class_by_node_name(self, name):
        if self._entries.has_key(name):
            return self._entries[name]
        return None

    def getNodeList(self):
        """Return list of (fully-qualified-class-name, name) pairs.

        The fully qualified class name can be used with
        createNodeByName() to retrieve an instance of the node.
        """
        nodes = {}
        for short_name, entry in self._entries.items():
            try:
                nodes[entry.get_full_name()] = short_name
            except ExpatError, e:
                message = "Invalid zml-file for function %s:\n%s" % (short_name, e)
                self._event_server.sendEvent(InfoEvent(None, "INFO", message))
            except MissingTagsError, e:
                message = "Invalid zml-file for function %s:\n%s" % (short_name, e)
                self._event_server.sendEvent(InfoEvent(None, "INFO", message))
            
        return nodes

    def __create_load_path_list(self, loadpath, functionpath):
        """Return list of current loadpaths."""

        loadPathList = []

        if loadpath:
            # correct possible wrong directory separators
            if   os.name == "nt":
                loadpath = loadpath.replace("/",  os.sep)
            elif os.name == "posix":
                loadpath = loadpath.replace("\\", os.sep)

            userPathList = loadpath.split(";")

            for path in userPathList:
                path = path.strip()

                if not path:
                    continue
        
                if functionpath and os.path.abspath(path) == os.path.abspath(functionpath):
                    continue

                if not os.path.exists(path):
                    print "Warning. Loadpath", path, "does not exist."
                    continue

                loadPathList.append(path)

        if functionpath:
            loadPathList.insert(0, functionpath)

        return loadPathList



    def _search_nodes(self):
        """Method that will search and find nodes"""
        nodes = {}        
        dict_ = copy.copy(globals())
        
        for key in dict_:    
            for attr in dir(dict_[key]):
                a = getattr(dict_[key], attr)
                if type(a) == types.ClassType and a.__name__ != "Node":
                    if issubclass(a, super_class):
                        # join e.g. ['EditNodes', 'MoleculeEditorNode'] to EditNodes.MoleculeEditorNode
                        dict_key = '.'.join(a.__module__.split(".")[2:])
                        nodes[dict_key] = a.__name__

        return nodes

