# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.
  

import re
import sys
from xml.dom.minidom import parseString
from xml.dom import minidom
from xml.parsers import expat

from System.Data.Parameter import Parameter
from System.Data.ports import NodePort
from System.events import ErrorEvent
from System.Function import Function

# domain exceptions
class MissingTagsError(Exception):
    """Essential tags missing when parsing xml."""
    pass


"""This class looks non-object-oriented because of the xml.dom.minidom that is not object-oriented. There has not been an easy way to inherit from all the methods of minidom and use those in an object-oriented class."""
# Note from semi-competent python coder: this was written by a Perl-coder. This class will be refactured.

def getText(nodelist):
    """
    Params:
           - nodelist : List - List of the nodes contained within the calling tag.
    Return:
           - String - Text that may be useful
           
           Extracts text from tags"""
    
    rc = ""
    regex = re.compile("^\s+|\s+$|\s{2,}")
    
    for node in nodelist:
        if node.nodeType == node.ELEMENT_NODE:
            children = node.childNodes
            for child in children:
                if child.nodeType == child.TEXT_NODE:
                    child.data = regex.sub("",child.data)
                    rc = rc + child.data
                
        elif node.nodeType == node.TEXT_NODE:
            node.data = regex.sub("",node.data)
            rc = rc + node.data
    return rc

def getCompulsoryElement(document, tag, filename):
    """Get an element from DOM-document, or raise an error if not found."""
    elements = document.getElementsByTagName(tag)
    if not elements:
        if filename:
            raise MissingTagsError, "Missing tag %s when parsing file %s." % (tag, filename)
        else:
            raise MissingTagsError, "Missing tag %s when parsing xml." % tag
    else:
        return elements[0]

def handleFunctions(function, filename=""):
    """
    Params:
           - function : DOM object - The function to be parsed
    Return:
           - Function - Representing the function
           """
           
    newFunction = Function()
    newFunction.setName(handleTitle(getCompulsoryElement(function, "title", filename)))
    newFunction.setPlugin(handlePlugin(getCompulsoryElement(function, "plugin", filename)))
    newFunction.setFileName(handleTag(getOptionalElement(function.getElementsByTagName("filename"))))
    newFunction.setDescription(handleDescription(getCompulsoryElement(function, "description", filename)))
    newFunction.setHtmlLink(handleTag(getOptionalElement(function.getElementsByTagName("htmllink"))))
    newFunction.setMenuAddress(handleTag(getOptionalElement(function.getElementsByTagName("menuaddress"))))
    newFunction.set_icon(handleTag(getOptionalElement(function.getElementsByTagName("icon"))))
    newFunction.set_doc_path(handleTag(getOptionalElement(
        function.getElementsByTagName("doc_path"))))
    newFunction.set_sub_objects(handle_sub_objects(\
        function.getElementsByTagName('subobjects')))
 
    inputs = function.getElementsByTagName("input")
    for input_ in inputs:
        handleInput(newFunction, input_)
    outputs = function.getElementsByTagName("output")
    for output in outputs:
        handleOutput(newFunction,output)
        
    return newFunction

def handleTitle(title):
    """
    Params:
           - title : DOM object - To be parsed
    Return:
           - String - The title
           """
    
    return getText(title.childNodes)
    
def handleDescription(d):
    """
    Params:
           - d : DOM object - To be parsed to search for a description
    Return:
           - String - The description
           """
    return getText(d.childNodes)

def handlePlugin(p):
    """
    Params:
           - p : DOM object - To be parsed to search for a plugin
    Return:
           - String - The name of the plugin
           """
    
    return getText(p.childNodes)

def handleInput(node, input_):
    """
    Params:
           - node : Node - The Node that will recieve inputs
           - input_ : DOM object - DOM objects representing an input port
    Handles an input port      """
    
    parameters = input_.getElementsByTagName("parameter")
    for parameter in parameters:
        # [NOTE] The actual Parameter object is not created by this call.
        #        Only the arguments are passed on, and the addParameter() is from Function.py.
        node.addParameter(handleParameter(parameter))

    ports = input_.getElementsByTagName("port")
    for port in ports:
        (variableName,validType,description) = handlePort(port)

        named = 0
        names = port.getElementsByTagName("isNamed")
        if names:
            isNamed = getText(names[0].childNodes)
            if isNamed == "1":
                named = "isNamed=1"

        if variableName:
            node.add_input_port((variableName,validType,description,named))
        
            
def handleOutput(node, output):
    """
    Params:
           - node : Node - The Node that will recieve outputs
           - output : DOM object - DOM objects representing an output port
    Handles an output port      """
    
    ports = output.getElementsByTagName("port")
    for port in ports:
        (variableName,validType,description) = handlePort(port)
        if variableName:
            node.add_output_port((variableName,validType,description))   

def handleParameter(parameter):    
    """
    Params:
           - parameter : DOM object - DOM object representing a parameter tag
    Return:
           - Parameter - The parameter that has been parsed and instanciated
    Handles a parameter and makes an object based on the parsed values       """
    
    
    variableName = str((handleParameterLabel(getOptionalElement(parameter.getElementsByTagName("name")))))
    description = str((handleParameterLabel(getOptionalElement(parameter.getElementsByTagName("description")))))
    type = str((handleParameterType(getOptionalElement(parameter.getElementsByTagName("type")))))
    isNamed = str(handleParameterLabel(getOptionalElement(parameter.getElementsByTagName("isNamed"))))
    restrictions = handleParameterRestrictions(getOptionalElement(parameter.getElementsByTagName("restrictions")))
    
    return (variableName,type,description,restrictions,isNamed)

def handlePort(port):
    """
    Params:
           - port : DOM object - DOM object representing a port
    Return:
           - variableName : String - The name of the port
           - dataType : String - The datatype of the port
           - description : String - A useful description of the port
    Handles a port and returns values needed when instanciating a new NodePort object.
    """
    
    variableName = handleTag(getOptionalElement(port.getElementsByTagName("name")))
    dataType = handleTag(getOptionalElement(port.getElementsByTagName("type")))
    description = handleTag(getOptionalElement(port.getElementsByTagName("description")))
    return (variableName,dataType,description)

def handleTag(tag):
    """
    Params:
           - tag : DOM object - A DOM object that will be used when parsing for a String
    Return:
           - String - The text that has been returned from the parsing
    Handles a general tag.
    """
    
    if tag:
        return getText(tag.childNodes)

def handleParameterLabel(label):
    """
    Params:
           - label :  DOM object - A DOM object that will be used when parsing for the label tag contents
    Return:
           - String - The label
           """
    
    if label:
        return getText(label.childNodes)

def handleParameterType(type):
    """
    Params:
           - label :  DOM object - A DOM object that will be used when parsing for the type tag contents
    Return:
           - String - The datatype of the parameter
           """

    if type:
        return str(getText(type.childNodes))

def handleParameterSingle(parameter,single):
    """
    Params:
           - parameter : Parameter - A Parameter instance that the Single will be added to.
           - single : DOM object - A DOM object that will be used when parsing for the 'single' tag contents
    Parses the single tag       """
    
    if single:
        si = Single(parameter)
        si.setType(single.getAttribute("type"))
        si.setDefault(getText(single.childNodes))
        parameter.setOption(si)
    
def handleParameterMultiple(parameter,multiple):
    """
    Params:
           - parameter : Parameter - A Parameter instance that the Multiple will be added to.
           - multiple : DOM object - A DOM object that will be used when parsing for the 'multiple' tag contents
    Parses the multiple tag       """

    if multiple:
        retval = None
        range = getOptionalElement(multiple.getElementsByTagName("range"))
        if range:
            multipleOption = handleParameterMultipleRange(parameter,range)
        else:
            options = multiple.getElementsByTagName("option")
            sendtmed = multiple.getAttribute("type")
            multipleOption = handleParameterMultipleOptions(parameter,sendtmed,options)
        parameter.setOption(multipleOption)
    
def handleParameterMultipleOptions(parameter,sendtmed,options):
    """
    Params:
           - parameter :  Parameter - A Parameter instance that the MultipleOption will be added to
           - sendtmed : String - The type of the visualisation of the parameter.
           - options : List[DOM object] - All the options the users may choose from.
    Return:
           - MultipleOption - A Multipleoption instance representing the options in the parameter.
           """
           
    if options:
        mo = MultipleOptions(parameter)
        if sendtmed:
            mo.setType(sendtmed)
        for option in options:
            handleParameterMultipleOption(mo,option)
        return mo

def handleParameterMultipleOption(mo,option):
    """
    Params:
           - mo : MultipleOption - The MultipleOption object that this option will be added to.
           - option : DOM object - An object representing the actual option.
    Parses a option tag and adds it to the parameter's options       """
    
    default = option.getAttribute("default")
    if default:
        mo.addDefault(default)
    text = getText(option.childNodes)
    mo.addOption(text)

def handleParameterMultipleRange(parameter,range):
    """
    Params:
           - parameter : Parameter - The parameter that this object will be added to.
           - range : DOM object - Representing the range used when parsing
    Return:
           - MultipleRange - The actual MultipleRange object fount in the XML.
    Parses a range tag and adds it to the parameter object       """
    
    if range:
        mr = MultipleRange(parameter)
        mr.setMax(range.getAttribute("max"))
        mr.setMin(range.getAttribute("min"))
        mr.setStep(range.getAttribute("step"))
        mr.setDefault(range.getAttribute("default"))
        return mr

def handleParameterComment(parameter,comment):
    """
    Params:
            - parameter : Parameter - The parameter that this object will be added to.
            - comment : DOM object - Representing the comment
    Parses a comment and adds it to the parameter object        """
    
    if comment:
        co = Comment(parameter)
        co.setDefault(getText(comment.childNodes))
        parameter.setOption(co)

def handleParameterRestrictions(restr):
    max_, step, default, min_, values = (None, None,None,None,[])    
    if restr:
        max_ = str(getText(restr.getElementsByTagName("max")))
        step = str(getText(restr.getElementsByTagName("step")))        
        default = str(getText(restr.getElementsByTagName("default")))        
        min_ = str(getText(restr.getElementsByTagName("min")))
        
        values = []
        valuesEl = restr.getElementsByTagName("values")
        if valuesEl:
            valueEl = valuesEl[0].getElementsByTagName("value")
            for value in valueEl:
                values.append(str(getText(value.childNodes)))
    return (max_, min_, default, step, values)

def handle_sub_objects(sub_objects):
    """Handles sub object tags. 

    Handles tags on the form:
    <subobjects>
      <object port='X'>car</object>
    </subobjects>

    Returnes an empty dict if no tags were found.

    """
    if not sub_objects:
        return {}

    object_dict = {}
    objects = sub_objects[0].getElementsByTagName('object')

    for object_ in objects:
        port = str(object_.getAttribute('port'))
        value = str(getText([object_])).split('$')

        object_dict[port] = value
        
    return object_dict
    
def getOptionalElement(elements):
    """
    Params:
           - elements : List[DOM element] - A list of elements taken from the XML file.
    Return:
           - DOM element - One element
    Returns the first element in the tag if i exists."""
    
    if elements:
        return elements[0]
    
def build(filename):
    """
    Params:
           - filename : String - The path to the actual file that will be parsed by the XML - parser.
    Return:
           - Function - A function object
    Exception:
           - ValueError - The XML file has errors. The internal xml error message is translated to a ValueError which says where (line no and character no) the error may be
    This is the method that needs to be known on the outside of this class. This method is doing what is needed by this class.
    """
    

    try:
        document = minidom.parse(filename)
    except expat.error, e:
        if expat.ErrorString(e.code) == expat.errors.XML_ERROR_NO_ELEMENTS:
            raise MissingTagsError, "No tags found in %s." % filename
        else:
            raise
    try:
        function_tag = document.getElementsByTagName("function")[0]
    except IndexError:
        raise MissingTagsError, "Missing function tag in file %s." % filename

    return handleFunctions(function_tag, filename)

