# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.Exceptions.ZXMLException import ZXMLException
from System.Data.ModuleObject import ModuleObject
from Library.Util.unrepr import unrepr

from dprint import dprint, flags
from xml.dom import minidom
from System import Tools

from qt import QFont, QColor

class XMLDiagram:

    def __init__(self, node_repository):
        self.nodeRep = node_repository

        self._nodes_by_id = {}
        self._connect_list = {}
        self._output_list = {}
        self._input_list = {}
        self._multi_input_connections = {}

    def build_diagram(self, module_diagram, xmlstring):
        try:
            dom = minidom.parseString(str(xmlstring))
            
        except Exception:
            return False
        
        self._handle_dom_object(module_diagram, dom)
        return True

    def _handle_dom_object(self, module_diagram, dom):
        """A DOM-object encompasses the entire parsed xml-tree.
        
        http://www.w3.org/TR/REC-DOM-Level-1/
        """
        # parse <modulediagram> tags        
        flist = dom.getElementsByTagName("modulediagram")
        if not flist:
            _raise_error("modulediagram")
        module_element = flist.item(0)

        # gets moduleDiagram properties!
        module_diagram.setName(_get_child_text(module_element, "name"))
        module_diagram.setDescription(description=_get_child_text(module_element,
                                                                 "description"))
        _handle_custom_xml(module_element, module_diagram)
        _handle_variables(module_element, module_diagram)
        
        # get Nodes, iterate through them, handle them
        node_elements = module_element.getElementsByTagName("node")
        for node_element in node_elements:
            self._handle_node_element(node_element, module_diagram)

        # get canvascoments tag, and handle the tags inside it
        canvas_comments_tag = module_element.getElementsByTagName("canvascomments")
        if canvas_comments_tag:
            self._handle_canvas_comments(canvas_comments_tag[0], module_diagram)
        
        self._connect_ports()
        self._handle_connectionless_nodelinks(module_diagram)


    def _handle_restriction_element(self, restriction_element, param):
        restriction_element = restriction_element[0]
        max_value = _get_child_text(restriction_element, "max",
                                 optional=True)
        min_value = _get_child_text(restriction_element, "min",
                                 optional=True)
        values = None
        step = _get_child_text(restriction_element, "step", optional=True)
        default = _get_child_text(restriction_element, "default",
                               optional=True)
        # get values element, parse it if it exists
        values_element = restriction_element.getElementsByTagName("values")
        if values_element:
            values_element = values_element[0]
            values = []

            # find all value-data statements and add them
            value_elements = values_element.getElementsByTagName("value") 

            for value in value_elements:
                for child in value.childNodes:
                    if child.nodeType == child.TEXT_NODE:
                        values.append(child.data)        

            # if no values appended; return values to None state
            if not values:
                values = None
        if max_value:
            max_value = int(max_value)
        if min_value:
            min_value = int(min_value)
        if step:
            step = int(step)
        param.restrictions = { 'max' : max_value, 'min' : min_value,
                               'values' : values, 'default' : default,
                               'step' : step }

    def _handle_parameter_element(self, parameter_element, node):
        name = _get_child_text(parameter_element, "name")
        description = _get_child_text(parameter_element, "description")
        param_type = _get_child_text(parameter_element, "type")
        value = _get_child_text(parameter_element, "value")
        is_named = _get_child_text(parameter_element, "isNamed", optional=True)
        
        param = node.addParameter(name, param_type, description, value=value)
        param.setIsNamed(is_named)

        _handle_custom_xml(parameter_element, param)
        _handle_variables(parameter_element, param)

        restriction = parameter_element.getElementsByTagName("restrictions")
        if restriction:
            self._handle_restriction_element(restriction, param)

    def _handle_input_port(self, input_port, node):
        name = _get_child_text(input_port, "name")
        description = _get_child_text(input_port, "description")
        output = _get_child_text(input_port, "outputport")
        is_named = _get_child_text(input_port, "isNamed", optional=True)
        is_optional = _get_child_text(input_port, "isOptional", optional=True)
        _type = _get_child_text(input_port, "type")
        id_ = int(_get_child_text(input_port, "id"))

        try:
            id_ = int(_get_child_text(input_port, "id"))
        except ValueError:
            # The ID tag for input ports must be an integer. It seems
            # like the diagram is corrupted.\nNOTE: Simpy inserting a
            # random integer will not correct the error! The diagram
            # might run, but the results will most likly be corrupted
            # or wrong.
            raise ZXMLException, "Non-integer id-tag for input port."
        # Make sure isOptional is True or False by comparing to
        # True explicitly
        port = node.add_input_port(name, _type, description,
                                   "FORCE,isNamed=%s" % is_named,
                                   is_optional=="True")
        port.set_id(str(id_))
        # Make sure id of next ports created don't collide
        ModuleObject.counter.set_next_value(max(ModuleObject.counter.get_next_value(),
                                                id_ + 1))
        _handle_custom_xml(input_port, port)
        _handle_variables(input_port, port)
        self._input_list[id_] = port
        # if connected: schedule for connecting later :-)
        if output:
            self._connect_list[id_] = int(output)

    def _handle_output_port(self, output_port, node):
        dprint(flags.SYSTEM.XMLREADER, "[XMLReader] Adding outputport on: ",
               node)
        name = _get_child_text(output_port, "name")
        description = _get_child_text(output_port, "description")
        id_ = int(_get_child_text(output_port, "id"))
        _type =  _get_child_text(output_port, "type")

        port = node.add_output_port(name, _type, description, "FORCE")
        port.set_id(str(id_))
        self._output_list[id_] = port
        # Make sure id of next ports created don't collide
        ModuleObject.counter.set_next_value(max(ModuleObject.counter.get_next_value(),
                                                id_ + 1))
                                                
        _handle_custom_xml(output_port, self._output_list[id_])
        _handle_variables(output_port, self._output_list[id_])

    def _handle_multi_input_port(self, port_element, node):
        name = _get_child_text(port_element, "name")
        port = node.add_multi_input_port(name)

        # fix connections
        connected_ports = port_element.getElementsByTagName("connected_port")
        connections = []
        for connected_element in connected_ports:
            connection_id = int(_get_text(connected_element))
            port_id = str(connected_element.getAttribute("id"))
            connections.append((port_id, connection_id))

        self._multi_input_connections[port] = connections
                        
    def _handle_node_element(self, node_element, module_diagram):
        # get regular Node properties
        node = _create_node(module_diagram,
                            _get_child_text(node_element, "class"),
                            _get_child_text(node_element,"name"))
        self._nodes_by_id[_get_child_text(node_element, "id")] = node
        node.setDescription(_get_child_text(node_element, "description"))
        
        _handle_custom_xml(node_element, node)
        _handle_variables(node_element, node)  
        
        outputs = node_element.getElementsByTagName("output").item(0)
        if outputs:
            for output_port in outputs.getElementsByTagName("port"):
                self._handle_output_port(output_port, node)
            
        inputs = node_element.getElementsByTagName("input").item(0)
        if inputs:
            for input_port in inputs.getElementsByTagName("port"):
                self._handle_input_port(input_port, node)
    
            for parameter_element in inputs.getElementsByTagName("parameter"):
                self._handle_parameter_element(parameter_element, node)

            for port_element in inputs.getElementsByTagName("multiport"):
                self._handle_multi_input_port(port_element, node)

        # start reading plot configuration (only 2D plot node for now)
        for tag, handler in (("plotnode_config", handle_plotnode_xml),
                             ("rdissection", _handle_rdissection_xml),
                             ("separator", _handle_text_input_xml),
                             ('query_setup', _handle_query_node_setup_xml)):
            xml = node_element.getElementsByTagName(tag).item(0)
            if xml:
                handler(xml, node)


    def _handle_canvas_comments(self, canvas_comments_tag, module_diagram):
        canvas_comments = canvas_comments_tag.getElementsByTagName("canvascomment")
        for canvas_comment in canvas_comments:            
            xpos = None
            ypos = None

            for (key, value) in canvas_comment.attributes.items():
                if (key == "xpos"):
                    xpos = value
                elif (key == "ypos"):
                    ypos = value
                else:
                    raise ZXMLException, "Unknown attribute found for <canvascomment>-tag"

            # getting text
            text = _get_text(canvas_comment.getElementsByTagName("text")[0])

            # getting font properties
            font_node = canvas_comment.getElementsByTagName("font")[0]

            font_family = _get_text(font_node.getElementsByTagName("family")[0])
            font_point_size = _get_text(font_node.getElementsByTagName("point_size")[0])
            font_weight = _get_text(font_node.getElementsByTagName("weight")[0])
            font_italic = _get_text(font_node.getElementsByTagName("italic")[0])
            font_under_line = _get_text(font_node.getElementsByTagName("under_line")[0])
            font_strike_out = _get_text(font_node.getElementsByTagName("strike_out")[0])

            font_colour_node = font_node.getElementsByTagName("colour")[0]

            font_colour_red = _get_text(font_colour_node.getElementsByTagName("red")[0])
            font_colour_green = _get_text(font_colour_node.getElementsByTagName("green")[0])
            font_colour_blue = _get_text(font_colour_node.getElementsByTagName("blue")[0])

            # getting background colour
            bg_colour_node = canvas_comment.getElementsByTagName("background_colour")[0]

            bg_colour_red = _get_text(bg_colour_node.getElementsByTagName("red")[0])
            bg_colour_green = _get_text(bg_colour_node.getElementsByTagName("green")[0])
            bg_colour_blue = _get_text(bg_colour_node.getElementsByTagName("blue")[0])

            # setting up variables for converting
            int_xpos = int_ypos = str_text = str_font_family = None
            int_font_point_size = int_font_weight = bool_font_italic = None
            bool_font_under_line = bool_font_strike_out = None
            int_font_colour_red = int_font_colour_green = None
            int_font_colour_blue = int_bg_colour_red = None
            int_bg_colour_green = int_bg_colour_blue = None

            every_item_set = False

            # creating a bool for checking if every items is set (except text
            # since this may be empty)
            every_item_set = xpos and ypos and font_family and \
                             font_point_size and font_weight and \
                             font_italic and font_under_line and \
                             font_strike_out and font_colour_red and \
                             font_colour_green and font_colour_blue and \
                             bg_colour_red and bg_colour_green and \
                             bg_colour_blue

            if every_item_set:
                try:
                    int_xpos              = int(xpos)
                    int_ypos              = int(ypos)
                    str_text              = str(text)
                    str_font_family       = str(font_family)
                    int_font_point_size   = int(font_point_size)
                    int_font_weight       = int(font_weight)
                    bool_font_italic      = unrepr(font_italic)
                    bool_font_under_line  = unrepr(font_under_line)
                    bool_font_strike_out  = unrepr(font_strike_out)
                    int_font_colour_red   = int(font_colour_red)
                    int_font_colour_green = int(font_colour_green)
                    int_font_colour_blue  = int(font_colour_blue)
                    int_bg_colour_red     = int(bg_colour_red)
                    int_bg_colour_green   = int(bg_colour_green)
                    int_bg_colour_blue    = int(bg_colour_blue)
                    
                except ValueError:
                    raise ZXMLException, "Tag values could not be converted to proper type in <canvascomment>-tag"
                
                # setting font
                font = QFont(str_font_family, int_font_point_size,
                             int_font_weight, bool_font_italic)
                font.setUnderline(bool_font_under_line)
                font.setStrikeOut(bool_font_strike_out)               

                # setting font colour
                font_colour = QColor(int_font_colour_red,
                                     int_font_colour_green,
                                     int_font_colour_blue)
                
                # setting background colour
                background_colour = QColor(int_bg_colour_red,
                                           int_bg_colour_green,
                                           int_bg_colour_blue)
                
                module_diagram.add_canvas_comment(text, int_xpos, int_ypos,
                                                  font, font_colour,
                                                  background_colour)
            else:
                raise ZXMLException, "Could not locate needed tags in <canvascomment>-tag"           


    def _connect_ports(self):
        for input_port_id in self._connect_list:
            outputport = self._output_list[self._connect_list[input_port_id]]
            inputport = self._input_list[input_port_id]
            inputport.connectToDataSource(outputport)
            
        for port, connections in self._multi_input_connections.items():
            for port_id, output_port_id in connections:
                emulated_port = port.add_data_source(self._output_list[output_port_id])
                if port_id:
                    emulated_port.setId(port_id)

    
    def _handle_connectionless_nodelinks(self, module_diagram):
        old_links = module_diagram.getVariable("connectionless_links")
        if old_links:
            # update node ids
            new_links = old_links
            for id in self._nodes_by_id.keys():
                new_links = new_links.replace(id, self._nodes_by_id[id].getId())
            module_diagram.setVariable("connectionless_links", new_links)


def _get_text(node):
    rc = ""
    flist = node.childNodes
    for node in flist:
        if node.nodeType == node.TEXT_NODE:
            rc = rc + node.data
    return str(rc)

def _get_child_text(node, childname, optional=0):

    # get list of all occurnces of 'childname'
    childlist = node.getElementsByTagName(childname)

    # choose the first occurence of 'childname' and scan it for text-nodes
    if not childlist:
        if optional:
            return None
        else:
            _raise_error(childname)
    return _get_text(childlist.item(0))

def _raise_error(tag):
    print "tag:", repr(tag), str(tag)
    raise ZXMLException, "Error building ModuleDiagram from xml:\nExpected tag '%s' does not exist." % tag

def _handle_custom_xml(element, module_object):
    var = element.getElementsByTagName("CustomXML")
    if var:
        lambdaish = Tools.xml2lambda(var[0])
        xml_string = Tools.lambda2xml(lambdaish)
        module_object.setCustomXML(xml_string)

def _handle_variables(element, module_object):
    variables = element.getElementsByTagName("variables")

    if not variables:
        return 

    for var in variables[0].childNodes:
        if var.nodeType != var.TEXT_NODE:
            module_object.setVariable(var.localName,
                                      _get_child_text(element, var.localName))

def _handle_rdissection_xml(rdissection_xml, rdissect_node):
    """This method will read the filter stored with the r-dissection node"""
    tags = rdissection_xml.getElementsByTagName("item")
    indexlist = []
    metalist = []
    for tag in tags:
        numlist = str(tag.attributes["index"].value)[1:-1].split(",")
        indexlist.append(tuple([int(siffer) for siffer in numlist if siffer]))
        metalist.append(str(tag.attributes["meta"].value))
    rdissect_node.from_xml(zip(indexlist, metalist))

def _handle_text_input_xml(text_input_xml, input_node):
    alfa = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
    tags = text_input_xml.getElementsByTagName("row")
    row = []
    for tag in tags:
        for i in range(8):
            row.append(int(str(tag.attributes[alfa[i]].value)))
        row_custom = str(tag.attributes["custom"].value)
    tags = text_input_xml.getElementsByTagName("col")
    col = []
    for tag in tags:
        for i in range(8):
            col.append(int(str(tag.attributes[alfa[i]].value)))
        col_custom = str(tag.attributes["custom"].value)
    all = {'lists': [row, col], 'custom': [row_custom, col_custom]}
    input_node.from_text_input_xml(all)
        
def handle_plotnode_xml(plotnode_xml, plotnode):
    """This method will read xml recursive through <plotnode_config>-tag"""
    widget_xmls = []
    plot_map_xml = None
    for tag in plotnode_xml.childNodes:
        # ignore TEXT_NODE which is the indent representation in the xml
        if tag.nodeType == tag.TEXT_NODE:
            continue

        if tag.tagName == "horisontal_widgets":
            hor_widgets = int(tag.firstChild.nodeValue)
        elif tag.tagName == "vertical_widgets":
            vert_widgets = int(tag.firstChild.nodeValue)
        elif tag.tagName == "widget_config":
            widget_xmls.append(tag)
        elif tag.tagName == "plot_map_config":
            plot_map_xml = tag
        elif tag.tagName == "input_ports_used":
            input_tags = tag.childNodes
            for input_tag in input_tags:
                if input_tag.nodeType == tag.TEXT_NODE:
                    continue
                if input_tag.tagName != "ip":
                    continue
                attrib_tags = input_tag.childNodes
                tempport = {}
                for attrib_tag in attrib_tags:
                    if attrib_tag.nodeType == tag.TEXT_NODE:
                        continue
                    if attrib_tag.tagName == "name":
                        tempport[1] = attrib_tag.firstChild.nodeValue
                    elif attrib_tag.tagName == "type":
                        if attrib_tag.firstChild:
                            tempport[2] = attrib_tag.firstChild.nodeValue
                        else:
                            tempport[2] = ""
                    elif attrib_tag.tagName == "description":
                        if attrib_tag.firstChild:
                            tempport[3] = attrib_tag.firstChild.nodeValue
                        else:
                            tempport[3] = ""

                plotnode.add_input_port(tempport[1], tempport[2],
                                        tempport[3])

        elif tag.tagName == "output_ports_used":
            output_tags = tag.childNodes
            for output_tag in output_tags:
                if output_tag.nodeType == tag.TEXT_NODE:
                    continue
                if output_tag.tagName != "ip":
                    continue
                attrib_tags = output_tag.childNodes
                tempport = {}
                for attrib_tag in attrib_tags:
                    if attrib_tag.nodeType == tag.TEXT_NODE:
                        continue
                    if attrib_tag.tagName == "name":
                        tempport[1] = attrib_tag.firstChild.nodeValue
                    elif attrib_tag.tagName == "type":
                        if attrib_tag.firstChild:
                            tempport[2] = attrib_tag.firstChild.nodeValue
                        else:
                            tempport[2] = ""
                    elif attrib_tag.tagName == "description":
                        if attrib_tag.firstChild:
                            tempport[3] = attrib_tag.firstChild.nodeValue
                        else:
                            tempport[3] = ""

                plotnode.add_output_port(tempport[1], tempport[2],
                                         tempport[3])

    plotnode.update_widgets(hor_widgets, vert_widgets)
         
    for widget_xml in widget_xmls:
        _handle_plot_widget_xml(widget_xml, plotnode)
        
    if plot_map_xml:
        _handle_plot_map_xml(plot_map_xml, plotnode)

def _parse_as_settings(dom):
    settings = {}
    for tag in dom.childNodes:
        if tag.nodeType != tag.TEXT_NODE:
            if tag.firstChild != None:
                settings[str(tag.tagName)] = str(tag.firstChild.nodeValue)
            else:
                settings[str(tag.tagName)] = ""
    return settings

def _handle_plot_map_xml(plot_map_xml, plotnode):
    """This method will read xml recursive through <widget_config>-tag"""
    map_xmls = []
    go_browser_map_settings = []
    for tag in plot_map_xml.childNodes:
        if tag.nodeType != tag.TEXT_NODE:
            if tag.tagName == "plot_map":
                map_xmls.append(tag)
            elif tag.tagName == "go_browser_map":
                go_browser_map_settings.append(_parse_as_settings(tag))

    plotnode.set_go_browser_map_settings(go_browser_map_settings)

    handler = plotnode.get_plot_map_handler()
    handler.set_enabled(bool(int(_get_child_text(plot_map_xml, "enabled"))))
    handler.set_auto_map_enabled(bool(int(_get_child_text(plot_map_xml,
                                                          "auto_map_enabled"))))
    plotnode.update_plot_maps(len(map_xmls))
    for i, item_xml in enumerate(map_xmls):
        plotnode.set_plot_map_settings(i, _parse_as_settings(item_xml))
    
def _handle_plot_widget_xml(widget_xml, plotnode):
    """This method will read xml recursive through <widget_config>-tag"""
    
    tags = widget_xml.childNodes
    settings = {}
    plot_xmls = []
    for tag in tags:
        if tag.nodeType != tag.TEXT_NODE:
            if tag.tagName == "x_coord":
                x_coord = int(tag.firstChild.nodeValue)
            elif tag.tagName == "y_coord":
                y_coord = int(tag.firstChild.nodeValue)
            elif tag.tagName == "plot_config":
                plot_xmls.append(tag)
                
            else:
                if tag.firstChild != None:
                    settings[str(tag.tagName)] = str(tag.firstChild.nodeValue)
                else:
                    settings[str(tag.tagName)] = ""

    plotnode.update_plots(x_coord, y_coord, len(plot_xmls))
    plotnode.set_widget_settings(x_coord, y_coord, settings)

    for plot_xml in plot_xmls:
        _handle_plot_xml(plot_xml, plotnode, x_coord, y_coord)


def _handle_plot_xml(plot_xml, plotnode, x_coord, y_coord):
    """This method will read xml recursive through <plot_config>-tag"""

    tags = plot_xml.childNodes
    settings = {}
    
    for tag in tags:
        if tag.nodeType != tag.TEXT_NODE:
            if tag.tagName == "index":
                index = int(tag.firstChild.nodeValue)
            elif tag.tagName == "plot_marker_settings":
                _handle_plot_marker_settings_xml(tag, x_coord, y_coord, index,
                                                 plotnode)
            else:
                if tag.firstChild != None:
                    settings[str(tag.tagName)] = str(tag.firstChild.nodeValue)
                else:
                    settings[str(tag.tagName)] = ""

    config = plotnode.get_plot_config(x_coord, y_coord, index)
    for key, value in settings.items():
        config.set_setting(key, value)

def _handle_plot_marker_settings_xml(marker_settings_xml, x_coord, y_coord,
                                     index, plotnode):
    """This method will read xml recursive through <plot_config>-tag"""

    tags = marker_settings_xml.childNodes
    settings = {}
    for tag in tags:
        if tag.nodeType != tag.TEXT_NODE:
            if tag.firstChild:
                value = str(tag.firstChild.nodeValue)
            else:
                value = ''

            settings[str(tag.tagName)] = value

    plotnode.set_marker_settings(x_coord, y_coord, index, settings)

def _handle_query_node_setup_xml(query_node_xml, query_node):
    """Handles the xml tag <query_setup> for UserQueryNode"""
    
    number_of_values = int(query_node_xml.getAttribute('number_of_values'))
    types = str(query_node_xml.getAttribute('types'))
    labels = str(query_node_xml.getAttribute('labels'))
    
    query_node.set_number_of_values(number_of_values)
    if types:
        query_node.set_value_types(types.split(":"))
    if labels:
        query_node.set_value_labels(labels.split(":"))

    if number_of_values == 0:
        return

    types = []
    values = {}

    for object_ in query_node_xml.getElementsByTagName('query_value'):
        id_ = object_.getAttribute('id')
        if object_.firstChild:
            value = str(object_.firstChild.nodeValue)


            if id_ not in values.keys():
                values[id_] = []
                
            values[id_].append(value)

    keys = values.keys()
    keys.sort()
    for key in keys:
        value = values[key]
        if len(value) == 1:
            query_node.add_values(value[0])
        else:
            query_node.add_values(value)

def _create_node(module_diagram, nodetype, nodename):
    nodesubtype = ""
    if nodetype.startswith("FileHandlers.Input."):
        nodesubtype = nodetype.split(".")[-1]
        nodetype = "FileHandlers.InputNode"
        nodename = "FileHandlers.InputNode " + str(nodename.split()[-1])
    if nodetype.startswith("FileHandlers.Output."):
        nodesubtype = nodetype.split(".")[-1]
        nodetype = "FileHandlers.OutputNode"
        nodename = "FileHandlers.OutputNode " + str(nodename.split()[-1])
    node = module_diagram.add_node_by_name(nodetype, nodename)
    if nodesubtype:
        node.setFileHandler(nodesubtype)
    return node

def convert_from_string_to_dom(xml_string):
    try:
        dom = minidom.parseString(str(xml_string))
            
    except Exception:
        return False

    return dom
