function [W,T,P,Q,B] = plsr(X,Y,aMax,mcent)
%PLSR algorithm that handles one or more Y-variables 
%
%[W,T,P,Q,B] = plsr(X,Y,aMax,meanCtr)
%
%Input variables:
%  X/Y:     Input data
%  aMax:    Number of components to calculate
%  mcent: If true, data are mean-centered prior to analysis
%
%Output variables:
%  W:       Loading weights
%  T:       Scores
%  P/Q:     X-/ Y-loadings
%  B:       Regression coefficients
%
%Wold's PLSR is performed. The algorithm is faster when the number
%of columns is larger than the number of rows in X (nVars>nSamples)
%
%Related functions:
%crossval.m, plsrnipals.m, plsrmartens.m, pcr.m

%author L. Gidskehaug, Oct 08, 2002

[nSamples,nVars] = size(X);

if mcent == 1,
  meanX = mean(X);
  meanY = mean(Y);
  X = X - ones(nSamples,1)*meanX;
  Y = Y - ones(nSamples,1)*meanY;
end %if

T = zeros(nSamples,aMax);
W = zeros(nVars,aMax);
E = X;
II = eye(nSamples);
for a=1:aMax
  [U,S,V] = svd(E'*Y,0);
  w = U(:,1);
  t = E*w;
  E = (II-t*t'/(t'*t))*E;
  W(:,a) = w;
  T(:,a) = t;
end %for

P = X'*T*inv(T'*T);
Q = Y'*T*inv(T'*T);
B = W*inv(P'*W)*Q';
